/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import dk.rode.thesis.factorymethod.Factory;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.prototype.StrictCopyable;
import dk.rode.thesis.singleton.SingletonRegistry;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>prototypical registry</i> creates instances of  
  * {@link StrictCopyable copyable} objects based on 
  * {@link #registerPrototype(Class, StrictCopyable) registered} 
  * prototypes. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Using the {@linkplain dk.rode.thesis.prototype Prototype} pattern as a means
  * to create different types of objects is an, perhaps safer, alternative 
  * to creating objects using reflection. This class to what Gamma et al. call
  * a <i>Prototype Manager</i> [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.121]. However, since this registry is 
  * based on class literals as keys to identify the prototype objects, generic types
  * cannot be created without requiring an unchecked cast. The {@link Factory} type 
  * illustrates how reflection can be used in the creation process (of a single type) 
  * in a type safe manner, even for generic types. Type literals could also be 
  * used as the keys to identify the stored prototype objects, which will solve
  * the casting problem, but that would require creation of type literals for all 
  * lookups and creation requests. <p>
  * 
  * The registry enforces a correlation between the class and prototypical object
  * type, but this could be relaxed to allow for adaptors, decorators, and proxies
  * to be registered for a given class. <p>
  * 
  * This registry does not enforce an upper bound on the prototypical objects to
  * create in a manner similar as the {@link SingletonRegistry}, but could 
  * easily have done so.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    PrototypicalFactory
  */
@ParticipantUsage(value = "ConcreteProduct", type = StrictCopyable.class)
@Participant("ConcreteFactory")
public class PrototypicalRegistry {

    /**
      * The prototypical objects to use registered by their class. <p>
      * 
      * Never null.
      */
    private final Map<Class<? extends StrictCopyable<?>>,StrictCopyable<?>> prototypes;
    
    /**
      * No-arg constructor. <p>
      * 
      * Prototypes must be registered with the 
      * {@link #registerPrototype(Class, StrictCopyable)} method.
      */
    public PrototypicalRegistry() {
        this.prototypes = new HashMap<Class<? extends StrictCopyable<?>>,StrictCopyable<?>>();
    }

    /**
      * Registers the prototypical object supplied as {@code prototype} to
      * this factory associated with the type supplied as {@code type}. <p>
      * 
      * {@code prototype} is copied so this factory does not keep a reference to it.
      * 
      * @param  <T>       The type of {@code prototype}. 
      * @param  type      A class literal representing the prototypical type; cannot be null.
      * @param  prototype The prototypical object; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #isRegistered(Class)
      */
    public <T extends StrictCopyable<?>> void registerPrototype(Class<T> type, T prototype) {
        if (type == null) {
            throw new NullPointerException("Type cannot be null");
        }
        this.prototypes.put(type, prototype.copy());
    }
    
    /**
      * Returns true if a prototypical object is registered for the type supplied
      * as {@code type}, false if not.
      * 
      * @param  <T>   The type of {@code prototype}. 
      * @param  type A class literal representing the prototypical type; cannot be null.
      * @return True if registered, false if not.
      * @throws NullPointerException If {@code type} is null.
      * @see    #registerPrototype(Class, StrictCopyable)
      * @see    #getRegisteredTypes()
      */
    public <T extends StrictCopyable<?>> boolean isRegistered(Class<T> type) {
        if (type == null) {
            throw new NullPointerException("Type cannot be null");
        }
        return this.prototypes.containsKey(type);
    }
    
    /**
      * Returns a read-only collection of all prototypical types registered
      * to this registry. <p>
      * 
      * @return The registered types; never null, but can be empty.
      * @see    #isRegistered(Class)
      */
    public Set<Class<? extends StrictCopyable<?>>> getRegisteredTypes() {
        return Collections.unmodifiableSet(this.prototypes.keySet());
    }
    
    /**
      * Creates a new instance of the prototypical type supplied
      * as {@code type}.
      * 
      * @param  <T>  The type of prototype.
      * @param  type The prototype type; cannot be null.
      * @return The new instance; never null.
      * @throws NullPointerException If {@code type} is null.
      */
    public <T extends StrictCopyable<?>> T create(Class<T> type) {
        if (type == null) {
            throw new NullPointerException("Type cannot be null");
        }
        StrictCopyable<?> prototype = this.prototypes.get(type);
        if (prototype == null) {
            throw new NullPointerException("Type not found: " + type.getName());
        }
        T object = type.cast(prototype.copy());
        LogFactory.getLog(this).println("Created object from prototype: ", type.getName(), " -> ", object);
        return object;
    }
        
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; prototypes: ").append(this.prototypes).append(']').toString();
    }
    
}
