/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.abstractfactory;

import dk.rode.thesis.bridge.SequenceValueRange;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>range value factory</i> creates {@link SequenceValueRange value range} 
  * generators. <p>
  *
  * This factory is thread-safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@ParticipantUsage(value = "ConcreteProduct", type = SequenceValueRange.class)
@Participant("ConcreteFactory")
public class RangeValueFactory implements GeneratorFactory<Integer,Integer[]>{

    /**
      * The default range for the created generators. <p>
      * 
      * Length = 1: minimum range, no upper value. <p>
      * 
      * Length = 2: minimum and maximum. <p>
      * 
      * Never null.
      */
    private final Integer[] defaultMinMax;
    
    /**
      * Constructor, which creates this factory to use the
      * minimum value of {@code minimum} for all created
      * generators.
      * 
      * @param minimum The minimum value.
      */
    public RangeValueFactory(int minimum) {
        this.defaultMinMax = new Integer[]{minimum};
    }
    
    /**
      * Constructor, which creates this factory to use the
      * minimum value of {@code minimum} for all created
      * generators.
      * 
      * @param  minimum The minimum value.
      * @param  maximum The maximum value.
      * @throws IllegalArgumentException If {@code minimum} is larger 
      *         than {@code maximum}.
      */
    public RangeValueFactory(int minimum, int maximum) {
        if (minimum > maximum) {
            throw new IllegalArgumentException("Maximum cannot be larger than minimum");
        }
        this.defaultMinMax = new Integer[]{minimum, maximum};
    }

    /**
      * 
      */
    public Integer[] getDefaultArgument() {
        return this.defaultMinMax;
    }
    
    /**
      * 
      */
    public SequenceValueRange createGenerator() {
        return this.createGenerator(this.getDefaultArgument());
    }
    
    /**
      * @param  minMax The range, where a length of one indicates a minimum
      *                only, and a length of two indicates a minimum
      *                and maximum value; can be null, in which case the
      *                default range is used. 
      * @throws IllegalArgumentException If {@code minimum} is larger 
      *         than {@code maximum}.
      */
    public SequenceValueRange createGenerator(Integer[] minMax) {
        if (minMax == null) {
            minMax = this.defaultMinMax;
        }
        SequenceValueRange range = null;
        if (minMax.length == 1) {
            range = new SequenceValueRange(minMax[0]);
        } else {
            range = new SequenceValueRange(minMax[0], minMax[1]);
        }
        LogFactory.getLog(this).println("Created range generator: ", range);
        return range;
    }
    
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        sb.append("; default-minimum: ").append(this.defaultMinMax[0]);
        if (this.defaultMinMax.length > 1) {
            sb.append("; default-maximum: ").append(this.defaultMinMax[1]);
        }
        return sb.append(']').toString();
    }
    
}
