/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.adapter;

import java.math.BigInteger;

/**
  * Simple adapter strategies used in conjunction with
  * {@link SequenceAdapter} instances. <p> 
  *
  * Each strategy is stateless, and thus thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Adapter composition is implemented according to the 
  * {@linkplain dk.rode.thesis.strategy Strategy} pattern. As the explicit 
  * type parameters are important and differs between each concrete 
  * strategy, theses strategies cannot be declared as values in an 
  * enumeration. A benefit of this is that the class is open for
  * sub-classing. <p>
  * 
  * Each strategy is implemented as an anonymous class implementing the 
  * {@link AdapterDelegate} interface, which is by it self a representation 
  * of adaptation. It is common practice in Java to utilise anonymous inner 
  * class to implement adapters. <p>
  * 
  * All strategies for adaptation defined here are <i>state 
  * preserving</i> and imply a simple representation shift <b>except</b> 
  * for the {@link #LONG_TO_INTEGER} and {@link #BIGINTEGER_TO_LONG} strategies, 
  * which may truncate values.   
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class AdapterStrategy {

    /**
      * A strategy to shift from an any type into {@code Object}. <p>
      * 
      * This is useful if wild-cards is to be used.
      */
    public final static AdapterDelegate<Object,Object> ANY_TO_OBJECT = 
        new AdapterDelegate<Object,Object>() {
            public Object adapt(Object value) {
                return value;
            }
            public String toString() {
                return "Any->Object";
            }
        };
    
    /**
      * A strategy to shift from an {@code Object} or wild-card representation
      * to a {@code String} representation by invoking {@code toString()}
      * on the object.
      */ 
    public final static AdapterDelegate<Object,String> ANY_TO_STRING = 
        new AdapterDelegate<Object,String>() {
            public String adapt(Object value) {
                return value == null ? null : value.toString();
            }
            public String toString() {
                return "Any->String";
            }
        };
    
    /**
      * A strategy to shift from a {@code CharSequence} representation
      * to a {@code String} representation.
      */ 
    public final static AdapterDelegate<CharSequence,String> CHAR_TO_STRING = 
        new AdapterDelegate<CharSequence,String>() {
            public String adapt(CharSequence value) {
                return value == null ? null : value.toString();
            }
            public String toString() {
                return "CharSequence->String";
            }
        };

    /**
      * A strategy to shift from a {@code String} representation
      * to a {@code CharSequence} representation via a simple
      * implicit cast.
      */ 
    public final static AdapterDelegate<String,CharSequence> STRING_TO_CHAR = 
        new AdapterDelegate<String,CharSequence>() {
            public CharSequence adapt(String value) {
                return value;
            }
            public String toString() {
                return "String->CharSequence";
            }
        };
        
    /**
      * A strategy to shift from a {@code Long} representation
      * to a {@code String} representation.
      */ 
    public final static AdapterDelegate<Long,String> LONG_TO_STRING = 
        new AdapterDelegate<Long,String>() {
            public String adapt(Long value) {
                return value == null ? null : value.toString();
            }
            public String toString() {
                return "Long->String";
            }
        };

    /**
      * A strategy to shift from a {@code Long} representation
      * to an {@code Integer} representation. <p>
      * 
      * The integer representation is simply truncated. Hence, this
      * strategy <b>cannot</b> be guaranteed to preserve the 
      * entire state, e.g. from 64 to 32 bit values.  
      */ 
    public final static AdapterDelegate<Long,Integer> LONG_TO_INTEGER = 
        new AdapterDelegate<Long,Integer>() {
            public Integer adapt(Long value) {
                return value == null ? null : value.intValue();
            }
            public String toString() {
                return "Long->Integer";
            }
        };
        
    /**
      * A strategy to shift from a {@code BigInteger} representation
      * to a {@code Long} representation. <p>
      * 
      * The long representation is simply truncated. Hence, this
      * strategy <b>cannot</b> be guaranteed to preserve the 
      * entire state.  
      */ 
    public final static AdapterDelegate<BigInteger,Long> BIGINTEGER_TO_LONG = 
        new AdapterDelegate<BigInteger,Long>() {
            public Long adapt(BigInteger value) {
                return value == null ? null : value.longValue();
            }
            public String toString() {
                return "BigInteger->Long";
            }
        };
        
    /**
      * A strategy to shift from a {@code Number} representation
      * to a {@code String} representation.
      */ 
    public final static AdapterDelegate<Number,String> NUMBER_TO_STRING = 
        new AdapterDelegate<Number,String>() {
            public String adapt(Number value) {
                return value == null ? null : value.toString();
            }
            public String toString() {
                return "Number->String";
            }
        };
        
    /**
      * A strategy to shift from a {@code CharSequence} representation
      * to a {@code Long} representation.
      */ 
    public final static  AdapterDelegate<CharSequence,Long> CHAR_TO_LONG = 
        new AdapterDelegate<CharSequence,Long>() {
            public Long adapt(CharSequence value) {
                return value == null ? null : Long.valueOf(value.toString());
            }
            public String toString() {
                return "CharSequence->Long";
            }
        };

    /**
      * No-arg sub-class constructor.
      */
    protected AdapterStrategy() {
        super();
    }
        
}
