/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.adapter;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.facade.FibonacciSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.DigitSequence;
import dk.rode.thesis.meta.model.HexDigitSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;

/**
  * <b>Adapter</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Adapter</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Adapter tests...");
            
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        //
        // Create adapters
        //
        
        List<Sequence<String>> sequences = new ArrayList<Sequence<String>>(); 
            
        // We have to adapt hex digit sequences to be able
        // to be applicable for the appender decorator...
        sequences.add( 
            new AppenderDecorator(
                new SequenceAdapter<CharSequence,String>(
                    new HexDigitSequence(), 
                    AdapterStrategy.CHAR_TO_STRING
                ),
            "Hex: ", null)
        );

        // We have to adapt long sequences to be able
        // to be applicable for the appender decorator...
        sequences.add( 
            new AppenderDecorator(
                new SequenceAdapter<Long,String>(
                    new LongSequence(), 
                    AdapterStrategy.LONG_TO_STRING
                ),
            "Number: ", null)
        );

        // We have to adapt integer sequences to be able
        // to be applicable for the appender decorator. Since
        // Integer is a Number, we can use the Number->String
        // handler...
        PrimeSequence ps = new PrimeSequence(250);
        sequences.add( 
            new AppenderDecorator(
                new SequenceAdapter<Integer,String>(
                    ps, 
                    AdapterStrategy.NUMBER_TO_STRING
                ),
            "Prime: ", null)
        );
        
        // Multiple adaptation...
        FibonacciSequence fs = new FibonacciSequence();
        sequences.add( 
            new AppenderDecorator(
                new SequenceAdapter<Long,String>(
                    new SequenceAdapter<BigInteger,Long>(
                        fs, 
                        AdapterStrategy.BIGINTEGER_TO_LONG
                    ),
                    AdapterStrategy.LONG_TO_STRING
                ),
            "Fibonacci: ", null)
        );
        
        // Multiple adaptation...
        sequences.add( 
            new AppenderDecorator(
                new SequenceAdapter<Long,String>(
                    new SequenceAdapter<CharSequence,Long>(
                        new DigitSequence(), 
                        AdapterStrategy.CHAR_TO_LONG
                    ),
                    // Super type again (i.e. Number is a super type of Long)...
                    AdapterStrategy.NUMBER_TO_STRING 
                ),
            "<[", "]>")
        );
        
        out.heading("Created sequences");
        
        out.println();
        for (int i = 0; i < sequences.size(); i++) {
            out.println("Sequence ", i, ": ", sequences.get(i));
        }
        
        //
        // Usage
        //
        
        out.heading("Sequence usage through class-like adapters");
        
        for (int i = 0; i < 30; i++) {
            out.println("Iteration: " + (i + 1));
            for (int j = 0; j < sequences.size(); j++) {
                Sequence<String> s = sequences.get(j);
                // Advance sequence containing string values...
                String value = s.next();
                out.println(j, ": ", s, " -> ", value);
            }
            out.println();
        }

        out.heading("Sequence usage through object adapters");

        // Adapt prime sequence to an iterator... 
        Iterator<Integer> iterator = new IteratorSequence<Integer>(ps);
        
        out.println();
        out.println("Adapting Sequence to Iterator: ", ps, " -> " + iterator);
        out.println();

        // Use iterator interface, not sequence...
        while (iterator.hasNext()) {
            int i = iterator.next();
            out.println("Iterating ", iterator, " -> ", i);
        }
        out.println();
        out.println("After iteration ", iterator, " -> ", ps);
    
        out.heading("Adapter tests succeeded!");
        
        return true;
    }
    
}
