/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Util.serializableCopy;

import java.io.Serializable;
import java.util.LinkedHashSet;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>sequence value linked hash set</i> stores sequence values 
  * in a {@link java.util.LinkedHashSet java.util.LinkedHashSet} instance. <p>
  * 
  * A sequence value linked hash set is always {@link #ordered() ordered},
  * {@link #sorted() unsorted}, and contains no {@link #duplicates() duplicates}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of serializable values stored in this value set.
  */
@Participant("ConcreteImplementor")
public class SequenceValueLinkedHashSet<E extends Serializable> 
     extends SequenceValueSet<E,LinkedHashSet<E>> {

    /**
      * Constructor, which creates this <i>sequence value linked 
      * hash set</i> to use the {@code LinkedHashSet} instance 
      * supplied as {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * @param  type   The type of values stored in this value set (and 
      *                {@code values}); cannot be null.
      * @param  values The actual {@code LinkedHashSet} instance storing 
      *                the sequence values; cannot be null, empty, or 
      *                contain null values.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, 
      *         contain a null value.
      */
    public SequenceValueLinkedHashSet(Class<E> type, LinkedHashSet<E> values) {
        this(type, values, 0);
    }

    /**
      * Constructor, which creates this <i>sequence value linked 
      * hash set</i> to use the {@code LinkedHashSet} instance 
      * supplied as {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * @param  type   The type of values stored in this value set (and 
      *                {@code values}); cannot be null.
      * @param  values The actual {@code LinkedHashSet} instance storing 
      *                the sequence values; cannot be null, empty, or 
      *                contain null values.
      * @param  number The number pointing the the {@code number}'th value
      *                to be returned by the next call to {@link #get()}.                
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, 
      *         contain a null value, or if {@code number} is illegal.
      */
    private SequenceValueLinkedHashSet(Class<E> type, LinkedHashSet<E> values, int number) {
        super(type, values, number);
    }
    
    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean ordered() {
        return true;
    }
    
    /**
      * Returns false.
      * 
      * @return False.
      */
    public boolean sorted() {
        return false;
    }
    
    /**
      * 
      */
    public SequenceValueLinkedHashSet<E> copy() {
        return new SequenceValueLinkedHashSet<E>(this.type, 
                                                 serializableCopy(this.values),
                                                 this.number);
    }
    
}
