/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Const.EOF;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Set;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>sequence value set</i> stores sequence values in a specific
  * {@link java.util.Set java.util.Set} implementation as specified by the
  * type parameter {@code C}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link #get(int)} and {@link #numberOf(Serializable, int)} methods
  * are not fast. They have a worst-case running time proportional to the
  * number of elements, <i>n</i>, in the value set, i.e. <i>O(n)</i>. However,
  * {@code numberOf} can quickly deduce if a given value is contained in 
  * the value set or not based on a hash lookup. For these simple classes, 
  * this implementation is deemed adequate.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of serializable values stored in this value set.
  * @param  <C> The type of set storing the values in this value set.
  */
@Participant("Implementor")
public abstract class SequenceValueSet<E extends Serializable,C extends Set<E>> 
              extends SequenceValueCollection<E,C> {

    /**
      * The current iterator used to deliver the next value
      * in {@link #first()}, {@link #get()}, or {@link #get(int)}. <p>
      * 
      * Never null.
      */
    private Iterator<E> iterator;
    
    /**
      * Constructor, which creates this <i>sequence value set</i> to
      * use the {@code java.util.Set} instance supplied as {@code values}. 
      * The actual type of {@code values} is determined by the type 
      * parameter {@code C}. <p>
      * 
      * The next value to return by {@link #get()} is the value that 
      * is the {@code number}'th value in this value set, and  
      * this constructor ensures that this value set is initialised 
      * accordingly! <p>
      *
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * @param  type   The type of values stored in this set (and 
      *                {@code values}); cannot be null.
      * @param  values The actual {@code Set} instance storing the sequence 
      *                values; cannot be null, empty, or contain null values.
      * @param  number The number pointing the the {@code number}'th value
      *                to be returned by the next call to {@link #get()}.                
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, 
      *         contain a null value, or if {@code number} is illegal.
      */
    protected SequenceValueSet(Class<E> type, C values, int number) {
        super(type, values, number);
        // Easy to test for null (sorted collections cannot contain null)...
        if (!this.sorted()) {
            if (this.values.contains(null)) {
                throw new IllegalArgumentException("Values cannot contain null");
            }
        }
        if (this.number > 1) {
            this.get(this.number);
        }
    }
     
    /**
      * Returns false.
      * 
      * @return False.
      */
    public final boolean duplicates() {
        return false;
    }
    
    /**
      * 
      */
    public final E first() {
        this.number = 0;
        this.iterator = this.values.iterator();
        return this.get();
    }
    
    /**
      * 
      */
    public final E get() {
        if (this.number >= this.values.size()) {
            this.number = 0;
        }
        this.number++;
        if ((this.iterator == null) || (!this.iterator.hasNext())) {
            this.iterator = this.values.iterator();
        }
        return next(this.iterator, 1);
    }

    /**
      * 
      */
    public final E get(int number) {
        checkNumber(number, true);
        int n = 0;
        if ((this.iterator == null) || (number <= this.number)) {
            this.iterator = this.values.iterator();
            n = number;
        } else {
            n = number - this.number; 
        }
        this.number = number;
        return this.next(this.iterator, n);
    }
       
    /**
      * 
      */
    public int numberOf(E value, int fromNumber) {
        if (!this.checkNumber(fromNumber, false)) {
            return EOF;
        } else if (!this.values.contains(value)) {
            return EOF; 
        }  
        // Sets are unique, so we can at most find E one time in the set...
        Iterator<E> iterator = this.values.iterator();
        int number = 0;
        while (iterator.hasNext()) {
            number++;
            if (value.equals(iterator.next())) {
                return (number < fromNumber) ? EOF : number;
            }
        }
        return EOF;
    }
    
    /**
      * Helper method that will call {@code next()} {@code count}
      * times on {@code iterator}.
      * 
      * @param  iterator The iterator; never null.
      * @param  count    The number of times to call {@code next()}.
      * @return The value from the last call to {@code next()}; never null.
      */
    private final E next(Iterator<E> iterator, int count) {
        E value = null;
        for (int i = 0; i < count; i++) {
            if (iterator.hasNext()) {
                value = iterator.next();
            } else {
                throw new InternalError("No more values");
            }
        }
        return value;
    }
    
}
