/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.CompareExpression;
import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.NextExpression;
import dk.rode.thesis.interpreter.SetExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.interpreter.CompareExpression.Comparison;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>counting comparable expression builder</i> performs the same 
  * functionality as a {@link ComparableExpressionBuilder standard} expression 
  * builder, but also registers the number of times each expression type is 
  * created based on expression {@link Expression#name() names}. <p>
  * 
  * The count per expression name is stored in the {@link #getContext() 
  * context} used, prefixed with {@link #PREFIX count.}, for example
  * {@code count.next} for {@link NextExpression}. The counts can
  * also be directly fetched using the {@link #getExpressionCount()}
  * method. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Since Java does not support multiple functional inheritance, this
  * class cannot inherit both {@link CountingExpressionBuilder} and
  * {@link ComparableExpressionBuilder}. As most of the functionality
  * is implemented in {@code CountingExpressionBuilder}, this is
  * the class chosen to inherit. The missing build-methods (only 3) 
  * thus has to create expressions themselves and not rely on the
  * super class to perform the construction.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The {@code Comparable} type of values the constructed expressions 
  *             operate on and/or produces.
  */
@Participant("ConcreteBuilder")
public class CountingComparableExpressionBuilder<E extends Comparable<? super E>> 
     extends CountingExpressionBuilder<E>
  implements ComparableExpressionBuilder<E> {
    
    /**
      * Constructor. <p>
      * 
      * A local {@link Context context} is used to register created 
      * {@link VariableExpression variables} and store 
      * {@link ConstantExpression constants}.
      * 
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.              
      * @see    #getContext()    
      */
    public CountingComparableExpressionBuilder(Sequence<? extends E> sequence) {
        super(sequence);
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link Context context} supplied as {@code context} is
      * used to register created {@link VariableExpression variables}
      * and store {@link ConstantExpression constants}. <p>
      * 
      * Note, that in case {@code context} has variables registered
      * with names prefixed with {@link #PREFIX}, this may affect
      * the counting performed by this builder (see the class documentation)!
      * 
      * @param  context  The context to use; cannot be null.
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #getContext()                  
      */
    public CountingComparableExpressionBuilder(Context context, Sequence<? extends E> sequence) {
        super(context, sequence);
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new {@link Context context} will be
      * created and used. Hence, all counts are cleared. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public CountingComparableExpressionBuilder(CountingComparableExpressionBuilder<E> builder) {
        super(builder.sequence);
    }
    
    /**
      * 
      */
    public CountingComparableExpressionBuilder<E> copy() {
        return new CountingComparableExpressionBuilder<E>(this);   
    }

    /**
      * 
      */
    public Expression<Boolean> buildSmallerThanExpression(Expression<E> first, Expression<E> second) {
        Expression<Boolean> expression = this.count(new CompareExpression<E>(Comparison.SMALLER, first, second));
        LogFactory.getLog(this).println("Building SMALLER THAN expression: ", expression);
        return expression;
    }
    
    /**
      * 
      */
    public Expression<Boolean> buildGreaterThanExpression(Expression<E> first, Expression<E> second) {
        Expression<Boolean> expression = this.count(new CompareExpression<E>(Comparison.GREATER, first, second));
        LogFactory.getLog(this).println("Building GREATER THAN expression: ", expression);
        return expression;
    }

    /**
      * 
      */
    public Expression<E> buildSetExpression(Expression<E> value) {
        Expression<E> expression = this.count(new SetExpression<E>(this.sequence, value));
        LogFactory.getLog(this).println("Building SET expression: ", expression);
        return expression;
    }
           
}
