/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import static dk.rode.thesis.meta.util.Throwables.toThrowableType;

import java.util.HashMap;
import java.util.Map;

import dk.rode.thesis.interpreter.ConstantExpression;
import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.NextExpression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>counting expression builder</i> performs the same functionality
  * as a {@link StandardExpressionBuilder standard} expression builder,
  * but also registers the number of times each expression type is 
  * created based on expression {@link Expression#name() names}. <p>
  * 
  * The count per expression name is stored in the {@link #getContext() 
  * context} used, prefixed with {@link #PREFIX count.}, for example
  * {@code count.next} for {@link NextExpression}. The counts can
  * also be directly fetched using the {@link #getExpressionCount()}
  * method. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Due to inheritance of the {@link StandardExpressionBuilder}, covariant
  * return types of the precise types of created expressions cannot
  * be made without casting to explicit types, for example to 
  * {@code NextExpression}. This is not always possible if the 
  * implementing types are hidden. <p>
  * 
  * The counts per expression type could easily have been stored in
  * a map instead, but by using the context, we at the same time
  * utilise the functionality exposed by it. <p> 
  * 
  * Counting functionality inspired by the Gamma et al. example 
  * in [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.104].
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CountingComparableExpressionBuilder
  * 
  * @param  <E> The type of value the evaluation of constructed 
  *             expressions produces.
  */
@Participant("ConcreteBuilder")
public class CountingExpressionBuilder<E> 
     extends StandardExpressionBuilder<E> {
    
    /**
      * The prefix used for all variables storing a count
      * for a given expression type. <p>
      * 
      * The value is {@code count.}, including the dot (.).
      */
    public final static String PREFIX = "count.";
    
    /**
      * Constructor. <p>
      * 
      * A local {@link Context context} is used to register created 
      * {@link VariableExpression variables} and store 
      * {@link ConstantExpression constants}.
      * 
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.              
      * @see    #getContext()    
      */
    public CountingExpressionBuilder(Sequence<? extends E> sequence) {
        super(sequence);
    }
    
    /**
      * Constructor. <p>
      * 
      * The {@link Context context} supplied as {@code context} is
      * used to register created {@link VariableExpression variables}
      * and store {@link ConstantExpression constants}. <p>
      * 
      * Note, that in case {@code context} has variables registered
      * with names prefixed with {@link #PREFIX}, this may affect
      * the counting performed by this builder (see the class documentation)!
      * 
      * @param  context  The context to use; cannot be null.
      * @param  sequence The sequence to be manipulated by constructed
      *                  terminal expressions; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @see    #getContext()                  
      */
    public CountingExpressionBuilder(Context context, Sequence<? extends E> sequence) {
        super(context, sequence);
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new {@link Context context} will be
      * created and used. Hence, all counts are cleared. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public CountingExpressionBuilder(CountingExpressionBuilder<E> builder) {
        super(builder.sequence);
    }
    
    /**
      * 
      */
    public CountingExpressionBuilder<E> copy() {
        return new CountingExpressionBuilder<E>(this);   
    }
    
    /**
      * Updates the number of times this builder has seen an expression
      * with the same {@link Expression#name() name} as {@code expression}. <p>
      * 
      * The count per expression name is stored in the {@link #getContext() 
      * context} used, prefixed with {@code count.}.
      * 
      * @param  <V> The type of value the evaluation of {@code expression} produces.
      * @param  <W> The actual type of {@code expression}.
      * @param  expression The expression to increase the count for; cannot be null.
      * @return {@code expression}; never null.
      * @throws NullPointerException If {@code expression} is null.
      */
    protected <V,W extends Expression<V>> W count(W expression) {
        try {
            String name = PREFIX + expression.name();
            VariableExpression<Integer> ve = new VariableExpression<Integer>(Integer.class, name);
            int count = 1;
            if (this.context.hasValue(ve)) {
                count = this.context.lookup(ve) + 1;
            }
            this.context.assign(ve, count);
            return expression;
            
        } catch (ExpressionException ee) {
            throw new RuntimeException(ee);
        }
    }
    
    /**
      * Returns a map containing as values the number of times
      * a given expression type has been created, excluding the
      * count for the {@link #getRootExpression() root} 
      * expression. <p>
      * 
      * The keys correspond to an expression {@link Expression#name() name}.
      * 
      * @return A map containing the count; never null, but can be
      *         empty.
      * @throws ExpressionException If the generation of the map fails.          
      */
    public Map<String,Integer> getExpressionCount() throws ExpressionException {
        try {
            Map<String,Integer> counts = new HashMap<String,Integer>();
            for (VariableExpression<?> ve : this.context) {
                // If the variable name starts with the prefix, we assume
                // it is a variable used for the counting, i.e. Integer type...
                if (ve.name.startsWith(PREFIX)) {
                    Object value = this.context.lookup(ve);
                    counts.put(ve.name.substring(PREFIX.length()), Integer.class.cast(value));
                }
            }
            return counts;
            
        } catch (Exception e) {
            throw toThrowableType(e, ExpressionException.class);
        }
    }
    
    //
    // Non Terminal Expressions
    //

    // Boolean expressions
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return this.count(super.buildAndExpression(first, second));
    }
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildNonShortCircuitAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return this.count(super.buildNonShortCircuitAndExpression(first, second));
    }
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return this.count(super.buildOrExpression(first, second));
    }
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildNonShortCircuitOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return this.count(super.buildNonShortCircuitOrExpression(first, second));
    }
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildNotExpression(Expression<Boolean> expression) {
        return this.count(super.buildNotExpression(expression));
    }
    
    /**
      * 
      */
    @Override
    public Expression<Boolean> buildEqualExpression(Expression<?> first, Expression<?> second) {
        return this.count(super.buildEqualExpression(first, second));
    }
       
    // Flow, break, conditional, and assignment
    
    /**
      * 
      */
    @Override
    public Expression<E> buildInitialisedFlowExpression(Expression<? extends E>... expressions) throws ExpressionException {
        return this.count(super.buildInitialisedFlowExpression(expressions));
    }

    /**
      * 
      */
    @Override
    public FlowExpression<E> buildFlowExpression() {
        return this.count(super.buildFlowExpression());
    }
    
    /**
      * 
      */
    @Override
    public Expression<E> buildBreakExpression(TypedExpression<E> expression) {
        return this.count(super.buildBreakExpression(expression));
    }

    /**
      * 
      */
    @Override
    public Expression<E> buildConditionalExpression(Expression<Boolean> condition, Expression<? extends E> first, Expression<? extends E> second) {
        return this.count(super.buildConditionalExpression(condition, first, second));
    }

    /**
      * 
      */
    @Override
    public Expression<E> buildAssignmentExpression(VariableExpression<E> variable, Expression<? extends E> expression) {
        return this.count(super.buildAssignmentExpression(variable, expression));
    }
    
    
    // Variables and constants

    /**
      * 
      */
    @Override
    public <V> VariableExpression<V> buildVariableExpression(Class<V> type, String name) throws ExpressionException {
        VariableExpression<V> ve = null;
        if (this.context.exist(name)) {
            ve = this.context.getVariable(type, name);
            LogFactory.getLog(this).println("Reusing VARIABLE expression: ", ve);
        } else {
            this.context.register(ve = new VariableExpression<V>(type, name));
            LogFactory.getLog(this).println("Building VARIABLE expression: ", ve);
            // Only count on creation, not on reuse...
            this.count(ve);
        }
        return ve;
    }

    /**
      * 
      */
    @Override
    public <V> Expression<V> buildConstantExpression(Class<V> type, V value) throws ExpressionException {
        VariableExpression<V> ce = new ConstantExpression<V>(type, value);
        if (this.context.exist(ce)) {
            ce = this.context.getVariable(type, ce.name);
            LogFactory.getLog(this).println("Reusing CONSTANT expression: ", ce);
        } else {
            this.context.assign(ce, value);
            LogFactory.getLog(this).println("Building CONSTANT expression: ", ce);
            // Only count on creation, not on reuse...
            this.count(ce);
        }
        return ce;
    }
    
    //
    // Terminal expressions
    //

    /**
      * 
      */
    @Override
    public Expression<E> buildCurrentExpression() {
        return this.count(super.buildCurrentExpression());
    }

    /**
      * 
      */
    @Override
    public Expression<E> buildResetExpression() {
        return this.count(super.buildResetExpression());
    }

    /**
      * 
      */
    @Override
    public Expression<Boolean> buildReverseExpression(Expression<Boolean> reverse) {
        return this.count(super.buildReverseExpression(reverse));
    }

    /**
      * 
      */
    @Override
    public Expression<E> buildNextExpression(Expression<? extends Number> count) {
        return this.count(super.buildNextExpression(count));
    }
    
}
