/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.ExpressionException;
import dk.rode.thesis.interpreter.FlowExpression;
import dk.rode.thesis.interpreter.InitialisableExpression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.interpreter.TypedExpressionDecorator;
import dk.rode.thesis.interpreter.VariableExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.proxy.ProxyFactory;

/**
  * A <i>typed expression builder</i> builds {@link TypedExpression typed}
  * expressions. <p>
  * 
  * An actual builder must be supplied at construction to perform the
  * construction of expressions, but this builder will use a local
  * {@link #getContext() context}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This builder uses covariant types for the created expressions
  * where possible. The expression type is narrowed to {@link TypedExpression}.
  * Narrowing it even further, i.e. to the exact expression type, is 
  * not always possible if the implementation types are hidden: how
  * can the builder cast to the exact type if only the interfaces are
  * known? As {@code TypedExpression} has a concrete decorator class,
  * {@link TypedExpressionDecorator}, the narrowing to {@code TypedExpression}
  * can be performed because an implementation is always known. <p>
  * 
  * This presents a problem with the {@link #buildFlowExpression()} method 
  * because it already has a narrowed return type, namely {@link FlowExpression},
  * <b>and</b> because it is created via composition; this builder has no control 
  * over the creation process. The {@code FlowExpression} type must not be lost, 
  * or the functionality offered by it cannot be used by the client.
  * This is not a problem for the {@link #buildInitialisedFlowExpression(Expression...)}
  * method, because it returns an {@code Expression} type like most other
  * builder methods. <p>
  *   
  * Inheritance, proxies, and/or decoration/adaptation must be used. Dynamic proxies can 
  * only be used of an interface is used. Decoration and adaptation can only be used 
  * if a non-final type is known. Inheritance only if a concrete class is known, 
  * regardless if an interface is known. As {@code FlowExpression} is concrete, 
  * but non final, we have to make a decorator wrapping the created 
  * {@code FlowExpression}, overriding all methods, and adapting it to the 
  * {@code TypedExpression} interface. This is tedious at best, but cannot be avoided.
  * The decorator class is {@link TypedFlowExpression}. <p>
  * 
  * Dynamic proxies created by a {@link ProxyFactory} cannot be used for several reasons. 
  * First, {@code FlowExpression} is a concrete type, so casting a proxy into the return 
  * type required by the {@code buildFlowExpression()} method will cause a class cast 
  * exception. Secondly, {@code FlowExpression} already implements 
  * {@link InitialisableExpression} that use covariant return types to refine its 
  * {@code copy()} method. {@code TypedExpression} also use covariant return types to 
  * to refine its {@code copy()} method. This works for {@code FlowExpression} because 
  * it is a concrete type known to implement both interfaces, but cannot work for dynamic 
  * proxies because if incompatible return types. See the JavaDoc for 
  * {@link TypedFlowExpression#copy()} to get an idea of the different narrowed return
  * types that fits the {@code TypedFlowExpression} class. <p>
  * 
  * Instead of inheritance, this builder relies on composition: the
  * actual builder to construct the expressions is supplied at construction
  * time. This can be viewed as an application of the 
  * {@linkplain dk.rode.thesis.decorator Decorator} pattern, and perhaps also
  * as a degenerate version of the {@linkplain dk.rode.thesis.bridge Bridge}
  * pattern where the abstraction and implementation share the same
  * interface.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of constructed 
  *             expressions produces.
  */
@Participant("ConcreteBuilder")
public class TypedExpressionBuilder<E> 
     extends AbstractExpressionBuilder<E>
  implements ExpressionBuilder<E> {
    
    /**
      * A class literal representing the type of value the 
      * evaluation of constructed expressions produces.
      * 
      * Supplied at construction time, never null.
      */
    protected final Class<E> type;
    /**
      * The expression builder used to construct the expressions. <p>
      * 
      * Never null.
      */
    protected final ExpressionBuilder<E> builder;
    
    /**
      * Constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new local {@link Context context} will be
      * created and used.
      *  
      * @param  type    The class literal representing the type of value the 
      *                 evaluation of constructed expressions produces; cannot
      *                 be null.
      * @param  builder The actual builder to use; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public TypedExpressionBuilder(Class<E> type, ExpressionBuilder<E> builder) {
        super(builder.getSequence()); // use local context
        if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        } 
        this.builder = builder;
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new local {@link Context context} will be
      * created and used. 
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public TypedExpressionBuilder(TypedExpressionBuilder<E> builder) {
        super(builder);
        this.type = builder.type;
        this.builder = builder.builder;
    }
    
    /**
      * 
      */
    public TypedExpressionBuilder<E> copy() {
        return new TypedExpressionBuilder<E>(this);   
    }

    /**
      * Returns a class literal representing the type of value the 
      * evaluation of constructed expressions produces.
      * 
      * @return The type; never null.
      */
    public final Class<E> type() {
        return this.type;
    }
    
    /**
      * Returns a {@link TypedExpression typed} expression of the
      * typed supplied as {@code type} based on {@code expression}.
      *
      * @param  <T>        The type.
      * @param  expression The expression to associate {@code type} with,
      *                    if not already; cannot be null.
      * @param  type       The type; cannot be null.
      * @return A typed expression; never null.
      * @throws NullPointerException If either argument is null.
      */
    protected final static <T> TypedExpression<T> getTypedExpression(Expression<T> expression, Class<T> type) {
        if (expression instanceof TypedExpression) {
            return (TypedExpression<T>)expression;
        } else {
            // Includes type information in symbol representation...
            return new TypedExpressionDecorator<T>(expression, type, true);
        }
    }
    
    //
    // Non Terminal Expressions
    //

    // Boolean expressions
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return getTypedExpression(this.builder.buildAndExpression(first, second), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildNonShortCircuitAndExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return getTypedExpression(this.builder.buildNonShortCircuitAndExpression(first, second), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return getTypedExpression(this.builder.buildOrExpression(first, second), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildNonShortCircuitOrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        return getTypedExpression(this.builder.buildNonShortCircuitOrExpression(first, second), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildNotExpression(Expression<Boolean> expression) {
        return getTypedExpression(this.builder.buildNotExpression(expression), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildEqualExpression(Expression<?> first, Expression<?> second) {
        return getTypedExpression(this.builder.buildEqualExpression(first, second), Boolean.class);
    }
       
    // Flow, break, conditional, and assignment
    
    /**
      * 
      */
    public TypedExpression<E> buildInitialisedFlowExpression(Expression<? extends E>... expressions) throws ExpressionException {
        return getTypedExpression(this.builder.buildInitialisedFlowExpression(expressions), this.type);
    }
    
    /**
      * 
      */
    public TypedFlowExpression<E> buildFlowExpression() { // must use explicit type!
        FlowExpression<E> fe = this.builder.buildFlowExpression();
        // Could already be typed...
        if (fe instanceof TypedFlowExpression) {
            return (TypedFlowExpression<E>)fe;
        }
        return new TypedFlowExpression<E>(this.type, fe);
    }
    
    /**
      * 
      */
    public TypedExpression<E> buildBreakExpression(TypedExpression<E> expression) {
        return getTypedExpression(this.builder.buildBreakExpression(expression), expression.type());
    }
    
    /**
      * 
      */
    public TypedExpression<E> buildConditionalExpression(Expression<Boolean> condition, Expression<? extends E> first, Expression<? extends E> second) {
        return getTypedExpression(this.builder.buildConditionalExpression(condition, first, second), this.type);
    }

    /**
      * 
      */
    public TypedExpression<E> buildAssignmentExpression(VariableExpression<E> variable, Expression<? extends E> expression) {
        return getTypedExpression(this.builder.buildAssignmentExpression(variable, expression), variable.type());
    }
    
    //
    // Terminal expressions
    //

    /**
      * 
      */
    public TypedExpression<E> buildCurrentExpression() {
        return getTypedExpression(this.builder.buildCurrentExpression(), this.type);
    }

    /**
      * 
      */
    public TypedExpression<E> buildResetExpression() {
        return getTypedExpression(this.builder.buildResetExpression(), this.type);
    }

    /**
      * 
      */
    public TypedExpression<Boolean> buildReverseExpression(Expression<Boolean> reverse) {
        return getTypedExpression(this.builder.buildReverseExpression(reverse), Boolean.class);
    }

    /**
      * 
      */
    public TypedExpression<E> buildNextExpression(Expression<? extends Number> count) {
        return getTypedExpression(this.builder.buildNextExpression(count), this.type);
    }
    
}
