/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.chainofresponsibility;

import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

/**
  * A <i>standard handler chain</i> is the standard implementation
  * of the {@link HandlerChain} interface. <p> 
  * 
  * A standard handler chain is <b>not</b> thread safe. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    WeakHandlerChain
  * 
  * @param  <R> The type of requests handled by this chain.
  */
public class StandardHandlerChain<R> extends AbstractHandlerChain<R> {

    /**
      * The ordered list of handlers, where the first handler
      * is the first inserted element. <p>
      *
      * <b>Implementation notes</b>:<br>
      * For small chains, a simple list will suffice, but it does 
      * not scale well: if each handler starting from handler 
      * <i>m</i> in the chain reinvokes the chain, we get a 
      * running time of <i>O(n*m)</i>, i.e. worst time <i>O(n*n)</i>, 
      * where <i>n</i> is the chain length. Existence, registration, 
      * and unregistration takes <i>O(n)</i>. <p>
      * 
      * A better solution is to use a linked hash set to avoid 
      * <i>O(n)</i> existence, registration, and unregistration 
      * time. This is the implementation here. <p>
      * 
      * An even faster and more scalable solution would be to use an index 
      * into a list, pointing to the start of the sub-list to return. 
      * A {@link java.util.NavigableMap navigable} map (tree map) can be used 
      * for this, i.e. {@code tailMap(..)} methods. For this simple example, 
      * we are satisfied with our simple solution. <p> 
      * 
      * Never null.
      */
    private final LinkedHashSet<Handler<R>> handlers;

    /**
      * No-arg constructor.
      */
    public StandardHandlerChain() {
        this.handlers = new LinkedHashSet<Handler<R>>(); 
    }
    
    /**
      * 
      */
    public List<Handler<R>> getHandlers(Handler<R> handler) {
        List<Handler<R>> handlers = new LinkedList<Handler<R>>();
        boolean add = (handler == null) || (!this.isRegistered(handler));
        for (Handler<R> h : this.handlers) {
            if (add) {
                handlers.add(h);
            // Do not add until next iteration...
            } else if (h.equals(handler)) {
                add = true;
            }
        }
        return handlers;
    }
    
    /**
      * 
      */
    public boolean isRegistered(Handler<R> handler) {
        return this.handlers.contains(handler);
    }
    
    /**
      * 
      */
    public int size() {
        return this.handlers.size();
    }
    
    /**
      * 
      */
    public void register(Handler<R> handler) {
        if (this.handlers.contains(handler)) {
            throw new IllegalArgumentException("Handler already registered");
        }
        this.handlers.add(handler);
    }
    
    /**
      * 
      */
    public void unregister(Handler<R> handler) {
        if (!this.handlers.contains(handler)) {
            throw new IllegalArgumentException("Handler not registered");
        }
        this.handlers.remove(handler);
    }
    
}
