/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import static dk.rode.thesis.meta.util.Const.SERIAL_VERSION_UID;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An <i>evil command</i> is a <b>test command</b> that will 
  * <b>always</b> fail during {@link #execute() execution} and perhaps 
  * on {@link #undo() undo} as well by specifying that it is 
  * {@link #isUndoable() undoable} while in reality it is not. <p>  
  *
  * The same instance of a {@link CommandException} is always thrown
  * to be easily identifiable in testing scenarios. <p> 
  * 
  * It still utilises a type parameter, {@code E}, so it can mimic 
  * commands delivering a certain type of value. <p>
  * 
  * Evil commands are thread-safe and can be shared.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of mimicked result obtained by executing 
  *             this command. This command type always return null 
  *             as the result.
  */
@Participant("ConcreteCommand")
public class EvilCommand<E> implements Command<E> {

    /**
      * The exception type thrown during {@link EvilCommand#execute() 
      * execution} and perhaps on {@link EvilCommand#undo() undo} of the
      * an evil command. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This exception type applies the {@linkplain dk.rode.thesis.prototype Singleton} 
      * pattern in its design. 
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class EvilCommandException extends CommandException {

        /**
          * Serial version id. 
          */
        final static long serialVersionUID = SERIAL_VERSION_UID;

        /**
          * Private constructor.
          */
        private EvilCommandException() {
            super("Evil command failed >:)");
        }
        
    }
    
    /**
      * This command exception instance always thrown during
      * the {@link #execute() execution} add perhaps during 
      * {@link #undo() undo} as well. <p>
      * 
      * Never null.
      */
    public final static EvilCommandException ERROR = new EvilCommandException();
    
    /**
      * True if this command will fail on undo, false if it 
      * only fails on execution. 
      */
    private final boolean failOnUndo;

    /**
      * Constructor. <p>
      * 
      * This evil command will not fail on undo.
      */
    public EvilCommand() {
        this(false);
    }
    
    /**
      * Constructor.
      * 
      * @param  failOnUndo True to fail on undo, false to only fail
      *                    during execution.
      */
    public EvilCommand(boolean failOnUndo) {
        this.failOnUndo = failOnUndo;
    }

    /**
      * 
      */
    public EvilCommand<E> copy() {
        return new EvilCommand<E>(this.failOnUndo);
    }
    
    /**
      * Always throws {@link #ERROR}.
      * 
      * @return Never returns.
      * @throws EvilCommandException Always thrown.
      */
    public List<Command<E>> execute() throws EvilCommandException {
        throw ERROR;
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean isUndoable() {
        return true;
    }

    /**
      * If this command is set to fail on undo, {@link #ERROR}
      * is thrown. <p>
      * 
      * Otherwise null is returned.
      * 
      * @return null if this command is not set to fail on undo.
      * @throws EvilCommandException If this command is set to fail on
      *         undo.
      */
    public E undo() throws EvilCommandException {
        if (this.failOnUndo) {
            throw ERROR;
        }
        return null;
    }

    /**
      * @return Always null. 
      */
    public E getResult() {
        return null;
    }
    
    /**
      * Returns the hash code of this evil command. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return Boolean.valueOf(this.failOnUndo).hashCode();
    }
    
    /**
      * Returns true if {@code object} is an evil command
      * set to fail in the same way as this command, otherwise
      * fails. 
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof EvilCommand)) {
            return false;
        }
        return this.failOnUndo == ((EvilCommand<?>)object).failOnUndo;
    }
    
    /**
      * Returns the string representation of this evil command.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; fail-on-undo: ").append(this.failOnUndo).append(']').toString();
    }
    
}
