/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>log command</i> simply logs the operation performed
  * on the receiver object. <p>
  * 
  * A log command carries no state, and is considered  
  * {@link #isUndoable() undoable} as to not interfere with
  * regular commands. <p>
  *
  * Log commands are thread-safe and can be shared as they
  * do not alter the receiver, only prints it to the log. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of result obtained by executing this command. 
  *             This command type always return null as the result.
  */
@Participant("ConcreteCommand")
public class LogCommand<E> implements Command<E> {
    
    /**
      * The receiver object. <p>
      * 
      * Never null.
      */
    @Participant("Receiver")
    private final Object receiver;
    
    /**
      * Constructor.
      * 
      * @param  receiver The receiver object; cannot be null.
      * @throws NullPointerException If {@code receiver} is null.
      */
    public LogCommand(@Participant("Receiver") Object receiver) {
        super();
        if ((this.receiver = receiver) == null) {
            throw new NullPointerException("Receiver cannot be null");
        }
    }

    /**
      * 
      */
    public LogCommand<E> copy() {
        return new LogCommand<E>(this.receiver);
    }
    
    /**
      * Logs the string representation of the receiver object
      * followed by "{@code .execute()}".
      * 
      * @return Always null.
      */
    public List<Command<E>> execute() {
        LogFactory.getLog(this).println(this.receiver, ".execute()");
        return null;
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean isUndoable() {
        return true;
    }

    /**
      * Logs the string representation of the receiver object
      * followed by "{@code .undo()}".
      * 
      * @return Always null.
      */ 
    public E undo() {
        LogFactory.getLog(this).println(this.receiver, ".undo()");
        return null;
    }
    
    /**
      * @return Always null. 
      */
    public E getResult() {
        return null;
    }
    
    /**
      * Returns the hash code of this log command. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return this.receiver.hashCode();
    }
    
    /**
      * Returns true if {@code object} is an instance
      * of a this class and has the same receiver (compared
      * using {@code equals(Object)}.
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof LogCommand)) {
            return false;
        }
        return this.receiver.equals(((LogCommand<?>)object).receiver);
    }
    
    /**
      * Returns the string representation of this log command.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; receiver: ").append(this.receiver).append(']').toString();
    }
    
}
