/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;

/**
  * A <i>reset command</i> command will invoke {@link Sequence#reset() 
  * reset()} on its receiving {@link Sequence sequence} when it is 
  * {@link #execute() executed}. <p>
  * 
  * Reset commands are {@link #isUndoable() undoable} if the receiving 
  * sequence is a {@link MemorizableSequence memorizable} sequence,
  * and/or if it is {@link Sequence#consistent() consistent} and
  * {@link Sequence#unique() unique}. Notice, however, that the
  * time to undo a non-memorizable sequence by this command is not
  * constant: it is linear with the number of times {@link Sequence#next()
  * next()} has to be invoked before the old sequence value is found
  * again.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of result obtained by executing this command.
  *             The type corresponds to the type of values delivered by
  *             the receiver sequence. 
  */
@Participant("ConcreteCommand")
public class ResetCommand<E> extends SequenceCommand<E> {
    
    /**
      * The sequence value before the last reset. <p>
      * 
      * Can be null.
      */
    private E previous;
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ResetCommand(@Participant("Receiver") Sequence<E> sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public ResetCommand<E> copy() {
        return new ResetCommand<E>(this.sequence);
    }
    
    /**
      * Resets the sequence value and sets the result to the (new) 
      * current value.
      */
    protected List<Command<E>> doExecute(boolean saved) {
        // Only need to save previous value in case the sequence
        // is not memorizable...
        if (!saved) {
            this.previous = this.sequence.current();
        }
        this.sequence.reset();
        this.result = this.sequence.current();
        return null;
    }

    /**
      * Returns true if this command has been executed and the 
      * receiving sequence used by this command is either 
      * {@link MemorizableSequence memorizable} and/or the sequence 
      * is {@link Sequence#bounded() bounded} and 
      * {@link Sequence#consistent() consistent}. 
      * 
      * @return True if undoable, false if not.
      */
    public boolean isUndoable() {
        return super.isUndoable()            || 
               ((this.sequence.unique())     && 
                (this.sequence.consistent()) && 
                (this.previous != null));
    }

    /**
      *  
      */
    public E undo() throws CommandException {
        // Try undo using super class, i.e. mementos...
        E value = super.undo();
        if (value != null) {
            return value;
        }
        Log out = LogFactory.getLog(this);
        // Invariant: will have a "previous" value here...
        out.println("Undoing consistent unique sequence: ", this.sequence, ": ", this.sequence.current(), " != ", this.previous);
        int time = 0, restart = 0;
        while (!(value = this.sequence.next()).equals(this.previous)) {
            if (time++ % 25 == 24) {
                out.warn("next() called ", time, " times, but still no match: ", this.sequence.current(), " != ", this.previous);
            }
            if (this.sequence.state() == State.RESTART) {
                if (++restart == 2) {
                    out.error("Previous value cannot be found in sequence: ", this.sequence, " -> ", this.previous);
                    throw new CommandException("Cannot find previous value to redo: " + this.previous);
                }
            }
        }
        out.println("Undoing time: ", time, " call(s) to next()");
        this.previous = null;
        return value;
    }
    
}
