/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Command</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.233]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Avoid coupling the sender of a request to its receiver by giving 
  *    more than one object a chance to handle the request. Chain the 
  *    receiving objects and pass the request along the chain until an 
  *    object handles it.
  *    
  * </i></ul>
  *  
  * Here, the {@link dk.rode.thesis.command.Command} interface 
  * represents the <b>Command</b> participant, and any concrete 
  * {@code Command} implementation represent the <b>ConcreteCommand</b>
  * participant, for example {@link dk.rode.thesis.command.NextCommand}
  * and {@link dk.rode.thesis.command.ReverseCommand}. <p> 
  * 
  * The <b>Receiver</b> participant is represented by instances of 
  * {@link dk.rode.thesis.meta.model.Sequence} for all
  * commands that operate on sequences, i.e. sub-classes of 
  * {@link dk.rode.thesis.command.SequenceCommand},
  * while it can be any object for {@link dk.rode.thesis.command.LogCommand}.
  * {@link dk.rode.thesis.command.NullCommand} 
  * and {@link dk.rode.thesis.command.EvilCommand} are
  * receiver-less. <p>
  * 
  * The {@link dk.rode.thesis.command.CommandProcessor} class 
  * represents the <b>Invoker</b> participant as well as implement a
  * local command <i>history list</i> for undo operations as described
  * by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.238]. <p>
  * 
  * The <b>Client</b> participant is represented by the test application, 
  * i.e. the {@link dk.rode.thesis.command.Main} class,
  * which creates commands and asks the command processor to execute (and
  * possibly undo) them. <p> 
  *   
  * Finally, {@link dk.rode.thesis.command.CompositeCommand} is 
  * a <i>macro command</i> as described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.241]. <p>
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/Command.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@code CommandProcessor} is a variant of the POSA Command Processor pattern 
  * suggested by Schmidt et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Schmidt00" target="thesis-bibliography">Schmidt00</a>, p.397] that also handles <i>commands 
  * spawned by other commands</i> during execution. This enables commands to express 
  * business-like semantics that can alter the behaviour at runtime depending on data. 
  * The commands to spawn can either be supplied at construction time or could be 
  * created on the fly. Here, the {@link dk.rode.thesis.command.ReverseCommand} may 
  * choose to spawn commands during execution, supplied at construction time, if any. <p>
  * 
  * The default strategy for undoing commands with {@code Sequence} objects
  * as receivers is to use the {@linkplain dk.rode.thesis.memento Memento} pattern: 
  * before execution, a {@link dk.rode.thesis.memento.SequenceMemento} will be 
  * created, if the receiver sequence is an instance of the 
  * {@link dk.rode.thesis.memento.MemorizableSequence} type, which can be then 
  * used to reset the original sequence value on undo. To handle non-memorizable
  * sequences, specific command types may naturally employ other means to support 
  * undo, as is for example the case with the {@link dk.rode.thesis.command.ResetCommand}.     
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Command", 
         scope        = Scope.Object,
         purpose      = Purpose.Behavioural,
         participants = {"Command", "ConcreteCommand", "Client", "Invoker", "Receiver"})
package dk.rode.thesis.command;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;






