/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.composite;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>abstract composite sequence</i> implements the basic traits
  * of the {@link CompositeSequence} interface. <p>
  * 
  * This implementation is <b>not</b> thread-safe.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this composite sequence.
  */
@Participant("Composite")
public abstract class AbstractCompositeSequence<E> extends AbstractSequence<E> implements CompositeSequence<E> {

    /**
      * The sequences associated with this sequence, in order. <p>
      * 
      * Is a random access list for fast traversal. <p>
      * 
      * Never null, but can be null.
      */
    protected final List<Sequence<? extends E>> sequences;
    /**
      * The strategy used for traversal starting from <b>this</b> 
      * composite unless a strategy is specifically requested. <p>
      * 
      * Never null.
      */
    private CompositeStrategy strategy;

    /**
      * No-arg constructor. <p>
      * 
      * The {@link CompositeStrategy#DEFAULT} strategy is used
      * for traversal.
      */
    public AbstractCompositeSequence() {
        this((CompositeStrategy)null);
    }
    
    /**
      * Constructor.
      * 
      * @param strategy  The traversal strategy; can be null, in which
      *                  case {@link CompositeStrategy#DEFAULT} is used.
      * @param sequences Sequences to add to this composite, if any; can be
      *                  null. Null entries are ignored.                  
      */
    public AbstractCompositeSequence(CompositeStrategy strategy, Sequence<? extends E>... sequences) {
        // Do not reset because we are not initialised completely...
        super(false);
        this.setCompositeStrategy(strategy);
        add(this.sequences = new ArrayList<Sequence<? extends E>>(), sequences);
    }
   
    /**
      * Copy constructor, which also copies associated sequences.
      * 
      * @param  sequence The composite sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null. 
      */
    public AbstractCompositeSequence(CompositeSequence<? extends E> sequence) {
        this.strategy = sequence.getCompositeStrategy();
        // Copy associated sequences...
        List<? extends Sequence<? extends E>> sequences = sequence.getSequences();
        this.sequences = new ArrayList<Sequence<? extends E>>(sequences.size());
        for (Sequence<? extends E> s : sequences) {
            this.sequences.add(s.copy());
        }
    }
    
    /**
      * 
      */
    public boolean addSequence(Sequence<? extends E> sequence) {
        if (sequence == null) {
            throw new NullPointerException("Sequence cannot be null");
        } else if (this.sequences.contains(sequence)) {
            return false;
        }
        // Returns true...
        return this.sequences.add(sequence);
    }

    /**
      * 
      */
    public boolean removeSequence(Sequence<?> sequence) {
        if (sequence == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        return this.sequences.remove(sequence);
    }
    
    /**
      * 
      */
    public List<Sequence<? extends E>> getSequences() {
        return new ArrayList<Sequence<? extends E>>(this.sequences);
    }

    /**
      * 
      */
    public Iterator<Sequence<? extends E>> iterator() {
        return this.getSequences().iterator();
    }
    
    /**
      * 
      */
    public <V extends Sequence<?>> List<V> getSequences(Class<V> clazz) {
        List<V> sequences = new ArrayList<V>(this.sequences.size());
        for (Sequence<? extends E> s : this.sequences) {
            if (clazz.isInstance(s)) {
                sequences.add(clazz.cast(s));
            }
        }
        return sequences;
    }

    /**
      * 
      */
    public List<Sequence<? extends E>> getSequences(CompositeStrategy strategy, boolean includeSelf, Boolean type) {
        List<Sequence<? extends E>> sequences = strategy.traverse(this, type, new ArrayList<Sequence<? extends E>>(), strategy);
        if (!includeSelf) {
            sequences.remove(this);
        }
        return sequences;        
    }

    /**
      * 
      */
    public <V extends Sequence<?>> List<V> getSequences(CompositeStrategy strategy, boolean includeSelf, Boolean type, Class<V> clazz) {
        List<Sequence<? extends E>> sequences = this.getSequences(strategy, includeSelf, type);
        List<V> list = new ArrayList<V>(sequences.size());
        for (int i = sequences.size() - 1; i >= 0; i--) {
            Sequence<?> s = sequences.get(i);
            if (clazz.isInstance(sequences.get(i))) {
                list.add(clazz.cast(s));
            }
        }
        return list;
    }
    
    /**
      * 
      */
    public CompositeStrategy getCompositeStrategy() {
        return this.strategy;
    }
    
    /**
      * 
      */
    public void setCompositeStrategy(CompositeStrategy policy) {
        this.strategy = (policy == null) ? CompositeStrategy.DEFAULT : policy;
    }

    /**
      * 
      */
    public int size() {
        return this.sequences.size();
    }

    /**
      * 
      */
    public boolean bounded() {
        for (Sequence<?> s : this.sequences) {
            if (!s.bounded()) {
                return false;
            }
        }
        return true;
    }

    /**
      * 
      */
    public boolean consistent() {
        for (Sequence<?> s : this.sequences) {
            if (!s.consistent()) {
                return false;
            }
        }
        return true;
    }

    /**
      * Returns false.
      * 
      * @return False.
      */
    public boolean unique() {
        return false;
    }
    
    /**
      * 
      */
    public void reset() {
        this.state = State.RESET;
        for (Sequence<?> s : this.sequences) {
            s.reset();
        }
    }
    
    /**
      * Calculates a "meaningful" state based on all associated
      * sequences: if all are reset, this sequence is reset. Otherwise
      * it functions normally. <p>
      * 
      * Sub-classes should invoke this method in their {@code next()}
      * implementation after the individual sequences have been advanced.
      */
    protected void calculateNextState() {
        if (this.bounded()) {
            for (Sequence<?> s : this.sequences) {
                if (s.state() != State.RESTART) {
                    this.state = State.NORMAL;
                    return;
                }
            }
            this.state = State.RESTART; // wow! :) 
        } else {
            this.state = State.NORMAL;
        }
    }
    
}
