/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.composite;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.observer.ObservableSequence;

/**
  * A <i>composite sequence</i> is a sequence that represents
  * an number of other contained sequences, possibly composite
  * as well. Sequence methods are propagated to each contained
  * sequence in either a depth or breath first manner, determined
  * via a changeable {@link CompositeStrategy strategy}. <p>
  * 
  * The value returned by the {@link Sequence#current() current()} and 
  * {@link Sequence#next() next()} methods is sub-class specific. <p>
  *  
  * Composite sequences are simple containers that do not maintain
  * parent references. Contained sequences do not know they are
  * contained unless this information is stored locally. Hence,
  * graph traversal is top-down only. <p>
  * 
  * Composite sequences are {@link #bounded() bounded} and
  * {@link #consistent() consistent} if each contained sequence is
  * bound and consistent, respectively, but composite sequences cannot
  * be guaranteed to be {@link #unique() unique} even
  * if all contained sequences are unique. Hence, composite sequences
  * are per definition not unique. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Any sequence delivering a sequence value of type {@code <? extends E>}
  * can be stored in a composite sequence of type {@code E}. This offers
  * greater flexibility, while ensuring that all sequences in a composite 
  * sequence will match the {@code E}, and can thus be "concatenated in a 
  * proper fashion", for example appending strings, adding different types 
  * of numbers, etc.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ObservableSequence
  * 
  * @param  <E> The type of values delivered by this composite sequence.
  */
@ParticipantUsage(value = {"Component", "Leaf"}, type = Sequence.class)
@Participant("Composite")
public interface CompositeSequence<E> extends Sequence<E>, Iterable<Sequence<? extends E>> {
    
    /**
      * Returns the {@link Sequence sequences} currently
      * associated with <b>this</b> composite sequence, in order. <p>
      * 
      * Modifying the returned collection will not affect 
      * this sequence.
      *  
      * @return The collection of associated sequences; never 
      *         null, but can be empty.
      */
    public List<Sequence<? extends E>> getSequences();
      
    /**
      * Returns the {@link Sequence sequences} currently
      * associated with <b>this</b> composite sequence, in order,
      * assignable to the type supplied as {@code clazz}. <p>
      * 
      * Invoking this method with the {@code Sequence} interface
      * corresponds to invoking {@link #getSequences()}. <p>
      * 
      * Modifying the returned collection will not affect 
      * this sequence.
      *  
      * @param  <V>   The type of applicable sequences.
      * @param  clazz The (super-)class of sequences to include; cannot be null.
      * @return The collection of associated sequences of the type
      *         supplied as {@code clazz}; never null, but can be 
      *         empty.
      * @throws NullPointerException If {@code clazz} is null.
      * 
      */
    public <V extends Sequence<?>> List<V> getSequences(Class<V> clazz);
    // Implementation note: has to be Sequence<?>, not Sequence<? extends E>,
    // because "clazz" can represent any type of sequence with any type of 
    // parameter...
    
    /**
      * Returns the {@link Sequence sequences} currently
      * reachable from this composite sequence, in the
      * order specified by {@code strategy}. <p>
      * 
      * This sequence is included in the returned list if 
      * {@code includeSelf} is true <b>and</b> it matches {@code type}; 
      * {@code type} determines the types of sequences to include in
      * the returned list: <ul><p>
      * 
      * <li> {@code TRUE}: only composite sequences ("nodes"). 
      * <li> {@code FALSE}: only non-composite sequences ("leafs").
      * <li> null: all sequences.
      * 
      * </ul><p>
      * 
      * If {@code strategy} is null, the {@link #getCompositeStrategy() default}
      * strategy for this sequence is used. <p>
      * 
      * Modifying the returned collection will not affect 
      * this sequence.
      *  
      * @param  strategy    The strategy dictating the traversal; can be null.
      * @param  includeSelf True to include this composite sequence in the
      *                     returned list (if it matches {@code type}),
      *                     false to exclude. 
      * @param  type        The type of sequences to include in the returned
      *                     list; can be null.
      * @return The list of reachable sequences; never null, 
      *         but can be empty.
      */
    public List<Sequence<? extends E>> getSequences(CompositeStrategy strategy, boolean includeSelf, Boolean type);
      
    /**
      * Returns the {@link Sequence sequences} currently
      * reachable from this composite sequence, in the order
      * specified by {@code strategy}. Each sequence included 
      * in the returned list is assignable to the type supplied 
      * as {@code clazz}. <p>
      * 
      * This sequence is included in the returned list if it
      * matches {@code clazz} and {@code type}; {@code type} determines 
      * the types of sequences to include in the returned list: <ul><p>
      * 
      * <li> {@code TRUE}: only composite sequences ("nodes"). 
      * <li> {@code FALSE}: only non-composite sequences ("leafs").
      * <li> null: all sequences.
      * 
      * </ul><p>
      * 
      * If {@code strategy} is null, the {@link #getCompositeStrategy() default}
      * strategy for this sequence is used. <p>
      * 
      * Modifying the returned collection will not affect 
      * this sequence.
      *  
      * @param  <V>         The type of applicable sequences.
      * @param  strategy    The strategy dictating the traversal; can be null.
      * @param  includeSelf True to include this composite sequence in the
      *                     returned list (if it matches {@code type} and
      *                     {@code clazz}), false to exclude. 
      * @param  type        The type of sequences to include in the returned
      *                     list; can be null.
      * @param  clazz       The (super-)class of sequences to include; cannot be null.
      * @return The collection of reachable sequences of the type
      *         supplied as {@code clazz}; never null, but can be 
      *         empty.
      * @throws NullPointerException If {@code clazz} is null.
      */
    public <V extends Sequence<?>> List<V> getSequences(CompositeStrategy strategy, boolean includeSelf, Boolean type, Class<V> clazz);
    // Implementation note: has to be Sequence<?>, not Sequence<? extends E>,
    // because "clazz" can represent any type of sequence with any type of 
    // parameter...
    
    /**
      * Associates the {@link Sequence sequence} supplied
      * as {@code sequence} to this composite sequence, if
      * not already. <p>
      * 
      * If already added, this method does nothing.  
      * 
      * @param  sequence The sequence to add; cannot be null.
      * @return True if {@code sequence} was added, false if not, 
      *         i.e. if already added.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IllegalArgumentException If {@code sequence} is not 
      *         a valid sequence to add to this composite sequence.
      * @see    #removeSequence(Sequence)         
      */
    public boolean addSequence(Sequence<? extends E> sequence);
    
    /**
      * Removes the {@link Sequence sequence} supplied
      * as {@code sequence} from this composite sequence, if
      * associated. <p>
      * 
      * If not associated, this method does nothing.  
      * 
      * @param  sequence The sequence to remove; cannot be null.
      * @return True if {@code sequence} was removed, false if not, 
      *         i.e. if not previously added.
      * @throws NullPointerException If {@code sequence} is null.
      * @see    #addSequence(Sequence)         
      */
    public boolean removeSequence(Sequence<?> sequence);

    /**
      * Returns the number of sequences associated to this
      * composite sequence.
      *  
      * @return The number of sequences.
      */
    public int size();
    
    /**
      * Return the current {@link CompositeStrategy strategy} used by
      * this composite sequence to traverse the associated sequence 
      * graph.
      *  
      * @return The strategy; never null.
      */
    public CompositeStrategy getCompositeStrategy();
    
    /**
      * Sets the current {@link CompositeStrategy strategy} used by
      * this composite sequence to traverse the associated sequence 
      * graph to {@code strategy}.
      *  
      * @param strategy The strategy; can be null, in which case the
      *                 default strategy is used.
      */
    public void setCompositeStrategy(CompositeStrategy strategy);
    
}
