/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.composite;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.model.Sequence;

/**
  * Strategies for traversing {@link CompositeSequence composite} sequence
  * graphs. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Composite strategies are implemented according to the 
  * {@linkplain dk.rode.thesis.strategy Strategy} pattern.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public enum CompositeStrategy {

    /**
      * Performs a depth first traversal returning all reachable
      * sequences. <p>
      * 
      * @see #DEPTH_FIRST
      */
    DEFAULT {
        protected <E,C extends List<Sequence<? extends E>>> C traverse(Sequence<? extends E> sequence, Boolean type, C sequences, CompositeStrategy strategy) {
            return DEPTH_FIRST.traverse(sequence, type, sequences, strategy);
        }
    },

    /**
      * Performs a depth first traversal. <p>
      */
    DEPTH_FIRST {
        protected <E,C extends List<Sequence<? extends E>>> C traverse(Sequence<? extends E> sequence, Boolean type, C sequences, CompositeStrategy strategy) {
            // Stop recursing if already processed...
            if (!sequences.contains(sequence)) {
                boolean composite = (sequence instanceof CompositeSequence);
                if (composite) {
                    CompositeSequence<? extends E> cs = (CompositeSequence<? extends E>)sequence;
                    // Recurse...
                    for (Sequence<? extends E> s : cs.getSequences()) {
                        strategy.traverse(s, type, sequences, strategy);
                    }
                }
                if ((type == null) || (composite == type)) {
                    sequences.add(sequence);
                }
            }
            return sequences;
        }
    },
    
    /**
      * Performs a breath first traversal. <p>
      */
    BREATH_FIRST {
        protected <E,C extends List<Sequence<? extends E>>> C traverse(Sequence<? extends E> sequence, Boolean type, C sequences, CompositeStrategy strategy) {
            // Stop recursing if already processed...
            if (!sequences.contains(sequence)) {
                boolean composite = (sequence instanceof CompositeSequence);
                if ((type == null) || (composite == type)) {
                    sequences.add(sequence);
                }
                if (composite) {
                    CompositeSequence<? extends E> cs = (CompositeSequence<? extends E>)sequence;
                    List<Sequence<? extends E>> composites = new ArrayList<Sequence<? extends E>>(cs.size());
                    for (Sequence<? extends E> s : cs.getSequences()) {
                        if (s instanceof CompositeSequence) {
                            composites.add(s);
                        } else if (!((Boolean.TRUE.equals(type)) || (sequences.contains(s)))) {
                            sequences.add(s);
                        }
                    }
                    // Recurse...
                    for (Sequence<? extends E> s : composites) {
                        strategy.traverse(s, type, sequences, strategy);
                    }
                }
            }
            return sequences;
        }
    },

    /**
      * When a {@link CompositeSequence} is reached, its 
      * {@link CompositeSequence#getCompositeStrategy() default} 
      * strategy is used to traverse onwards. 
      */
    COMBINED {
        protected <E,C extends List<Sequence<? extends E>>> C traverse(Sequence<? extends E> sequence, Boolean type, C sequences, CompositeStrategy strategy) {
            // Use policy from composite sequence...
            if (sequence instanceof CompositeSequence) {
                CompositeSequence<? extends E> cs = (CompositeSequence<? extends E>)sequence;
                // Use default...
                strategy = cs.getCompositeStrategy();
                // Cannot use combined because of infinite recursion...
                if ((strategy == null) || (strategy == COMBINED)) {
                    strategy = DEPTH_FIRST;
                }
                strategy.traverse(cs, type, sequences, COMBINED);
                
            // Non-composite - simply add if not already and if we 
            // search for all or non-composite types...
            } else if ((!Boolean.TRUE.equals(type)) && (!sequences.contains(sequence))) {
                sequences.add(sequence);
            }
            return sequences;
        }
    };
    
    /**
      * Returns an ordered list containing all {@link Sequence sequences}
      * reachable from {@code root}, including {@code root}. <p>
      * 
      * {@code type} determines the types of sequences to include in
      * the returned list: <ul><p>
      * 
      * <li> {@code TRUE}: only composite sequences ("nodes"). 
      * <li> {@code FALSE}: only non-composite sequences ("leafs").
      * <li> null: all sequences.
      * 
      * </ul>
      * 
      * @param  <E>  The type of values delivered by {@code sequence}.
      * @param  root The composite sequence to start from; cannot be null.
      * @param  type The type of traversal; can be null, i.e. return
      *              all reachable sequences.
      * @return An ordered list of reachable sequences; never null, but
      *         can be empty.
      * @throws NullPointerException If {@code root} is null.
      */
    public final <E> List<Sequence<? extends E>> traverse(Sequence<? extends E> root, Boolean type) {
        return this.traverse(root, type, new ArrayList<Sequence<? extends E>>(), this);
    }
    
    /**
      * Adds all {@link Sequence sequences} reachable from {@code sequence}
      * to {@code sequences} as dictated by this strategy, including
      * {@code sequence} if not already added. <p>
      * 
      * {@code type} determines the types of sequences to include in
      * the returned list: <ul><p>
      * 
      * <li> {@code TRUE}: only composite sequences ("nodes"). 
      * <li> {@code FALSE}: only non-composite sequences ("leafs").
      * <li> null: all sequences.
      * 
      * </ul>
      * 
      * @param  <E>       The type of sequence values.
      * @param  <C>       The actual type of list to add the sequences to.
      * @param  sequence  The current sequence; cannot be null.
      * @param  type      The type of traversal; can be null, i.e. return
      *                   all reachable sequences.
      * @param  sequences The list to add the reached sequences; never null.
      * @param  strategy  The strategy to apply for all attached sequences,
      *                   if any; never null.
      * @return The {@code sequences} list; cannot be null.                                   
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected abstract <E,C extends List<Sequence<? extends E>>> C traverse(Sequence<? extends E> sequence, Boolean type, C sequences, CompositeStrategy strategy);

}
