/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.decorator;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>appender decorator</i> returns the String value
  * from a decorated sequence prefixed and/or postfixed with
  * values supplied at construction time, respectively. Given 
  * the prefix and postfix values {@code xx} and {@code zz}, 
  * the value returned will thus be {@code xxYYzz}, where 
  * {@code YY} is the value returned by the decorated sequence. <p>
  * 
  * Only sequences returning strings can be decorated. <p>
  *  
  * The default stringable policy used is {@link DecoratedPolicy#HIDDEN}.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteDecorator")
public class AppenderDecorator extends SequenceDecorator<String> implements Sequence<String> {

    /**
      * Prefix to any value returned by the decorated sequence. <p>
      * 
      * Can be null, i.e. not used.
      */
    private final CharSequence prefix;
    /**
      * Postfix to any value returned by the decorated sequence. <p>
      * 
      * Can be null, i.e. not used.
      */
    private final CharSequence postfix;
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence to decorate; cannot be null.
      * @param  prefix   The prefix to use; null means none.
      * @param  postfix  The postfix to use; null means none.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public AppenderDecorator(@Participant("ConcreteComponent") Sequence<String> sequence, CharSequence prefix, CharSequence postfix) {
        super(sequence);
        this.prefix  = prefix;
        this.postfix = postfix;
    }

    /**
      * 
      */
    public AppenderDecorator copy() {
        return new AppenderDecorator(this.getSequence(true), this.prefix, this.postfix);
    }

    /**
      * 
      */
    public String current() {
        return enclose(this.prefix, super.current(), this.postfix);
    }

    /**
      * 
      */
    public String next() {
        return enclose(this.prefix, super.next(), this.postfix);
    }

    /**
      * Returns a string having one of the following 
      * formats depending on whether {@code prefix} and
      * {@code postfix} are null, respectively: <ul><p>
      * 
      * <li> {@code value}: {@code prefix} and {@code postfix} are null.
      * 
      * <li> {@code prefix} + {@code value}: {@code postfix} is null.
      *    
      * <li> {@code value} + {@code postfix}: {@code prefix} is null.
      * 
      * <li> {@code prefix} + {@code value} + {@code postfix}: neither is null.
      * 
      * </ul>
      * 
      * @param  prefix The prefix, if any; can be null.
      * @param  value  The value; cannot be null.
      * @param  postfix The postfix, if any; can be null.
      * @return The concatenated string; never null.
      */
    private final static String enclose(CharSequence prefix, CharSequence value, CharSequence postfix) {
        StringBuilder sb = new StringBuilder();
        if (prefix != null) {
            sb.append(prefix);
        }
        sb.append(value);
        if (postfix != null) {
            sb.append(postfix);
        }
        return sb.toString();
    }
    
}
