/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.decorator;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * A <i>sequence decorator</i> is a decorator that forwards all
  * request to the decorated {@link Sequence} <b>without</b> performing
  * additional operations except for {@code toString()}. <p>
  * 
  * Sub-classes can override {@link #current()} and {@link #next()} to
  * add functionality, but <b>must</b> implement {@link #copy()}. <p>
  * 
  * A sequence decorator exhibit the same {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link Sequence#unique() unique}
  * properties as the decorated sequence. <p> 
  * 
  * The default stringable policy used is {@link DecoratedPolicy#VISIBLE}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The decorated sequence must have the exact same type parameter, {@code E},
  * and not {@code <? extends E>} is not allowed. To facilitate different
  * type parameters, the {@linkplain dk.rode.thesis.adapter Adapter} pattern
  * can be used.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence.
  */
@ParticipantUsage(value = {"Component", "ConcreteComponent"}, type = Sequence.class)
@Participant("Decorator")
public abstract class SequenceDecorator<E> implements Sequence<E> {
    
    /**
      * The decorated sequence. <p>
      * 
      * Never null.
      */
    final Sequence<E> sequence;
    /**
      * The default sequence policy used for formatting. <p>
      *
      * Can be null, in which case the default policy of decorated
      * sequence is used. 
      */
    private final StringablePolicy<Sequence<?>> policy;
    
    /**
      * Constructor, which creates this decorator to decorate
      * the {@link Sequence} supplied as {@code sequence}. <p>
      * 
      * The default stringable policy used is {@link DecoratedPolicy#VISIBLE}.
      * 
      * @param  sequence The sequence; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected SequenceDecorator(@Participant("Component") Sequence<E> sequence) {
        this(sequence, DecoratedPolicy.VISIBLE);
    }

    /**
      * Constructor, which creates this decorator to decorate
      * the {@link Sequence} supplied as {@code sequence} and
      * use the default stringable policy supplied as {@code policy}.
      * 
      * @param  sequence The sequence; cannot be null.
      * @param  policy   The default policy; can be null. 
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected SequenceDecorator(@Participant("Component") Sequence<E> sequence, StringablePolicy<Sequence<?>> policy) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        // Null = default policy from "sequence"...
        this.policy = policy;
    }
    
    /**
      * Returns a <b>copy</b> of the decorated sequence
      * if {@code copy} is true, otherwise the decorated
      * sequence it self. <p>
      * 
      * @param  copy True to return a copy of the decorated sequence,
      *              false to return a reference to the sequence.
      * @return The decorated sequence, as a copy if {@code copy}
      *         is true; never null.
      */
    protected Sequence<E> getSequence(boolean copy) {
        return copy ? this.sequence.copy() : this.sequence;  
    }
       
    /**
      * 
      */
    public boolean bounded() {
        return this.sequence.bounded();
    }

    /**
      *
      */
    public boolean consistent() {
        return this.sequence.consistent();
    }

    /**
      *
      */
    public boolean unique() {
        return this.sequence.unique();
    }
    
    /**
      * 
      */
    public E current() {
        return this.sequence.current();
    }

    /**
      * 
      */
    public E next() {
        return this.sequence.next();
    }

    /**
      * 
      */
    public void reset() {
        this.sequence.reset();
    }

    /**
      * 
      */
    public State state() {
        return this.sequence.state();
    }
    
    /**
      * 
      */
    public int hashCode() {
        return this.sequence.hashCode();
    }

    /**
      * 
      */
    public boolean equals(Object object) {
        if (!(object instanceof SequenceDecorator)) {
            return false;
        }
        SequenceDecorator<?> sd = (SequenceDecorator<?>)object;
        return this.sequence.equals(sd.sequence); 
    }
    
    /**
      * 
      */
    public String toString() {
        return this.toString(null).toString();
    }
    
    /**
      * 
      */
    public StringablePolicy<? super Sequence<E>> getStringablePolicy(StringablePolicy<? super Sequence<E>> policy) {
        if (policy == null) {
            if (this.policy == null) {
                return this.sequence.getStringablePolicy(null);
            }
            return this.policy;
        }
        return policy;
    }
    
    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }
    
}
