/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.meta.reflect.TypeLiteral;

/**
  * A <i>generic factory</i> that will create products of type {@code T} while
  * only accepting constructor arguments of type {@code P} (for all arguments). <p>
  * 
  * Usage: to create instances of {@code Foo<String>} using the constructor
  * accepting two {@code String} arguments: 
  * <pre>
  *   // Create factory...
  *   TypedFactory&lt;Foo&lt;String>,String> factory1 = 
  *     new {@link #TypedFactory(Class,int) TypedFactory}(String.class, 2){}; // anonymous class
  *     
  *   // Create products...  
  *   Foo&lt;String> foo    = factory1.{@link #createInstance(Object...) createInstance}("Bar", "BarBar");
  *   Foo&lt;String> foobar = factory1.{@link #createInstance(Object...) createInstance}("FooBar", "GooBar");
  * </pre>
  * 
  * To create instances of {@code Goo<String>} using the constructor
  * accepting a the generic {@code FooBar<V>} type, e.g. {@code FooBar<Integer>}: 
  * <pre>   
  *   // Argument(s) type and values...
  *   TypeLiteral&lt;FooBar&lt;Integer>> type = new {@link TypeLiteral}&lt;FooBar&lt;Integer>>(){}; // anonymous class
  *  
  *   FooBar&lt;Integer> fb1 = new FooBar&lt;Integer>(..);
  *   FooBar&lt;Integer> fb2 = new FooBar&lt;Integer>(..);
  *   
  *   // Factory...
  *   TypedFactory&lt;Goo&lt;String>,FooBar&lt;Integer>> factory2 = 
  *     new {@link #TypedFactory(TypeLiteral) TypedFactory}(type){}; // e.g. TypedFactory(type, 1)
  *   
  *     
  *   // Create products...  
  *   Goo&lt;String> goo1 = factory2.{@link #createInstance(Object...) createInstance}(fb1); // FooBar&lt;Integer>
  *   Goo&lt;String> goo2 = factory2.{@link #createInstance(Object...) createInstance}(fb2); // FooBar&lt;Integer>
  *   
  * </pre>
  *
  * The constructor to create new instances of {@code T} is
  * fetched ignoring primitive types, e.g. a formal parameter {@code P} of type
  * {@code Integer} will match both {@code Integer} and {@code int}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This factory is classified as a Factory Method, not an 
  * {@linkplain dk.rode.thesis.abstractfactory Abstract Factory},
  * because its traits most closely resemble those described for Factory
  * Method by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.107-116], but it is a judgement call. 
  * We see it as a variant of the <i>Creator</i> class [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.113], which 
  * uses templates to determine the actual type of product to create. We still 
  * use it in the Abstract Factory implementation as well, though. These are
  * the reasons why it is named {@code Factory}, and not {@code Creator}. <p>
  * 
  * While the creation process can enforce proper type on supplied factory
  * method arguments, the number of arguments supplied may still be illegal.
  * This can be handled by making explicit factory methods, i.e. specific
  * factories, or perhaps by ignoring redundant arguments.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    InstantiableTypeLiteral
  * 
  * @param  <T> The product type created by this factory.
  * @param  <P> The single type of parameters used for the constructor.
  */
@Participant("Creator")
public abstract class TypedFactory<T,P> extends Factory<T> {

    /**
      * Constructor, which creates this factory to create instances
      * of the type represented by the type parameter {@code T} using
      * the declared constructor accepting a single parameter of type
      * {@code parameterType}. <p>
      * 
      * The constructor is made accessible if not already. New instances
      * of {@code T} must be created with the {@link #createInstance(Object...)}
      * factory method.
      *        
      * @param  parameterType The formal parameter type for the 
      *                       declared constructor to use; cannot be null.
      * @throws NullPointerException If {@code parameterType} is null.                       
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, or if the raw raw represents 
      *         a non instantiable class.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected TypedFactory(Class<P> parameterType) throws NoSuchMethodException {
        this(parameterType, 1);
    }

    /**
      * Constructor, which creates this factory to create instances
      * of the type represented by the type parameter {@code T} using
      * the declared constructor accepting the number of parameters
      * supplied as {@code parameters} of the type {@code parameterType}. <p>
      * 
      * The constructor is made accessible if not already. New instances
      * of {@code T} must be created with the {@link #createInstance(Object...)}
      * factory method.
      *        
      * @param  parameterType The formal parameter types for the 
      *                       declared constructor to use; cannot be null.
      * @param  parameters    The number of constructor parameters; cannot be negative.                       
      * @throws NullPointerException If {@code parameterType} is null.                       
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, if the raw raw represents 
      *         a non instantiable class, or if {@code parameters} is negative.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected TypedFactory(Class<P> parameterType, int parameters) throws NoSuchMethodException {
        super(toClassArray(parameterType, parameters));
    }
    
    /**
      * Constructor, which creates this factory to create instances
      * of the type represented by the type parameter {@code T} using
      * the declared constructor accepting a single parameter of type
      * {@code parameterType}. <p>
      * 
      * The constructor is made accessible if not already. New instances
      * of {@code T} must be created with the {@link #createInstance(Object...)}
      * factory method.
      *        
      * @param  parameterType The formal parameter type for the 
      *                       declared constructor to use; cannot be null.
      * @throws NullPointerException If {@code parameterType} is null.                       
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, if the raw raw represents 
      *         a non instantiable class, or if {@code P} has no raw type.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected TypedFactory(TypeLiteral<P> parameterType) throws NoSuchMethodException {
        this(parameterType, 1);
    }

    /**
      * Constructor, which creates this factory to create instances
      * of the type represented by the type parameter {@code T} using
      * the declared constructor accepting the number of parameters
      * supplied as {@code parameters} of the type {@code parameterType}. <p>
      * 
      * The constructor is made accessible if not already. New instances
      * of {@code T} must be created with the {@link #createInstance(Object...)}
      * factory method.
      *        
      * @param  parameterType The formal parameter types for the 
      *                       declared constructor to use; cannot be null.
      * @param  parameters    The number of constructor parameters; cannot be negative.                       
      * @throws NullPointerException If {@code parameterType} is null.                       
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, if the raw raw represents 
      *         a non instantiable class, if {@code P} has no raw type, or if 
      *         {@code parameters} is negative.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected TypedFactory(TypeLiteral<P> parameterType, int parameters) throws NoSuchMethodException {
        super(toClassArray(parameterType, parameters));
    }
    
    /**
      * Creates an array of length {@code length} storing the raw type of
      * {@code typeLiteral} at each index. 
      * 
      * @param  typeLiteral The type literal representing the raw type; cannot be null.
      * @param  length      The length of the array; cannot be negative. 
      * @return An array of length {@code length} containing the corresponding raw type; 
      *         never null.
      * @throws NullPointerException If {@code typeLiteral} is null.
      * @throws IllegalArgumentException If {@code typeLiteral} do not represent
      *         a raw type, or if {@code length} is negative.                
      */
    private static Class<?>[] toClassArray(TypeLiteral<?> typeLiteral, int length) {
        Class<?> type = typeLiteral.getRawType();
        if (type == null) {
            throw new IllegalArgumentException("Parameter type literal does not represent a raw type: " + typeLiteral);
        }
        return toClassArray(type, length);
    }
    
    /**
      * Creates an array of length {@code length} storing the type supplied
      * as {@code type} at each index. 
      * 
      * @param  type   The type; cannot be null. 
      * @param  length The length of the array; cannot be negative. 
      * @return An array of length {@code length} containing {@code type}; 
      *         never null.
      * @throws NullPointerException If {@code type} is null.
      * @throws IllegalArgumentException If {@code length} is negative.
      */
    private static Class<?>[] toClassArray(Class<?> type, int length) {
        if (type == null) {
            throw new NullPointerException("Parameter type cannot be null");
        } else if (length < 0) {
            throw new IllegalArgumentException("Parameter length cannot be negative: " + length);
        }
        Class<?>[] types = new Class[length];
        for (int i = 0; i < length; i++) {
            types[i] = type;
        }
        return types;
    }
    
    /**
      * Returns a new instance of the type represented by
      * this factory using the arguments supplied as {@code arguments}. <p> 
      *  
      * @param  <S> The actual type of the arguments. 
      * @param  arguments The arguments to supply to the constructor. 
      * @return A new instance of the type represented by this factory; never null.
      * @throws Exception If the construction fails.
      * @see    #newInstance(Object...) 
      */
    public <S extends P> T createInstance(S... arguments) throws Exception {
        return super.newInstance(arguments);
    }
        
}
