/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import java.util.Locale;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.meta.util.Arguments.Argument;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Flyweight</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Client")
public class Main implements Testable {

    /**
      * Executes the <b>Flyweight</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * {@inheritDoc} <p>
      * 
      * Arguments to control log-level: <ul><p>
      * 
      * <li> {@code -log}: a boolean value (as a string) indicating if 
      *      class logging is enabled for relevant classes. <p> 
      * 
      * <li> {@code -log.verbose}: a boolean value (as a string) indicating
      *      if class logging is enabled for relevant classes that generates
      *      verbose output. <p>  
      * 
      * </ul>
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Flyweight tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        //
        // Create sentence(s)
        //
        
        // Create factory to use...
        CharacterFactory cf = new CharacterFactory();

        // Supplied arguments?
        if (arguments.size() > 0) {
            out.heading("Creating flyweights: sentences (unshared), words, and characters from supplied arguments");
            for (Argument argument : arguments) {
                
                if (argument.value != null) {
                    parse(out, cf, argument.value); 
                }
            }
        }
        
        // Parse standard arguments...
        out.heading("Creating flyweights: sentences (unshared), words, and characters");
        Sentence sentence = parse(out, cf, "Number 1: this is, indeed, a very SMALL test! :)  ");
        parse(out, cf, "  Second test   is\talso very (very)  small  , but illustrative!");
        parse(out, cf, " Three (3) is\na charm!!!  ");
        parse(out, cf, " 5 4 3 2 1 !");
        parse(out, cf, " \n  \t   "); 
        
        //
        // Test copying
        //
        
        // Copy: Sentence->String->Sentence (should not create new objects!)
        out.println();
        out.println("Copying: ", (sentence = sentence.copy()));
        out.println();
        out.println("Re-parsing: ");
        parse(out, cf, sentence.toString());
        
        //
        // Usage of flyweight methods
        //
        
        out.heading("Use flyweight method explicitly: CharSequence.charAt(0)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.charAt(0));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.charAt(0));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.charAt(0));
            }
        }

        out.heading("Use flyweight method explicitly: CharSequence.length()");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.length());
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.length());
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.length());
            }
        }

        out.heading("Use flyweight method explicitly: CharSequence.subSequence(0,length/2)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.subSequence(0, sentence.length() / 2));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.subSequence(0, w.length() / 2));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.subSequence(0, c.length() / 2));
            }
        }
        
        out.heading("Use flyweight method explicitly: Stringable.toString(null)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.toString(null));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.toString(null));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.toString(null));
            }
        }

        out.heading("Use flyweight method explicitly: Stringable.toString(VALUE_UPPERCASE)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.toString(SequencePolicy.VALUE_UPPERCASE));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.toString(SequencePolicy.VALUE_UPPERCASE));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.toString(CharacterPolicy.VALUE_UPPERCASE));
            }
        }
        
        out.heading("Use flyweight method explicitly: Comparable.compareTo(this)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.compareTo(sentence));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.compareTo(w));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.compareTo(c));
            }
        }

        Locale locale = Locale.getDefault();
        out.heading("Use flyweight method explicitly: Textual.lowerCaseFirst(Locale)")
           .println("Sentence:            ", sentence)
           .println("Default locale used: ", locale)
           .println();
        
        out.println("Sentence -> ", sentence.lowerCaseFirst(locale));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.lowerCaseFirst(locale));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.lowerCaseFirst(locale));
            }
        }

        out.heading("Use flyweight method explicitly: Textual.upperCaseFirst(Locale)");
        out.println("Sentence: ", sentence);

        out.println();
        out.println("Sentence -> ", sentence.upperCaseFirst(locale));
        for (Word w : sentence) {
            out.println("Word: ", w, " -> ", w.upperCaseFirst(locale));
            for (Character c : w) {
                out.println("Character: ", c, " -> ", c.upperCaseFirst(locale));
            }
        }

        CharSequence[] tokens = {"iS", "s"};
        out.heading("Use flyweight method explicitly: Textual.indexOfIgnoreCase(CharSequence, int, Locale)")
           .println("Sentence:            ", sentence)
           .println("Default locale used: ", locale);

        for (CharSequence token : tokens) {
            out.println().println("Search for: ", token).println();
            out.print("Sentence: ", sentence, " -> ");
            indexOfIgnoreCase(sentence, token, locale, out);
            for (Word w : sentence) {
                out.print("Word: ", w, " -> ");
                indexOfIgnoreCase(w, token, locale, out);
                for (Character c : w) {
                    out.print("Character: ", c, " -> ");
                    indexOfIgnoreCase(c, token, locale, out);
                }
            }
        }
        
        out.heading("All created flyweight objects");

        out.println();
        out.println("Words (", cf.countWords(), "):");
        for (Word w : cf.getWords()) {
            out.println(w.toString(SequencePolicy.VERBOSE), " -> ", w);
        }
        
        out.println();
        out.println("Characters (", cf.count(Character.class), "):");
        for (Character c : cf.getCharacters()) {
            out.println(c.toString(CharacterPolicy.IDENTITY), " -> ", c);
        }
        
        out.heading("Flyweight tests succeeded!");
        
        return true;
    }

    /**
      * Finds the indexes of {@code token} in {@code textual} and logs
      * them to {@code log}.
      *  
      * @param textual The textual object; never null.
      * @param token   The token to search for; never null.
      * @param locale  The local to use; never null.
      * @param out     The log to use.
      */
    private final static void indexOfIgnoreCase(Textual<?> textual, CharSequence token, Locale locale, Log out) {
        int index = 0;
        boolean any = false;
        do {
            index = textual.indexOfIgnoreCase(token, index, locale);
            if (any) {
                out.print(", ");
            }
            out.print(index);
            any = true;
        } while (index++ != -1);
        out.println();
    }
    
    /**
      * Parses the string supplied as {@code string} into a 
      * sentence, which is printed and then returned. 
      * 
      * @param  out     The log to use; never null.
      * @param  factory The character factory to use; never null.
      * @param  string  The string to parse; never null.
      * @return The constructed {@code Sentence} object; never null.
      */
    private final static Sentence parse(Log out, CharacterFactory factory, String string) {
        // Print info...
        out.println();
        out.println("Parsing: \"", string, "\"");

        // Parse...
        Sentence sentence = factory.getSentence(string);
        
        // Factory status...
        out.println("Factory: " + factory);
        out.println("- Words:      ", factory.countWords());
        out.println("- Letters:    ", factory.count(Letter.class));
        out.println("- Symbols:    ", factory.count(Symbol.class));
        out.println("- Whitespace: ", factory.count(Whitespace.class));
        out.println("- Total:      ", factory.count(Character.class));
        out.println();

        // Print it...
        for (Word w : sentence) {
            for (Character c : w) {
                out.print(c);
                out.print("|");
            }
            out.println();
        }
        return sentence;
    }
}
