/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;

import dk.rode.thesis.iterator.SequenceIterator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Strings;

/**
  * A <i>sentence</i> represents an ordered sequence of  
  * {@link Word words} making up that sentence. <p>
  *  
  * A sentence is {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, but not 
  * {@link #unique() unique} if the same word is contained more
  * than once, including {@link Word#isWhiteSpace() white space}!
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("UnsharedConcreteFlyweight")
public class Sentence 
     extends ArraySequence<Word> 
  implements Sequence<Word>, Iterable<Word>, Textual<Sentence> {

    /**
      * The hash code of this sentence. 
      */
    private final int hashCode;
    /**
      * True if this sentence is {@link #unique() unique}, false
      * if not. <p>
      * 
      * If the same {@link Word word} is contained
      * more than once, this sentence is not unique. 
      */
    private final boolean unique;
   
    /**
      * Constructor, which creates this sentence to be comprised
      * of the words in {@code words}.
      * 
      * @param  words The words; cannot be null or empty.
      * @throws NullPointerException If {@code words} is null.
      * @throws IllegalArgumentException If {@code words} is empty.
      */
    Sentence(Word[] words) {
        super(words);
        this.hashCode = this.current().hashCode();
        boolean unique = true;
        Set<Word> set = new HashSet<Word>();
        for (Word w : words) {
            if (set.contains(w)) {
                unique = false;
            }
            set.add(w);
        }
        this.unique = unique;
        
    }

    /**
      * Copy constructor.
      * 
      * @param  sentence The sentence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    Sentence(Sentence sentence) {
        super(sentence);
        this.hashCode = sentence.hashCode;
        this.unique = sentence.unique;
    }

    /**
      * 
      */
    public Sentence copy() {
        // Fine to copy into new Sentence object, but we do not
        // copy words or characters, because they are flyweights.
        return new Sentence(this);
    }

    /**
      * Returns true if any given {@link Word word} in this
      * sentence is only contained once, including white
      * space, false otherwise.
      * 
      * @return True if unique, false if not.
      */
    public boolean unique() {
        return this.unique;
    }
    
    /**
      * 
      */
    public Iterator<Word> iterator() {
        return SequenceIterator.iterator(this);
    }

    /**
      * 
      */
    public String lowerCaseFirst(Locale locale) {
        return this.elements[0].lowerCaseFirst(locale);
    }
    
    /**
      * 
      */
    public String upperCaseFirst(Locale locale) {
        return this.elements[0].upperCaseFirst(locale);
    }
    
    /**
      * 
      */
    public int indexOfIgnoreCase(CharSequence sequence, int index, Locale locale) {
        // Can use "this" because we implement CharSequence!
        return Strings.indexOfIgnoreCase(this, sequence, index, locale);
    }
    
    /**
      * 
      */
    public char charAt(int index) {
        return this.toString().charAt(index);
    }

    /**
      * 
      */
    public int length() {
        return this.toString().length();
    }

    /**
      * 
      */
    public CharSequence subSequence(int start, int end) {
        return this.toString().subSequence(start, end);
    }

    /**
      * 
      */
    public final int hashCode() {
        return this.hashCode; // hash code of first word in this sentence
    }

    /**
      * 
      */
    public final boolean equals(Object object) {
        if (!(object instanceof Sentence)) {
            return false;
        } else if (this == object) {
            return true;
        }
        return (this.compareTo((Sentence)object) == 0);
    }
    
    /**
      * 
      */
    public int compareTo(Sentence sentence) {
        // Invariant: words are bounded!

        // Simulates complex behaviour! Could just have
        // used this.toString().compareTo(sentence.toString())
        // or traversed the alphabet directly!
        
        Iterator<Word> i = this.iterator(),
                       j = sentence.iterator();

        boolean iNext = true, jNext = true;
        
        // Iterate through all words...
        while ((iNext = i.hasNext()) & (jNext = j.hasNext())) { // No short circuit!
            int n = i.next().compareTo(j.next());
            // Zero = equals...
            if (n != 0) {
                return n;
            }
        }

        // this sentence contains fewer words... 
        if (jNext) {
            return 1;
            
        // "sentence" contains fewer words...
        } else if (iNext) {
            return -1;
            
        // Same number of words...
        } else {
            return 0;
        }
    }
    
    /**
      * Returns this sentence as a string.
      * 
      * @return This sentence as a string; never null.
      */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (Word word : this) {
            sb.append(word);
        }
        return sb.toString();
    }
    
}
