/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import java.util.Locale;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>textual</i> object represents a flyweight object that is
  * a char sequence and comparable to its own type. <p>
  * 
  * The <i>value</i> of a textual object is its {@code toString()}
  * representation. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @param  <T> The type of this textual object.
  */
@Participant("Flyweight") 
public interface Textual<T> extends CharSequence, Comparable<T> {

    /**
      * Return the first character of this textual representation 
      * lower cased. <p>
      * 
      * Note, that in some locales, shifting case of a given
      * character may cause it to expand into more than one 
      * character! Hence, the return type is a string.
      * 
      * @param  locale The locale to use; can be null, in which
      *                case the default locale is used.
      * @return The lower cased character; never null.
      */
    public String lowerCaseFirst(Locale locale);
    
    /**
      * Return the first character of this textual representation 
      * upper cased. <p>
      * 
      * Note, that in some locales, shifting case of a given
      * character may cause it to expand into more than one 
      * character! Hence, the return type is a string.
      * 
      * @param  locale The locale to use; can be null, in which
      *                case the default locale is used.
      * @return The upper cased character; never null.
      */
    public String upperCaseFirst(Locale locale);
    
    /**
      * Finds the first index where {@code sequence} is present in this
      * textual representation from the index supplied, regardless of case, 
      * or -1 if {@code sequence} cannot be found. <p>
      * 
      * {@code index} is allowed to be larger than the length of
      * this textual representation, in which case this method returns -1. 
      *  
      * The supplied locale is used to perform the case conversion.
      * 
      * @param  sequence The char sequence to find; cannot be null.
      * @param  index    The index to start at; cannot be negative.
      * @param  locale   The locale to use; can be null, in which case
      *                  the default locale is used.
      * @return The first index of {@code sequence} from {@code index}, 
      *         regardless of case, or -1 if not found.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IndexOutOfBoundsException If {@code index} is negative.
      */
    public int indexOfIgnoreCase(CharSequence sequence, int index, Locale locale);
    
}
