/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An <i>assignment expression</i> assigns the result of the
  * evaluation of an expression to a {@link VariableExpression 
  * variable}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The assignment expression cannot inherit {@link BinaryExpression}
  * because of the {@code Expression<? extends E>} bound on
  * the first operand for any sub-class of it. Hence, it cannot 
  * be guaranteed to store a variable of type {@code E}, which 
  * is required when assigning the value to the {@link Context context} 
  * on evaluation.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
public class AssignmentExpression<E>
     extends AbstractExpression<E> 
  implements NonTerminalExpression<E> {
    
    /**
      * The variable expression to be assigned the value
      * of {@link #expression}. <p>
      * 
      * Never null.
      */
    private final VariableExpression<E> variable;
    /**
      * The second operand. <p>
      * 
      * Never null.
      */
    private final Expression<? extends E> expression;
    
    /**
      * Constructor.
      * 
      * @param  variable   The variable to be assigned the result of the
      *                    evaluation of {@code expression}; cannot be null,
      *                    or be a {@link ConstantExpression constant}.      
      * @param  expression The expression to deliver the variable value when
      *                    evaluated; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code variable} is a constant.                   
      */
    public AssignmentExpression(VariableExpression<E> variable, Expression<? extends E> expression) {
        if ((this.variable = variable) == null) {
            throw new NullPointerException("First expression cannot be null");
        } else if ((this.expression = expression) == null) {
            throw new NullPointerException("Second expression cannot be null");
        } else if (this.variable instanceof ConstantExpression) {
            throw new IllegalArgumentException("Variable cannot be a constant");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public AssignmentExpression(AssignmentExpression<E> expression) {
        this.variable = expression.variable.copy();
        this.expression = expression.expression.copy();
    }

    /**
      *
      */
    public AssignmentExpression<E> copy() {
        return new AssignmentExpression<E>(this);
    }

    /**
      *
      */
    public List<Expression<?>> operands() {
        return add(new ArrayList<Expression<?>>(2), this.variable, this.expression);
    }
    
    /**
      * @return {@code assignment}
      */
    public final String name() {
        return "assignment";
    }

    /**
      * 
      */
    public E evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        E value = this.expression.evaluate(context);
        context.assign(this.variable, value);
        return value;
    }
    
    /**
      * Returns {@code x = y}, where {@code x} is the symbolic 
      * representation of the first expression operand and {@code y} 
      * of the the second.  
      */
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        // First...
        SymbolIdiom.asSymbol(context, this, this.variable, sb);
        // Separator...
        sb.append(" = ");
        // Second...
        return SymbolIdiom.asSymbol(context, this, this.expression, sb).toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; variable: ").append(this.variable)
                 .append("; expression: ").append(this.expression)
                 .append(']').toString();
    }
    
}
