/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.newIdentityHashSet;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.util.KeyValue;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>context</i> represents an evaluation context where 
  * {@link VariableExpression variables} (and 
  * {@link ConstantExpression constants}) are stored during
  * {@link Interpreter interpretation}. <p>
  * 
  * A context aids in the generation of symbolic representations
  * of expressions via the {@link #touch(Expression)} and
  * {@link #reset()} methods. <p>
  * 
  * A context is <b>not</b> thread-safe! <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The context could easily be expanded to store any kind of
  * {@link TypedExpression}, but the problem is what the
  * associated value should be. A value (constant) or another
  * expression, and what would be the point (replacement of
  * expressions, perhaps)?
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Interpreter
  */
@Participant("Context")
public class Context implements Iterable<VariableExpression<?>> {

    /**
      * The map storing the values. <p>
      * 
      * Format: {@code (key, value) = (String, KeyValue<VariableExpression<X>,X>)}. <p>
      * 
      * Keys are expression names. <p>
      * 
      * If a variable/constant has only been {@link #register(VariableExpression) registered}, the 
      * associated value is null, and no value can be retrieved. <p>
      * 
      * Never null.
      */
    private final Map<String,KeyValue<VariableExpression<?>,Object>> values;
    /**
      * A set storing currently touched expressions based
      * on their <b>identity</b>. <p>
      * 
      * Never null, but can be empty.
      * 
      * @see #reset()
      * @see #touch(Expression)
      * @see Expression#asSymbol(Context)
      */
    private final Set<Expression<?>> touched;
    
    /**
      * No-arg constructor.
      */
    public Context() {
        this.values = new HashMap<String,KeyValue<VariableExpression<?>,Object>>();
        this.touched = newIdentityHashSet();
    }

    /**
      * Copy constructor. <p>
      * 
      * The set of touched expressions are not copied.
      * 
      * @param  context The context to copy; cannot be null.
      * @throws NullPointerException if {@code context} is null.
      */
    public Context(Context context) {
        this();
        // Copy values... Constants will not be copied, but used as is, but that is ok...
        for (Map.Entry<String,KeyValue<VariableExpression<?>,Object>> entry : context.values.entrySet()) {
            KeyValue<VariableExpression<?>,?> keyValue = entry.getValue();
            KeyValue<VariableExpression<?>,Object> newKeyValue = 
                new KeyValue<VariableExpression<?>,Object>(keyValue.getKey().copy(), keyValue.getValue());
            this.values.put(entry.getKey(), newKeyValue);
        }
    }
    
    /**
      * Clears all stored variables or {@link #touch(Expression) touched}
      * expressions.
      *  
      * @return This context; never null.
      */
    public Context clear() {
        this.values.clear();
        this.touched.clear();
        return this;
    }
    
    /**
      * Resets the internal set of touched expressions. <p>
      * 
      * The actual calling context should <b>always</b> call this method 
      * before generating {@link Expression#asSymbol(Context) symbolic} 
      * representations because cyclic expression references may occur.
      * 
      * @return The context; never null.
      * @see    #touch(Expression)
      */
    public Context reset() {
        this.touched.clear();
        return this;
    }
    
    /**
      * Touches the expression supplied as {@code expression}
      * based on its <b>identity</b> and return false if it was 
      * the first time {@code expression} was touched since the 
      * last call to {@link #reset()}, true otherwise. 
      * 
      * @param  expression The expression to touch; cannot be null.
      * @return False if this was the first time {@code expression}
      *         was touched, true if touched before.
      * @see    Expression#asSymbol(Context)          
      */
    public boolean touch(Expression<?> expression) {
        return !this.touched.add(expression);
    }

    /**
      * Returns an iterator to iterate over all the contained 
      * {@link VariableExpression variables} and 
      * {@link ConstantExpression constants} stored in
      * this context, in no particular order. <p>
      * 
      * Modifying the returned iterator will not affect this
      * context.
      * 
      * @return The iterator to iterate over the contained expressions; 
      *         never null, but can be empty.
      * @see    #getVariables()
      */
    public Iterator<VariableExpression<?>> iterator() {
        return this.getVariables().iterator();
    }
    
    /**
      * Returns a set containing all the contained 
      * {@link VariableExpression variables} and 
      * {@link ConstantExpression constants} stored in
      * this context, in no particular order. <p>
      * 
      * Modifying the returned set will not affect this
      * context.
      * 
      * @return A set containing the expressions; never null,
      *         but can be empty.
      * @see    #iterator()        
      */
    public Set<VariableExpression<?>> getVariables() {
        Set<VariableExpression<?>> expressions = new HashSet<VariableExpression<?>>();
        for (KeyValue<VariableExpression<?>,?> keyValue : this.values.values()) {
            expressions.add(keyValue.getKey());
        }
        return expressions;
    }
    
    /**
      * Returns a set containing all the contained 
      * {@link VariableExpression variables} and 
      * {@link ConstantExpression constants} stored in
      * this context, in no particular order, delivering 
      * the type of values supplied as {@code type}. <p>
      * 
      * Variables that are only {@link #register(VariableExpression)
      * registered} are also included in the returned list. <p>
      * 
      * Modifying the returned set will not affect this
      * context.
      *
      * @param  <E>  The type of values returned by the expressions.
      * @param  type The type of values to be returned by the 
      *              expressions; cannot be null.
      * @return A set containing the expressions; never null,
      *         but can be empty.
      * @throws NullPointerException If {@code type} is null.         
      */
    public <E> Set<VariableExpression<E>> getVariables(Class<E> type) {
        HashSet<VariableExpression<E>> variables = new HashSet<VariableExpression<E>>();
        for (KeyValue<VariableExpression<?>,?> keyValue : this.values.values()) {
            VariableExpression<?> expression = keyValue.getKey();
            if (type.equals(expression.type())) {
                // Known to succeed because of the above check and the
                // rules enforced when adding expressions to this context...
                @SuppressWarnings("unchecked") VariableExpression<E> ve = 
                    (VariableExpression<E>)expression;
                variables.add(ve);
            }
        }
        return variables;
    }

    /**
      * Returns the {@link VariableExpression variable} <b>or</b> 
      * {@link ConstantExpression constant} stored in
      * this context associated with the type and name
      * supplied as {@code type} and {@code name}, respectively. <p>
      * 
      * A variable need only be {@link #register(VariableExpression)
      * registered} to be retrieved with this method.
      *
      * @param  <E>  The type of value returned by the expression.
      * @param  type The type of value of the expression; cannot be null.
      * @param  name The name the expression; cannot be null.
      * @return The matching expression; never null.
      * @throws NullPointerException If either argument is null.         
      * @throws ExpressionException If no such expression exist.
      */
    public <E> VariableExpression<E> getVariable(Class<E> type, String name) throws ExpressionException {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        KeyValue<VariableExpression<?>,?> keyValue = this.values.get(name);
        if (keyValue != null) {
            VariableExpression<?> expression = keyValue.getKey();
            if (type.isAssignableFrom(expression.type())) {
                // Known to succeed because of the above check and the
                // rules enforced when adding expressions to this context...
                @SuppressWarnings("unchecked") VariableExpression<E> ve = 
                    (VariableExpression<E>)expression;
                return ve; 
            }
        }
        throw new ExpressionException("Name not found: " + name);
    }
    
    /**
      * Returns true if the {@link VariableExpression variable} 
      * expression supplied as {@code expression} has an 
      * associated {@link #assign(VariableExpression, Object) 
      * value} in this context, false if not. <p>
      * 
      * Variables that only have been {@link #register(VariableExpression)
      * registered} does <b>not</b> have an associated value!
      *  
      * @param  expression The expression to test; cannot be null.
      * @return True if {@code expression} has an associated value
      *         in this context, false if not.
      * @throws NullPointerException If {@code expression} is null.         
      * @see    #hasValue(String)
      * @see    #lookup(VariableExpression)         
      */
    public boolean hasValue(VariableExpression<?> expression) {
        if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        }
        KeyValue<?,?> value = this.values.get(expression.name);
        return ((value != null) && (value.getValue() != null));
    }
    
    /**
      * Returns true if a {@link VariableExpression variable} 
      * expression having the name supplied as {@code name} 
      * has an associated {@link #assign(VariableExpression, Object)
      * value} in this context, false if not.
      *  
      * Variables that only have been {@link #register(VariableExpression)
      * registered} does <b>not</b> have an associated value!
      *  
      * @param  name The variable name; cannot be null.
      * @return True if {@code expression} has an associated value
      *         in this context, false if not.
      * @throws NullPointerException If {@code name} is null.         
      * @see    #hasValue(VariableExpression)
      * @see    #lookup(String)         
      */
    public boolean hasValue(String name) {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        KeyValue<VariableExpression<?>,?> keyValue = this.values.get(name);
        if (keyValue != null) {
            return this.hasValue(keyValue.getKey());
        }
        return false;
    }
    
    /**
      * Returns true if the {@link VariableExpression variable} 
      * expression supplied as {@code expression} has been 
      * {@link #register(VariableExpression) registered} <b>or</b> has 
      * an associated {@link #assign(VariableExpression, Object)
      * value} in this context, false if not.
      *  
      * @param  expression The expression to test; cannot be null.
      * @return True if {@code expression} has been registered or has 
      *         an associated value in this context, false if not.
      * @throws NullPointerException If {@code expression} is null.         
      * @see    #exist(String)
      * @see    #lookup(VariableExpression)         
      */
    public boolean exist(VariableExpression<?> expression) {
        if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        }
        return this.values.containsKey(expression.name);
    }

    /**
      * Returns true if a {@link VariableExpression variable} 
      * expression having the name supplied as {@code name} 
      * has been {@link #register(VariableExpression) registered}
      * <b>or</b> has an associated {@link #assign(VariableExpression, Object) 
      * value} in this context, false if not.
      *  
      * @param  name The variable name; cannot be null.
      * @return True if such an expression has an associated value
      *         in this context, false if not.
      * @throws NullPointerException If {@code name} is null.         
      * @see    #exist(VariableExpression)
      * @see    #lookup(String)         
      */
    public boolean exist(String name) {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        return this.values.containsKey(name);
    }
    
    /**
      * Returns the type of the value associated with the 
      * {@link VariableExpression variable} expression
      * identified by the variable name supplied as {@code name}, 
      * or throws an {@link ExpressionException} in case no 
      * such value exist. <p>
      * 
      * A variable need only be {@link #register(VariableExpression)
      * registered} for its type to be looked up. 
      *  
      * @param  name The variable name; cannot be null.
      * @return The type of the associated value; never null.
      * @throws NullPointerException If {@code name} is null.         
      * @throws ExpressionException If no such value exist.
      * @see    #exist(String)
      */
    public Class<?> lookupType(String name) throws ExpressionException {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        KeyValue<VariableExpression<?>,?> keyValue = this.values.get(name);
        if (keyValue != null) {
            return keyValue.getKey().type();
        }
        throw new ExpressionException("Name not found: " + name);
    }
    
    /**
      * Returns the value currently associated with the 
      * {@link VariableExpression variable} expression
      * supplied as {@code expression}. <p>
      * 
      * A value that has a type assignable to {@code E} must have 
      * been {@link #assign(VariableExpression, Object) assigned} 
      * to {@code expression} in this context or an {@link ExpressionException} 
      * will be thrown.  
      *  
      * @param  <E>        The type of value the evaluation of 
      *                    {@code expression} produces.
      * @param  expression The expression to find the value of; 
      *                    cannot be null.
      * @return The associated value, having the proper type; 
      *         never null.
      * @throws ExpressionException If no such value exist, or if
      *         {@code E} does not match the type.
      * @see    #lookup(String)
      * @see    #hasValue(VariableExpression)
      * 
      */
    public <E> E lookup(VariableExpression<E> expression) throws ExpressionException {
        if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        }
        KeyValue<?,?> keyValue = this.values.get(expression.name);
        if ((keyValue == null) || (keyValue.getValue() == null)) {
            throw new ExpressionException("Expression not found: " + expression);
        }
        // This may cause a class cast exception...
        try {
            return expression.type().cast(keyValue.getValue());
        } catch (ClassCastException cce) {
            throw new ExpressionException(cce);
        }
    }

    /**
      * Returns the value currently associated with the 
      * {@link VariableExpression variable} expression
      * identified by the variable name supplied as {@code name}. <p>
      * 
      * A value must have been {@link #assign(VariableExpression, Object) assigned} 
      * to {@code expression} in this context or an {@link ExpressionException} 
      * will be thrown.  
      *  
      * @param  name The variable name; cannot be null.
      * @return The associated value; never null.
      * @throws NullPointerException If {@code name} is null.         
      * @throws ExpressionException If no such value exist.
      * @see    #lookup(VariableExpression)
      * @see    #hasValue(String)         
      */
    public Object lookup(String name) throws ExpressionException {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        KeyValue<VariableExpression<?>,?> keyValue = this.values.get(name);
        if (keyValue != null) {
            return this.lookup(keyValue.getKey());
        }
        throw new ExpressionException("Name not found: " + name);
    }

    /**
      * Registers the {@link VariableExpression variable} expression
      * supplied as {@code expression} in this context with <b>no</b> 
      * associated value. <p>
      * 
      * Registered variables can tested for {@link #exist(VariableExpression)
      * existence} and {@link #lookupType(String) type}; can be
      * retrieved using the {@link #getVariable(Class, String)} method;
      * but an associated value <b>cannot</b> be 
      * {@link #lookup(VariableExpression) retrieved}! <p> 
      * 
      * Once {@code expression} has been {@link #assign(VariableExpression, Object)
      * assigned} a value, invoking this method again will cause an
      * exception to be thrown. Registering {@code expression} several
      * times is allowed. 
      * 
      * @param  expression The expression to register; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      * @throws ExpressionException If {@code expression} has an associated
      *         value.
      * @see    #exist(String)
      * @see    #exist(VariableExpression)
      * @see    #lookupType(String)
      * @see    #getVariable(Class, String)        
      */
    public void register(VariableExpression<?> expression) throws ExpressionException {
        if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        }
        KeyValue<VariableExpression<?>,Object> keyValue = this.values.get(expression.name);
        if (keyValue == null) {
            this.values.put(expression.name, new KeyValue<VariableExpression<?>,Object>(expression));
        } else if (keyValue.getValue() != null) {
            throw new ExpressionException("Expression has an associated value");    
        }
    }
    
    /**
      * Assigns the value supplied as {@code value} to the
      * {@link VariableExpression variable} expression supplied
      * as {@code expression}.
      * 
      * @param  <E>        The type of value.
      * @param  expression The expression to assign {@code value} to;
      *                    cannot be null.
      * @param  value      The value to associate with {@code expression};
      *                    cannot be null.
      * @return The old value associated with {@code expression}, or
      *         null.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the old value has an incompatible type
      *         compared to the type used by {@code expression}, or if
      *         {@code expression} represents a constant with a 
      *         different value. 
      */
    public <E> E assign(VariableExpression<E> expression, E value) throws ExpressionException {
        if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        } else if (value == null) {
            throw new NullPointerException("Value cannot be null");
        }
        Log out = LogFactory.getLog(this);
        
        KeyValue<VariableExpression<?>,Object> keyValue = this.values.get(expression.name);
        Object oldValue = keyValue == null ? null : keyValue.getValue();
        E old = null;
        if (oldValue != null) {
            try {
                // If the cast fails, an incompatible value is tried assigned!
                old = expression.type().cast(oldValue);
                // A constant cannot change its value...
                if (expression instanceof ConstantExpression) {
                    if (!oldValue.equals(value)) {
                        LogFactory.getLog(this).error("Constant cannot change its value: ", oldValue, " != ", value);
                        throw new ExpressionException("Constants are immutable: " + expression);
                    }
                }
                
            } catch (ClassCastException cce) {
                out.error("Old value has different type: ", oldValue, " != ", value).error(cce);
                throw new ExpressionException("Expression cannot change type: " + expression);
            }
        }
        // Check constant value...
        if (expression instanceof ConstantExpression) {
            ConstantExpression<E> ce = (ConstantExpression<E>)expression;
            if (!ce.value.equals(value)) {
                out.error("Constant values different: ", ce.value, " != ", value);
                throw new ExpressionException("Illegal constant value: " + value);
            }
        }
        
        // Add...
        if (keyValue == null) {
            this.values.put(expression.name, new KeyValue<VariableExpression<?>,Object>(expression, value));
            
        // Update, since already registered...
        } else if (oldValue == null) {
            keyValue.setValue(value);
        }
        out.println("Assigned: ", expression, " -> ", value);
        
        return old;
    }
    
    /**
      * Returns the string representation of this context. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; size: ").append(this.values.size())
                 .append("; values: ").append(this.getVariables())
                 .append(']').toString();
    }
    
}
