/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An <i>interpreter</i> can interpret {@link Expression expressions}. <p>
  * 
  * Only expressions evaluating to values of a given type can be used.
  * The type is supplied at {@link #Interpreter(Class) construction} time. <p>
  * 
  * An interpreter supports {@link BreakExpression break} expressions 
  * storing {@link TypedExpression typed} expressions with a matching 
  * (sub-) {@link #type() type}. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @see    Context
  * 
  * @param  <T> The type of value the evaluated expressions evaluate to. 
  */
public class Interpreter<T> {

    /**
      * The super-type of expression values handled by this 
      * interpreter. <p>
      * 
      * Never null.
      */
    private final Class<T> type;
    
    /**
      * Constructor.
      * 
      * @param  type The super-type of expression values handled by this 
      *              interpreter. 
      * @throws NullPointerException If {@code type} is null.
      */
    public Interpreter(Class<T> type) {
        if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        }
    }

    /**
      * Constructor.
      * 
      * @param  type The super-type of expression values handled by this 
      *              interpreter. 
      * @throws NullPointerException If {@code type} is null.
      */
    public Interpreter(TypeLiteral<T> type) {
        if ((this.type = type.asClass()) == null) {
            throw new NullPointerException("Type literal cannot be null");
        }
    }
    
    /**
      * Return the super-type of expression values handled by this 
      * interpreter. 
      * 
      * @return The type; never null.
      * @see    TypedExpression#type()
      */
    public Class<T> type() {
        return this.type;
    }
    
    /**
      * Interpret the {@link Expression expression} syntax tree having
      * the root supplied as {@code expression}, and returns the 
      * result. <p>
      * 
      * The is no guarantee that the evaluation will ever complete!
      * 
      * @param  <E>        The type of value the evaluation of {@code expression} 
      *                    produces.
      * @param  context    The variable context to use; cannot be null.
      * @param  expression The root expression; cannot be null.
      * @return The result of the evaluation, or null if the evaluation
      *         was {@link BreakExpression exited} prematurely.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the evaluation fails.
      *
      */
    public <E extends T> T interpret(Context context, Expression<E> expression) throws ExpressionException {
        return this.interpret(context, expression, this.type);
    }

    /**
      * Interpret the {@link Expression expression} syntax tree having
      * the root supplied as {@code expression}, and returns the 
      * result. <p>
      * 
      * The is no guarantee that the evaluation will ever complete!
      * 
      * @param  <E>        The type of value the evaluation of {@code expression} 
      *                    produces.
      * @param  context    The variable context to use; cannot be null.
      * @param  expression The root expression; cannot be null.
      * @param  type       The type of value the evaluation of {@code expression}
      *                    produces; cannot be null.
      * @return The result of the evaluation, or null if the evaluation
      *         was {@link BreakExpression exited} prematurely.
      * @throws NullPointerException If either argument is null.
      * @throws ExpressionException If the evaluation fails.
      * 
      */
    public <E extends T> E interpret(Context context, Expression<E> expression, Class<? super E> type) throws ExpressionException {
        if (context == null) {
            throw new NullPointerException("Context cannot be null");
        } else if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        } else if (type == null) {
            throw new NullPointerException("Type cannot be null");
        }
        
        Log out = LogFactory.getLog(this);
        String exp = null;
        try {
            out.println("Context:         ", context);
            out.println("Interpreting:    ", expression.name(), " -> \"", exp = expression.asSymbol(context.reset()), "\"");
            out.println("Expression type: ", type.getName());
            
            return expression.evaluate(context);
         
        // Crude and ugly, but effective...
        } catch (BreakExpression.BreakException be) {
            if (be.expression != null) {
                if (type.isAssignableFrom(be.expression.type())) {
                    if (!type.equals(be.expression.type())) {
                        out.warn("Refining expression type: ", type.getName(), " -> ", be.expression.type().getName());
                    }
                    out.warn("Breaking and transferring control to: ", be.expression.asSymbol(context.reset()));
                    // Known to succeed because of the above check, though the actual
                    // type may not be E. We do not use the type from "be.expression" 
                    // because the re-interpretation may break again...
                    @SuppressWarnings("unchecked") Expression<E> e =(Expression<E>)be.expression;
                    return this.interpret(context, e, type);
                } else {
                    out.warn("Cannot handle expression of type: ", be.expression.type(), " -> ", be.expression.asSymbol(context.reset()));
                    // Fall-through
                }
            }
            out.warn("Interpretation exited");
            // Done..!
            return null;
            
        } catch (ExpressionException ee) {
            out.println().error("Interpretation failed", exp != null ? ": " + exp : "").error(ee);
            throw ee;
        }
    }

    
    /**
      * Returns the string representation of this interpreter. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; type: ").append(this.type.getName()).append(']').toString();
    }
    
}
