/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;

/**
  * A <i>logical or expression</i> for two expressions evaluating 
  * to a {@code Boolean} value. <p>
  * 
  * The logical OR operation may or may not be <i>short-circuit</i>, 
  * which is determined at {@link #OrExpression(boolean, Expression, Expression)
  * construction} time. <p<
  * 
  * The evaluation of a logical short-circuit OR expression is not 
  * commutative.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    AndExpression
  * @see    NotExpression
  * @see    CompareExpression
  */
@Participant("NonTerminalExpression")
public class OrExpression extends BinaryExpression<Boolean,Boolean> implements NonTerminalExpression<Boolean> {
    
    /**
      * True if a <i>short-circuit</i> logical OR operation is used, 
      * i.e. {@code ||}, false if not, i.e. {@code |}.  
      */
    public final boolean shortCircuit;
    
    /**
      * Constructor. <p>
      * 
      * The logical OR operation performed is <i>short-circuit</i>, 
      * i.e. {@code ||}.
      * 
      * @param  first  The first expression operand; cannot be null.
      * @param  second The second expression operand; cannot be null.
      * @throws NullPointerException If either expression is null.
      */
    public OrExpression(Expression<Boolean> first, Expression<Boolean> second) {
        this(true, first, second);
    }
    
    /**
      * Constructor. <p>
      * 
      * The logical OR operation performed is <i>short-circuit</i>, i.e. 
      * {@code ||}, if {@code shortCircuit} is true, otherwise not, 
      * i.e {@code |}.
      *
      * @param  shortCircuit True for short-circuit ({@code ||}).
      * @param  first        The first expression operand; cannot be null.
      * @param  second       The second expression operand; cannot be null.
      * @throws NullPointerException If either expression is null.
      */
    public OrExpression(boolean shortCircuit, Expression<Boolean> first, Expression<Boolean> second) {
        super(first, second);
        this.shortCircuit = shortCircuit;
    }
    
    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public OrExpression(OrExpression expression) {
        super(expression);
        this.shortCircuit = expression.shortCircuit;
    }

    /**
      *
      */
    public OrExpression copy() {
        return new OrExpression(this);
    }
 
    /**
      * @return {@code or}
      */
    public final String name() {
        return "or";
    }
    
    /**
      * 
      */
    public Boolean evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        if (this.shortCircuit) {
            return this.first.evaluate(context) || this.second.evaluate(context);
        } else {
            return this.first.evaluate(context) | this.second.evaluate(context);
        }
    }

    /**
      * Returns {@code (x || y)} if <i>short-circuit</i>, or {@code (x | y)},
      * where {@code x} is the symbolic representation of the
      * first expression operand and {@code y} of the second.  
      */
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder().append('(');
        return this.asSymbol(context, this.shortCircuit ? "||" : "|", sb).append(')').toString();
    } 
    
}
