/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.iterator;

import static dk.rode.thesis.meta.util.Const.INFINITE;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>logging value processor</i> logs the first {@code X} number of
  * sequence values during processing, where {@code X} is supplied
  * at construction time. <p>
  * 
  * A logging processor can handle any type of sequence values. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This processor mimics the functionality of the internal iterator 
  * described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.268].
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class LoggingValueProcessor implements ValueProcessor<Object> { 

    /**
      * The maximum number of sequence values to log. <p>
      * 
      * A value of {@link Const#INFINITE} means that the processing
      * will not be stopped by this processor. 
      */
    private final int maximum;
    /**
      * The current number of processed sequence values. <p>
      * 
      * Range: {@code 0 <= count <= min(}{@link #maximum}, {@code size-of-sequence)} <p>
      * 
      * This count will be set to zero on {@link #initialise() 
      * initialisation}.
      */
    private int count;

    /**
      * Constructor. <p>
      * 
      * All sequence values will be logged.
      */
    public LoggingValueProcessor() {
        this(INFINITE);
    }

    /**
      * Constructor.
      * 
      * @param  maximum The maximum number of sequence values to log;
      *                 must be larger than zero or {@link Const#INFINITE}.
      * @throws NullPointerException If {@code maximum} is illegal.                
      */
    public LoggingValueProcessor(int maximum) {
        if (maximum == INFINITE) {
            this.maximum = INFINITE;
        } else if (maximum < 1) { 
            throw new IllegalArgumentException("Illegal maximum value: " + maximum);
        } else {
            this.maximum = maximum;
        }
    }

    /**
      * Resets the internal counter to zero. 
      */
    public void initialise() {
        this.count = 0;
    }

    /**
      * 
      */
    public boolean process(Object value) {
        if (this.maximum == INFINITE) {
            LogFactory.getLog(this).println(value);
            return true;
        } else if (this.count < this.maximum) {
            this.count++;
            LogFactory.getLog(this).println(value);
            return true;
        }
        return false;
    }

    /**
      * Returns the string representation of this processor.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        if (this.maximum != INFINITE) {
            sb.append("; count: ").append(this.count)
              .append("; maximum: ").append(this.maximum);
        }
        return sb.append(']').toString();
    }
    
}
