/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.iterator;

import static dk.rode.thesis.strategy.SequencePolicy.IDENTITY;

import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>processable sequence</i> represents an <b>internal 
  * iterator</b> that can iterate over <b>bounded</b> 
  * {@link Sequence sequences} using a value {@link ValueProcessor 
  * processor} to perform the actual processing of delivered 
  * sequence values. The processing <b>will</b> advance the sequence. <p>
  * 
  * The iteration is performed using the {@link #process(ValueProcessor)} 
  * method. The processing will stop when the processable sequence has
  * no more values, or in case the processor returns false. <p>
  * 
  * The processing performed by a processable sequence is not 
  * thread-safe. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This implementation uses the {@linkplain dk.rode.thesis.decorator 
  * Decorator} and {@linkplain dk.rode.thesis.strategy Strategy} 
  * patterns as opposed to inheritance: the processing is delegated to a 
  * {@link ValueProcessor} supplied at iteration time, and hence this 
  * iterator need not be sub-classed. The internal iterator example 
  * provided by Gamma et al. uses inheritance, but Gamma et al. note 
  * that it might be feasible to parameterise the iterator with the 
  * processing operation in languages with different features than
  * those offered by C++ [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.267-269]. <p>
  * 
  * This is a specialised case of using composition instead of inheritance.
  * More specifically, the internal iterators presented by Gamma et al. 
  * can be viewed as variants of the 
  * {@linkplain dk.rode.thesis.templatemethod Template Method} pattern, and the
  * use of the Template Method pattern can thus be replaced by composition,
  * providing all required information is supplied as arguments. <p>
  * 
  * The is no interface in Java to support internal iterators.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    IterableSequence
  * @see    SequenceIterator
  * 
  * @param  <E> The type of values processed by this processable sequence.
  */
public class ProcessableSequence<E> extends SequenceDecorator<E> implements Sequence<E> {
    
    /**
      * Constructor, which creates this processable sequence to process
      * through the sequence values from {@code sequence}. <p>
      * 
      * Before each time this sequence is processed, it will be reset. <p>
      * 
      * Modifying {@code sequence} externally will affect this iterator.
      * 
      * @param  sequence The sequence to make iterable; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IllegalArgumentException If {@code sequence} is unbounded.
      */
    public ProcessableSequence(final Sequence<E> sequence) {
        super(sequence instanceof Iterable ? sequence : new IterableSequence<E>(sequence), IDENTITY);
    }

    /**
      * 
      */
    public ProcessableSequence<E> copy() {
        return new ProcessableSequence<E>(this.getSequence(true));
    }
    
    /**
      * Process values in this sequence using the {@link ValueProcessor 
      * processor} supplied as {@code processor} until all values have been
      * exhausted, or {@code processor} returns false. <p>
      * 
      * This sequence will be {@link Sequence#reset() reset} before 
      * the processing is performed, and the processor
      * {@link ValueProcessor#initialise() initialised}. <p>
      * 
      * The processing <b>will</b> advance this sequence. <p>
      *
      * @param  processor The value processor to use; cannot be null.
      * @return True if all sequence values were processed, false if
      *         {@code processor} terminated the iteration by
      *         returning false.
      * @throws NullPointerException If {@code processor} is null.         
      */
    public boolean process(ValueProcessor<? super E> processor) {
        // Cast known to succeed based on the rules enforced in
        // the constructor...
        IterableSequence<E> sequence = (IterableSequence<E>)this.getSequence(false);
        // Restart sequence...
        sequence.reset();
        // ..and (re-)initialise processor...
        processor.initialise();
        // Now process...
        for (E value : sequence) {
            if (!processor.process(value)) {
                return false;
            }
        }
        return true;
    }

}
