/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Memento</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.283]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Without violating encapsulation, capture and externalise 
  *    an objects internal state so that the object can be restored 
  *    to this state later. 
  *    
  * </i></ul>
  * 
  * Here, the <b>Memento</b> participant is represented by the 
  * {@link dk.rode.thesis.memento.SequenceMemento} interface,
  * which is applicable to {@link dk.rode.thesis.meta.model.Sequence}
  * objects only. <p>
  * 
  * The <b>Originator</b> participant is represented by any 
  * {@link dk.rode.thesis.memento.MemorizableSequence} 
  * implementation. Two specific implementations are defined in
  * this package, namely {@link dk.rode.thesis.memento.RangeSequence}
  * and {@link dk.rode.thesis.memento.MemorizableEnglishAlphabetSequence}. <p>
  * 
  * The test application, i.e. the {@link dk.rode.thesis.memento.Main}
  * class, acts as the <b>Caretaker</b> participant. <p>
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/Memento.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br/>
  * Internally, sequence mementos utilise the 
  * {@linkplain dk.rode.thesis.prototype Prototype} pattern to save
  * the sequence state. <p>
  * 
  * All memento functionality has to be declared public if it must
  * be used outside this package. It cannot have "two interfaces" in 
  * Java, a narrow (public) and a wide (using friends to access
  * private part), as described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.285-287]. 
  * This implementation therefore divides the memento functionality into 
  * two interfaces, namely
  * {@link dk.rode.thesis.memento.SequenceMemento} (private parts of wide) 
  * and {@link dk.rode.thesis.memento.MemorizableSequence} (narrow, public). 
  * However, the state methods {@link dk.rode.thesis.memento.SequenceMemento#setSequence(Sequence)},
  * {@link dk.rode.thesis.memento.SequenceMemento#getSequence()}, and 
  * {@link dk.rode.thesis.memento.SequenceMemento#current()} still have to 
  * be public. To enforce an access check at runtime, the 
  * {@link dk.rode.thesis.memento.GuardedSequenceMemento} can be used
  * in place of {@code SequenceMemento}. It identifies the 
  * {@link dk.rode.thesis.meta.reflect.CallerClass caller}
  * of the state methods and only allows calls from valid contexts. We
  * call types enforcing such runtime access checks for 
  * <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.2.1" target="thesis"><i>Guarded Types</i></a>. <p>
  * 
  * The {@code RangeSequence} implementation also supports Java's built 
  * in <i>serialization mechanism</i> by implementing the 
  * {@link java.io.Serializable java.io.Serializable} interface. It can thus 
  * be written to and read from {@link java.io.ObjectOutputStream java.io.ObjectOutputStream}
  * and {@link java.io.ObjectInputStream java.io.ObjectInputStream} objects,
  * respectively, to save and restore its state. Hence, the 
  * {@code java.io.Serializable} interface can also be considered a 
  * representative of the <b>Originator</b> participant, and the <i>target</i>
  * of the {@code OutputObjectStream} used, e.g. a file, therefore represents 
  * the <b>Memento</b> participant as well.  
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Memento", 
         scope        = Scope.Object,
         purpose      = Purpose.Behavioural,
         participants = {"Memento", "Originator", "Caretaker"})
package dk.rode.thesis.memento;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;
import dk.rode.thesis.meta.model.Sequence;
