/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

import static dk.rode.thesis.meta.util.Strings.empty;
import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;
import static dk.rode.thesis.meta.util.Throwables.toStackTrace;

import java.io.IOException;
import java.util.concurrent.atomic.AtomicInteger;

import dk.rode.thesis.meta.reflect.CallerClass;
import dk.rode.thesis.meta.util.Primitive;

/**
  * An <i>abstract log</i> implements the basic traits of 
  * any {@link Log} implementation. <p>
  * 
  * The logger is <b>not</b> thread-safe!
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public abstract class AbstractLog implements Log {

    /**
      * The class to which this log is associated, if any. <p>
      * 
      * Null indicates the global log.
      */
    private Class<?> clazz;
    /**
      * If not null, the line number representing the current
      * value is logged. <p>
      * 
      * If used, even empty lines logged will have a line number. <p>
      * 
      * Default false.
      */
    private final AtomicInteger lineNumber; // Could have used a Sequence... :)
    /**
      * True if active, i.e. logs messages, false if inactive.
      */
    private boolean active;
    /**
      * True if the last logged message ended with a newline, false
      * if not.  
      */
    private boolean flushed;
    
    /**
      * Constructor. <p>
      * 
      * Line numbers are logged if {@code lineNumbers} is not null.
      * 
      * @param  clazz      The class to which this log is associated, if any.
      *                    Null indicates global log.
      * @param  active     True if this log is initially active, false if not.
      * @param  lineNumber The line number values; can be null, in which
      *                    case line numbers are not generated.
      * @see    #isActive()
     */
    protected AbstractLog(Class<?> clazz, boolean active, AtomicInteger lineNumber) {
        this.clazz = clazz;
        this.setActive(active);
        this.flushed = true;
        this.lineNumber = lineNumber;
    }

    /**
      * 
      */
    public Log heading(Object... heading) {
        this.forceNewLine();
        if (this.active) {
            this.println();
            StringBuilder sb = new StringBuilder();
            for (Object o : heading) {
                sb.append(this.format(o));
            }
            int m = sb.length();
            this.println(sb);
            sb = new StringBuilder();
            for (int i = 0; i < m; i++) {
                sb.append('=');
            }
            this.println(sb);
            this.println();
            if (this.lineNumber == null) {
                this.println();
            }
        }
        return this;
    }

    /**
      * Prints a new line if this log is not flushed. 
      */
    private void forceNewLine() {
        if (this.active) {
            if (!this.flushed) {
                try {
                    this.log("\n");
                    this.flushed = true;
                } catch (Exception e) {
                    throw toRuntimeException(e);
                }
            }
        }
    }
    
    /**
      * 
      */
    public Log error(Object... messages) {
        this.forceNewLine();
        return this.println("> ERROR - ", messages);
    }

    /**
      * 
      */
    public Log warn(Object... messages) {
        this.forceNewLine();
        return this.println("> WARN - ", messages);
    }
    
    /**
      * 
      */
    public final Log print(Object... messages) {
        if (this.active) {
            if (messages.length > 0) {
                StringBuilder sb = new StringBuilder();
                for (Object message : messages) {
                    if (message == null) {
                        sb.append("null");
                        continue;
                    }
                    if (message.getClass().isArray()) {
                        for (Object m : (Object[])message) {
                            sb.append(this.format(m));
                        }
                    } else {
                        sb.append(this.format(message));
                    }
                }
                try {
                    if (empty(sb)) {
                        if ((this.lineNumber != null) && (this.flushed)) {
                            this.log(this.lineNumber.getAndIncrement(), '\t');
                        }
                        this.log(sb);
                        
                    } else if (!this.flushed) {
                        this.log(sb);
                        
                    } else {
                        this.log(this.getPrefix(), sb);
                    }
                    this.flushed = false;
                    
                } catch (Exception e) {
                    throw toRuntimeException(e);
                }
            }
        }
        return this;
    }

    /**
      * 
      */
    public final Log print(boolean condition, Object... messages) {
        if (condition) {
            this.print(messages);
        }
        return this;
    }
    
    /**
      * 
      */
    public final Log println(Object... messages) {
        this.print(messages, '\n');
        this.flushed = true;
        return this;
    }

    /**
      * 
      */
    public final Log println(boolean condition, Object... messages) {
        if (condition) {
            this.print(messages, '\n');
        } else {
            this.print('\n');
        }
        this.flushed = true;
        return this;
    }
    
    /**
      * Performs the actual logging of the messages supplied
      * as {@code messages} with <b>no new line postfixed</b>. <p>
      * 
      * All messages part are assumed to have a meaningful 
      * {@code toString()} representation and are not null, 
      * and there is no need for explicit formatting.
      * 
      * @param  messages The messages to log; never null.
      * @throws IOException In case an exception occurs while logging.
      */
    protected abstract void log(Object... messages) throws IOException;
    
    /**
      * 
      */
    public String getPrefix() {
        String prefix = getPrefix(this.clazz);
        if (this.lineNumber != null) {
            prefix = this.lineNumber.getAndIncrement() + "\t" + prefix;
        }
        return prefix;
    }
    
    /**
      * Returns the current prefix to prefix any log message based
      * on the class supplied as {@code clazz}. <p>
      * 
      * The prefix will include the name of the member and line number
      * of the calling context that issued the original log request.
      *   
      * @param  clazz The current class; can be null.
      * @return The current prefix; never null, and never empty.
      * @see    CallerClass
      */
    public final static String getPrefix(Class<?> clazz) {
        StringBuilder prefix = new StringBuilder();
        if (clazz == null) {
            prefix.append("Main:");
        } else {
            prefix.append(clazz.getSimpleName()).append('.');            
        }
        CallerClass caller = new CallerClass().getNonAssignableCaller();
        return prefix.append(caller.element.getMethodName())
                     .append(':')
                     .append(caller.element.getLineNumber())
                     .append(":: ").toString();
    }
    
    /**
      * Formats the object supplied as {@code object} into
      * a char sequence representation that can be logged.
      * 
      * @param  object The object; can be null.
      * @return The char sequence representation; never null.
      */
    protected CharSequence format(Object object) {
        if (object == null) {
            return "null";
            
        } else if (object.getClass().isArray()) {
            StringBuilder sb = new StringBuilder();
            if (object.getClass().getComponentType().isPrimitive()) {
                sb.append(Primitive.toString(object));
            } else {
                for (Object o : (Object[])object) {
                    sb.append(this.format(o));
                }
            }
            return sb;
            
        } else if (object instanceof Throwable) {
            return toStackTrace((Throwable)object);
        }
        return object.toString();
    }
    
    /**
      * 
      */
    public int getLineNumber() {
        return this.lineNumber == null ? 0 : this.lineNumber.get(); 
    }
    
    /**
      * 
      */
    public final boolean isActive() {
        return this.active;
    }
    
    /**
      * 
      */
    public final Log setActive(boolean active) {
        this.active = active;
        return this;
    }

    /**
      * Returns the string representation of this log. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        return new StringBuilder()
              .append(this.getClass().getSimpleName())
              .append("[associated: ")
              .append(this.clazz == null ? "global" : this.clazz.getName())
              .append("; active: ")
              .append(this.active)
              .append(']').toString();
    }
    
}
