/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

import java.io.Closeable;
import java.io.IOException;
import java.util.concurrent.atomic.AtomicInteger;

/**
  * An <i>appendable log</i> uses an {@code Appendable} object as
  * the target for log messages, for example a {@code PrintStream} 
  * such as {@code System.out}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class AppendableLog extends AbstractLog implements Log {

    /**
      * The target for log messages. <p>
      * 
      * Never null.
      */
    private final Appendable out;
    /**
      * True if {@link #out} should be closed when this log goes
      * out of scope, false if not.  
      */
    private final boolean close;
    
    /**
      * Constructor, which creates this log as active. <p>
      * 
      * {@code out} is not closed by this log. <p>
      * 
      * No line numbers are logged.
      * 
      * @param  out   The target for log messages; cannot be null.
      * @param  clazz The class this log is associated with, if any;
      *               can be null.
      * @throws NullPointerException If {@code out} is null.
      */
    AppendableLog(Appendable out, Class<?> clazz) {
        this(out, false, clazz, true, null);
    }
    
    /**
      * Constructor, which creates this log as active if {@code active}
      * is true. <p>
      * 
      * {@code out} is closed by this log if {@code close} is true.
      * 
      * @param  out         The target for log messages; cannot be null.
      * @param  close       True to close {@code out} when this log goes
      *                     out of scope.
      * @param  clazz       The class this log is associated with, if any;
      *                     can be null.
      * @param  active      True if this log is initially active, false if not.
      * @param  lineNumber  Line number values; can be null, in which case
      *                     line numbers are not generated.
      * @throws NullPointerException If {@code out} is null.
      * @see    #isActive()
      */
    AppendableLog(Appendable out, boolean close, Class<?> clazz, boolean active, AtomicInteger lineNumber) {
        super(clazz, active, lineNumber);
        if ((this.out = out) == null) {
            throw new NullPointerException("Appendable cannot be null");
        }
        this.close = close;
    }
    
    /**
      * 
      */
    protected void log(Object... messages) throws IOException {
        for (Object message : messages) {
            this.out.append(message.toString());
        }
    }
    
    /**
      * Closes the appendable used by this logger if so specified
      * at construction time and if the appendable is closeable. 
      */
    protected void finalize() {
        if (this.close) {
            if (this.out instanceof Closeable) {
                try {
                    ((Closeable)this.out).close();
                } catch (IOException ioe) {
                    // Intentionally left blank...
                }
            }
        }
    }
    
}
