/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.model;

import java.util.Arrays;

/**
  * A simple abstract <i>array sequence</i> to deliver all the 
  * entries in a given array, in order. An invocation of
  * {@link #next()} returns the next <i>element</i> in the array. <p>
  * 
  * Sub-classes need only supply the actual array on construction. <p>
  * 
  * An array sequence is {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and unless otherwise
  * specified {@link #unique()}. <p>
  * 
  * Array sequences are by default equal to any array sequence 
  * of the <b>same type</b> using an <b>equivalent array</b> that 
  * is at the <b>same sequence index</b>. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
public abstract class ArraySequence<E> 
              extends AbstractSequence<E> 
           implements Sequence<E> {

    /**
      * The array used. <p>
      * 
      * Never null, and never empty. <p>
      */
    protected final E[] elements;
    /**
      * The current index. <p>
      * 
      * Invariant: 0 <= index < {@link #elements}.length
      */
    protected int index; 
    
    /**
      * Constructor, which creates this sequence to use the
      * array supplied as {@code array}. <p>
      * 
      * The array is not copied, but used as is.
      * 
      * @param  array The array; cannot be null or empty.
      * @throws NullPointerException If {@code array} is null.
      * @throws IllegalArgumentException If {@code array} is empty.
      */
    protected ArraySequence(E... array) {
        super();
        if ((this.elements = array) == null) {
            throw new NullPointerException("Array cannot be null");
        } else if (this.elements.length == 0) {
            throw new IllegalArgumentException("Array cannot be empty");
        }
    }

    /**
      * Copy constructor. <p>
      * 
      * The array is not copied, but used as is. This sequence starts 
      * at the current index of {@code sequence}.
      * 
      * @param  sequence The array sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected ArraySequence(ArraySequence<E> sequence) {
        super(sequence);
        this.elements = sequence.elements;
        this.index = sequence.index;
    }
    
    /**
      * Returns the number of elements in this array sequence.
      *  
      * @return The number of elements.
      */
    public int size() {
        return this.elements.length;
    }
    
    /**
      * @return True, as the array used by this sequence is finite.
      * @see    #size()
      */
    public final boolean bounded() {
        return true;
    }

    /**
      * @return True, as the array used by this sequence is finite.
      */
    public final boolean consistent() {
        return true;
    }

    /**
      * @return True.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * 
      */
    public void reset() {
        super.reset();
        this.index = 0;
    }
    
    /**
      *  
      */
    public E current() {
        return this.elements[this.index];
    }

    /**
      * 
      */
    public E next() {
        this.state = (this.index == this.elements.length - 1) ? State.RESTART : State.NORMAL;
        return this.elements[this.index = (this.index + 1) % this.elements.length];
    }
    
    /**
      * Returns the hash code of this sequence.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.elements.hashCode();
    }
    
    /**
      * Returns true if {@code object} is an array sequence having the 
      * <b>same type</b> as this array sequence, using an <b>equivalent 
      * array</b>, and is at the <b>same sequence index</b>, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
        } else if (this.getClass().equals(object.getClass())) {
            ArraySequence<?> as = (ArraySequence<?>)object;
            if (this.index != as.index) {
                return false;   
            } else if (Arrays.equals(this.elements, as.elements)) {
                return true;
            }
        }
        return false;
    }
    
}
