/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.model;

/**
  * A <i>reversible sequence</i> allows retrieval of previous
  * values <b>if and only if</b> such previous values are 
  * {@link #reversible() ready} for delivery, i.e. have 
  * already been calculated. <p>
  * 
  * By invoking {@link #reverse()} on a reversible sequence,
  * the order to advance is reversed, from "next" to "previous" value, 
  * and back to "next" from "previous" the next time {@code reverse()}
  * is invoked, and so forth. <p> 
  * 
  * Reversible sequences are by nature <b>always</b> 
  * {@link Sequence#bounded() bounded} and {@link Sequence#consistent() consistent}. <p> 
  * 
  * <b>Examples</b>:
  * <pre>
  *    Consistent, bounded, unique, reversible number sequence: [0, 1, 2, 3, 4, 5]
  *    
  *    current(): 0   
  *    next():    1               
  *    next():    2               
  *    next():    3               
  *    current(): 3               
  *    next():    4               
  *    next():    5  
  *    next():    1     
  *    next():    2               
  *    reverse(): 2
  *    next():    1               
  *    reset():   0    
  *    next():    5
  *    next():    4
  *    next():    3
  *    next():    2
  *    reverse(): 2
  *    next():    3
  *    reset():   0
  *    next():    1
  *    reset():   0
  *    reverse(): 0
  *    next():    5
  *    ..
  * </pre>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this reversible sequence. 
  */
public interface ReversibleSequence<E> extends Sequence<E> {
    
    /**
      * Reverses this reversible sequence, if currently
      * {@link #reversible() reversible}. <p>
      * 
      * The {@link Sequence#current() current} value is <b>not</b>
      * affected by this method, and is returned by this method.
      *  
      * @return The current sequence value; never null.
      * @throws IllegalStateException If this sequence cannot be
      *         reversed at this point.
      */
    public E reverse();

    /**
      * Returns true if this reversible sequence can be reversed
      * in its current state, false if not. <p>
      * 
      * If this method returns true, it is safe to invoke
      * {@link #reverse() reverse()}. If it returns false, an illegal
      * state exception will be thrown by {@code reverse()}.
      * 
      * @return True if this reversible sequence can be reversed in
      *         its current state, false if not.
      */
    public boolean reversible();
    
    /**
      * Returns a copy of this reversible sequence that will start 
      * at the same sequence index as this sequence in the same
      * direction. 
      *  
      * @return A copy of this sequence; never null.
      */
    public ReversibleSequence<E> copy();
    
}
