/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import static dk.rode.thesis.meta.util.Util.copy;

import java.util.Arrays;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.util.Throwables;

/**
  * An <i>identified stack trace element</i> is used to identify a 
  * stack trace element representing the <i>calling context</i>. <p>
  * 
  * The calling context can be: <ol> <p>
  *   <li> A static initialisation block (class);
  *   <li> An initialisation block (instance); 
  *   <li> A constructor; or
  *   <li> A method.
  * </ol>
  * 
  * An identified stack trace element <i>per default</i> identifies the 
  * stack trace element corresponding to the context where an instance 
  * of this class was constructed. The element thus represents the 
  * <i>immediate calling context</i>, e.g. the static initialisation block,
  * instance initialisation block, constructor, or method from where the 
  * constructor of this class was invoked. The identified element is 
  * stored in {@link #element}, and the entire call-stack in 
  * {@link #stack}. The zeroth element of the call-stack represents the 
  * top of the stack, corresponding to one of the constructors declared 
  * in this class <i>or</i> in a sub-class. <p>
  * 
  * However, a <i>stack offset</i> can be specified at construction time 
  * to identify a stack trace element from anywhere within the call-stack 
  * (from the top of the stack), corresponding to a <i>relative calling 
  * context</i>. The field {@link #offset} represents the stack offset used, 
  * relative to the top of the stack: zero = top of stack, 1 = second element, 
  * 2 = third element, etc. <p>   
  * 
  * All instances of this class are immutable. If an instance can
  * be created, a stack trace element will have been identified. <p>
  * 
  * <b>Example</b>: a class {@code Foo} defined in the {@code Foo.java} file:
  * <pre>
  * 01  class Foo {
  * 02    // Static initialisation block...
  * 03    static { 
  * 04      System.out.print(new IdentifiedStackTraceElement().toString());
  * 05    }
  * 06
  * 07    // Instance initialisation block... 
  * 08    {
  * 09      System.out.print(new IdentifiedStackTraceElement().toString());
  * 10    }
  * 11  
  * 12    // Constructor... 
  * 13    public Foo() {
  * 14      System.out.print(new IdentifiedStackTraceElement().toString());
  * 15    }  
  * 16
  * 17    public void bar() {
  * 18      System.out.print(new IdentifiedStackTraceElement().toString());
  * 19      this.barbar();
  * 20    }
  * 21
  * 22    public void barbar() {
  * 23      System.out.print(new IdentifiedStackTraceElement(1).toString());
  * 24    }
  * 25  }
  * </pre>
  * Line 4 generates the following output (the precise format of the 
  * output is dependent on the JVM):
  * <pre>
  *   "Foo.&lt;clinit>(Foo.java:4)"
  * </pre>
  * Line 9 and 14 generates the following output, respectively:
  * <pre>
  *   "Foo.&lt;init>(Foo.java:9)"       and
  *   "Foo.&lt;init>(Foo.java:14)"
  * </pre>
  * Line 18 and 23 generate the <b>same</b> output (provided access through 
  * {@code bar()}), since a stack offset (1) is specified in {@code barbar()}:
  * <pre>
  *   "Foo.bar(Foo.java:18)"
  * </pre>
  *
  * Note, that a JVM is not required to supply stack traces!
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class IdentifiedStackTraceElement {
    
    /**
      * The identified stack trace element. <p>
      * 
      * It represents the point of origin where this instance 
      * was constructed from relative to the {@link #offset} 
      * supplied at construction time. <p>
      * 
      * Never null.
      */
    public final StackTraceElement element;
    /**
      * The offset of {@link #element} from the top of the call-stack
      * stored in {@link #stack}. <p>
      * 
      * A value of zero represents the top of the call-stack. <p>
      * 
      * Invariant: {@link #stack}{@code [offset] == }{@link #element} is 
      * <b>always</b> true.  
      */
    public final int offset;
    /**
      * The call-stack used to identify the stack trace element  
      * stored in {@link #element}. <p>
      * 
      * The zeroth element of this call-stack represents the top of 
      * the call-stack, corresponding to one of the constructors 
      * declared in this class or in a sub-class. Hence, {@link #offset} zero = 
      * top of stack, 1 = second element, 2 = third element, etc. <p>
      *  
      * Never null.
      */
    protected final StackTraceElement[] stack;
    /**
      * The throwable that supplied the stack trace 
      * stored in {@link #stack}. <p>
      * 
      * Never null.
      */
    protected final Throwable throwable;
    
    /**
      * No-arg constructor, which identifies the stack trace element 
      * corresponding to the <i>calling context</i> that invoked 
      * this constructor.
      */  
    public IdentifiedStackTraceElement() {
        // +1 for this constructor...
        this(1);
    }
    
    /**
      * Constructor, which identifies the stack trace element 
      * at {@code offset} from the top of the current call-stack. <p>
      * 
      * Invoking with a value of zero corresponds to invoking the 
      * default constructor, i.e., the identified stack trace element 
      * corresponds to the context that invoked this constructor; a 
      * value of one will resolve the context to the calling context 
      * of this constructor; and so forth.
      *   
      * @param  offset The offset from the top of the call-stack; 
      *                must be within bounds.
      * @throws IndexOutOfBoundsException If {@code offset} does not 
      *         denote a valid stack element.
      */
    public IdentifiedStackTraceElement(int offset) {
        // +1 for this constructor and +1 for the 
        // constructor about to be invoked...
        this(new Throwable(), offset + 2);  
    }    

    /**
      * Constructor, which identifies the stack trace element 
      * at {@code offset} from the top of the call-stack acquired
      * via {@code throwable}; hence, this constructor need
      * not be in the call-stack at all! <p>
      * 
      * @param  throwable The throwable to supply the call-stack;
      *                   cannot be null.
      * @param  offset    The offset from the top of the call-stack; 
      *                   must be within bounds.
      * @throws NullPointerException If {@code throwable} is null.
      * @throws IndexOutOfBoundsException If {@code offset} does not 
      *         denote a valid stack element.
      */
    public IdentifiedStackTraceElement(Throwable throwable, int offset) {
        if ((this.throwable = throwable) == null) {
            throw new NullPointerException("Throwable cannot be null");
        }
        this.stack = this.throwable.getStackTrace();
        if (this.stack.length == 0) {
            throw new Error("Stack trace has zero length");      
        } 
        this.element = this.stack[this.offset = offset]; 
    }
    
    /**
      * Copy constructor, which identifies the <b>same</b> stack 
      * trace element as {@code element} in the <b>same</b> 
      * call-stack at the <b>same</b> offset.
      * 
      * @param  element The identified stack trace element to copy; 
      *                 cannot be null.
      * @throws NullPointerException If {@code element} is null.
      */
    public IdentifiedStackTraceElement(IdentifiedStackTraceElement element) {
        this(element, element.offset);
    }
 
    /**
      * Constructor, which identifies the stack trace element 
      * at {@code offset} in the call-stack used by
      * {@code element}. <p>
      * 
      * The offset is used as is.
      * 
      * @param  element The identified stack trace element; 
      *                 cannot be null.
      * @param  offset  The offset from the top of the call-stack; 
      *                 must be within bounds.
      * @throws NullPointerException If {@code element} is null.
      * @throws IndexOutOfBoundsException If {@code offset} does not 
      *         denote a valid stack element.
      */
    IdentifiedStackTraceElement(IdentifiedStackTraceElement element, int offset) {
        if (element == null) {
            throw new NullPointerException("Element cannot be null");
        }
        this.stack = element.stack;
        this.element = this.stack[this.offset = offset];
        this.throwable = element.throwable;
    }
     
    /**
      * Returns a copy of the call-stack used by this 
      * identified stack trace element.
      * 
      * @return A copy; never null.
      */
    public StackTraceElement[] getStackTrace() {
        return copy(this.stack, new StackTraceElement[this.stack.length]);
    }
    
    /**
      * Prints the call-stack to {@code out} with a
      * severity level of <i>error</i>.
      * 
      * @param  out The log to use; cannot be null.
      * @throws NullPointerException If {@code out} is null. 
      */
    public void printStackTrace(Log out) {
        out.error(Throwables.toStackTrace(this.throwable));
    }
        
    /**
      * Returns the hash code of this identified 
      * stack trace element.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.element.hashCode();   
    }
     
    /**
      * Returns true if {@code object} has the same type
      * as this element, same offset, same element, and
      * same stack trace, otherwise false.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not. 
      */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
        } else if (this.getClass().equals(object.getClass())) {
            IdentifiedStackTraceElement iste = (IdentifiedStackTraceElement)object;
            return ((this.offset == iste.offset) &&
                    (this.element.equals(iste.element)) &&
                    (Arrays.equals(this.stack, iste.stack)));
        }
        return false;
    }
    
    /**
      * Returns the string representation of this identified stack
      * trace element. <p>
      * 
      * The representation is that of the identified {@code StackTraceElement}.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        return this.element.toString();
    }
      
}
