/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.reflect.proxy;

import static dk.rode.thesis.meta.util.Util.add;

import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.CallerClass;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.meta.util.Primitive;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>proxy factory</i> creates <i>dynamic proxies</i>, which
  * can also be used for <i>duck typing</i>. <p>
  *  
  * The proxy objects are based on the 
  * {@link java.lang.reflect.Proxy java.lang.reflect.Proxy} 
  * class, and must therefore adhere to the rules set forth by that
  * class. For example, an implication of this is that even if a 
  * specific instance is supplied as the {@code object} argument
  * to the {@link #getProxy(Object, InvocationHandler, Class...)}
  * method, i.e. a specific class, the returned (proxy) object 
  * <b>cannot</b> be cast back into that type (the specific type 
  * information is "lost" as the class returned is an anonymous 
  * {@code Proxy} class). <p>
  * 
  * Per default, the class loader used to load this class is used
  * as the class loader to define the created proxies, but a specific 
  * class loader can be supplied at {@link #ProxyFactory(ClassLoader)
  * construction} time. <p>
  * 
  * This class defines four specific proxy methods: <ol>
  * 
  * <li> {@link #getLoggableObject(Object, Class...)}: returns a proxy
  *      that will record and time all access to the proxied object
  *      <b>if and only if</b> all access is through the proxy only. <p> 
  *      
  * <li> {@link #getSynchronisedObject(Object, Object, Class...)}: returns
  *      a proxy that guarantees thread-safe access to the proxied object
  *      <b>if and only if</b> all access is through the proxy only. <p> 
  * 
  * <li> {@link #getSharedObject(Object, String, Class[], String...)}: returns
  *      a proxy of a given object that can be shared until a 
  *      <i>mutator</i> method is invoked, causing the original referenced
  *      object to be copied (<i>smart reference</i>), <b>if and only if</b>
  *      all access is through the proxy or copies of it only. <p>
  *       
  * <li> {@link #getNullObject(Class)}: returns a to this factory unique 
  *      <i>null object</i> of the type specified as the argument. Only a 
  *      single instance of the type in question will be created, and the
  *      type may be either an interface or class. This can be used to
  *      mask null where null is not allowed. <p>
  *      
  * </ol>
  *  
  * A proxy factory only maintain weak references to proxied objects, though
  * individual handlers may store strong references! <p>
  * 
  * A proxy factory is open for sub-classing. <p>
  * 
  * Please refer to the JavaDoc of the {@code java.lang.reflect.Proxy} 
  * class for in-depth semantics regarding proxies. The JavaDoc for
  * the three methods described above contains examples on general
  * usage of {@code Proxy} objects as well. <p>   
  * 
  * <b>Implementation notes</b>:<br>
  * The nested classes defined in this class represent applications
  * of the {@linkplain dk.rode.thesis.adapter Adapter} and 
  * {@linkplain dk.rode.thesis.decorator Decorator} patterns, while this
  * factory itself is an application of the {@linkplain dk.rode.thesis.proxy Proxy}
  * pattern. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class ProxyFactory {

    /**
      * All proxies created by a proxy factory will be tagged
      * with this adapter interface, allowing for later retrieval of
      * the actual proxy {@link #getProxyFactory() factory} that 
      * created the proxy as well as the unique {@link #getHandler()
      * handler} associated with it. <p>
      * 
      * The interface is private, so it will not be accessible
      * outside the proxy factory class.
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private interface Proxy {

       /**
         * Returns the proxy {@link ProxyFactory factory} that created 
         * this proxy.
         *  
         * @return The proxy factory; never null.
         */
       ProxyFactory getProxyFactory();
       
       /**
         * Returns the unique {@link Handler handler} used by this proxy. 
         *  
         * @return The unique handler; never null.
         */
       Handler getHandler();
       
    }
    
    /**
      * A key adapts a {@link Proxy proxy} to ensure consistent {@code hashCode()} 
      * and {@code equals(Object)} behaviour when proxies are stored 
      * internally by proxy factories, because proxies 
      * may override {@code hashCode} and {@code equals(Object)}
      * in unpredictable ways. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private final static class Key {
     
        /**
          * The actual proxy object. <p>
          * 
          * Never null.
          */
        private final Proxy proxy;

        /**
          * Constructor. 
          * 
          * @param proxy The proxy to store; never null.
          */
        private Key(Proxy proxy) {
            this.proxy = proxy;
        }
        
        /**
          * Returns the hash code of this key.
          * 
          * @return The hash code.
          */
        public int hashCode() {
            return this.proxy.getHandler().hashCode();
        }
        
        /**
          * True if {@code object} is a key representing the
          * same exact proxy instance as the proxy represented
          * by this key. <p>
          * 
          * @param  object The object to test; can be null.
          * @return True if equal, false if not.
          */
        public boolean equals(Object object) {
            if (!(object instanceof Key)) {
                return false;
            }
            Key key = (Key)object;
            return (this.proxy.getHandler() == key.proxy.getHandler());
        }
        
        /**
          * Returns the string representation of this key.
          * 
          * @return The string representation; never null.
          */
        public String toString() {
            return new StringBuilder()
                  .append(this.proxy.getHandler())
                  .append(": ").append(ObjectPolicy.ID.toString(this.proxy))
                  .append(" -> ").append(this.proxy).toString();
        }
    }
    
    /**
      * Reference to the {@link Proxy#getProxyFactory} method 
      * defined in the {@link Proxy} interface. <p>
      * 
      * Never null.
      */
    private final static Method factoryMethod;
    /**
      * Reference to the {@link Proxy#getHandler} method 
      * defined in the {@link Proxy} interface. <p>
      * 
      * Never null.
      */
    private final static Method handlerMethod; 
    static {
        try {
            factoryMethod = Proxy.class.getMethod("getProxyFactory");
            handlerMethod = Proxy.class.getMethod("getHandler");
        } catch (NoSuchMethodException nsme) {
            throw new Error(nsme); // will never happen...
        }
    }
    
    /**
      * A <i>handler</i> is used to decorate all invocation handlers 
      * used by proxies created by any proxy factory. <p>
      * 
      * A handler logs access to any method accessed via the 
      * proxy, and ensures correct {@link Proxy} functionality.
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private class Handler implements InvocationHandler {

        /**
          * The decorated handler. <p>
          * 
          * Never null.
          */
        private final InvocationHandler handler;
        
        /**
          * Constructor.
          * 
          * @param  handler The actual invocation handler; cannot be null.
          * @throws NullPointerException If {@code handler} is null.
          */
        private Handler(final InvocationHandler handler) {
            if ((this.handler = handler) == null) {
                throw new NullPointerException("Handler cannot be null");
            }
        }
        
        /**
          * Logs each method invoked. 
          */
        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
            if (handlerMethod.equals(method)) {
                return this; 
                
            } else if (factoryMethod.equals(method)) {
                return ProxyFactory.this; // non-static inner class gives direct access! 
            }
            
            // May be null because of weak references or if not registered...
            Object object = ProxyFactory.this.getObject(proxy);
            CharSequence id = ObjectPolicy.ID.toString(object == null ? this : object); 
            Log out = LogFactory.getLog(InvocationHandler.class).println(id, " -> execute start: ", method);
            long start = System.nanoTime();
            
            // Method combination!
            try {
                return this.handler.invoke(proxy, method, args);
                
            } catch (Throwable t) {
                out.warn(id, " -> execute exception: ").warn(t);
                throw t;
                
            } finally {
                long time = System.nanoTime() - start;
                out.println(id, " -> execute end: ", method, ": ", time, " ns");
            }
        }

        /**
          * Return the string representation of this handler.
          * 
          * @return The string representation; never null.
          */
        public String toString() {
            return ObjectPolicy.ID.asString(this);
        }
    }
    
    /**
      * A map containing the proxies created by this factory. <p>
      * 
      * The actual proxied objects are stored as weak 
      * references, so this factory will not keep a reference to 
      * them. Individual handlers may still keep references to the
      * proxied objects, however. <p>
      * 
      * Never null, but can be empty. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * The keys, i.e. proxies, should be stored as weak references
      * as well, but this is tricky as weak references compare on
      * identity in {@code equals(Object)}. A {@code WeakHashMap}
      * wraps any type of key as a weak reference, and hence creating new
      * {@link Key keys} to look-up existing proxies will fail if 
      * such a map is used! 
      */
    private final Map<Key,WeakReference<Object>> proxies;
    /**
      * A map containing all unique <i>null objects</i> created by this
      * factory. <p>
      * 
      * A null object (value) is a unique object based on identity for its type (key), 
      * which can be used to mask null where null is not allowed. Hence, at most one 
      * instance of a given type will be created by this factory. <p>
      * 
      * Will be null if no such null objects have been created.
      */
    private Map<Class<?>,Object> nulls;
    /**
      * The class loader to use when creating the proxy objects. <p>
      * 
      * Never null.
      */
    private final ClassLoader classLoader;
    
    /**
      * No-arg constructor, which creates this proxy factory to
      * use the class loader that loaded this factory.
      */
    public ProxyFactory() {
        this(ProxyFactory.class.getClassLoader());
    } 
    
    /**
      * Constructor, which creates this factory to use the 
      * class loader supplied as {@code classLoader} when
      * creating new proxy objects.
      * 
      * @param  classLoader The class loader; cannot be null.
      * @throws NullPointerException If {@code classLoader} is null.
      */
    public ProxyFactory(ClassLoader classLoader) {
        if ((this.classLoader = classLoader) == null) {
            throw new NullPointerException("Class loader cannot be null");
        }
        this.proxies = new HashMap<Key,WeakReference<Object>>();
        this.nulls = null;
    }
    
    /**
      * Returns the proxy factory that created the proxy supplied
      * as {@code proxy}, or null if {@code proxy} is not a
      * proxy created by a proxy factory.
      * 
      * @param  proxy The proxy object; cannot be null.
      * @return The proxy factory, or null.
      * @throws NullPointerException If {@code proxy} is null.
      * @throws UnsupportedOperationException If this method
      *         is invoked from a non sub-class context.
      */
    protected final static ProxyFactory getProxyFactory(Object proxy) {
        // Guard against malicious reflection invocation. 
        // An assignable caller can only be sub-class as this
        // factory does not implement any interfaces. Notice
        // that it also works for static methods...
        if (new CallerClass().getAssignableCaller() == null) {
            throw new UnsupportedOperationException();
            
        } else if (proxy instanceof Proxy) { 
            return ((Proxy)proxy).getProxyFactory();
        }
        return null;
    }
        
    /**
      * Registers the proxied object supplied as {@code object}
      * for the proxy supplied as {@code proxy}.
      * 
      * @param  proxy  The proxy; cannot be null.
      * @param  object The proxied object; cannot be null.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code proxy} already
      *         has an associated proxied object. 
      * @throws UnsupportedOperationException If this method
      *         is invoked from a non sub-class context.
      */
    protected void register(Object proxy, Object object) {
        // Guard against malicious reflection invocation. 
        if (new CallerClass().getAssignableCaller() == null) {
            throw new UnsupportedOperationException();
        } else if (proxy == null) {
            throw new NullPointerException("Proxy cannot be null");
        } else if (object == null) {
            throw new NullPointerException("Object cannot be null");
        } else if (this.proxies.containsKey(proxy)) {
            throw new IllegalArgumentException("Proxy already registered");
        } 
        Proxy p = (Proxy)proxy;
        this.proxies.put(new Key(p), new WeakReference<Object>(object));
    }
    
    /**
      * Returns a set containing the interfaces to use for the
      * proxied object supplied as {@code object}. <p>
      * 
      * If {@code interfaces} is not null, the contained interfaces
      * will be used in that order. If null, each interface implemented 
      * by {@code object} will be used, in the order implemented
      * by {@code object}. 
      *  
      * @param  object     The proxied object; cannot be null.
      * @param  interfaces The list of specified interfaces, if any;
      *                    can be empty.
      * @return An ordered set containing the interfaces to use; never null,
      *         but can be empty.
      * @throws NullPointerException If {@code object} is null, or
      *         {@code interfaces} contains a null entry.         
      */
    protected final static Set<Class<?>> getInterfaces(Object object, Class<?>... interfaces) { 
        return (interfaces.length == 0)
               ? Reflection.getInterfaces(object.getClass())
               : add(new LinkedHashSet<Class<?>>(), interfaces);
    }
    
    
    /**
      * Creates a new proxy object implementing the interfaces supplied 
      * as {@code interfaces}, in order. <p>
      * 
      * The proxy object is <b>not</b> registered to any real object.
      * 
      * @param  handler    The invocation handler to use by the
      *                    proxy; cannot be null.
      * @param  interfaces The interfaces to implement by the returned
      *                    proxy; cannot be null, empty, or contain null 
      *                    entries or non-interface classes. 
      * @return A new proxy instance; never null.
      * @throws NullPointerException If either argument is null, or if
      *         {@code interfaces} contains a null entry.
      * @throws IllegalArgumentException If {@code interfaces} contain
      *         a non-interface class.
      * @throws ClassCastException If the returned proxy is cast
      *         to a non-interface class, or an interface not implemented
      *         by the proxy.
      */
    public Object getProxy(final InvocationHandler handler, Class<?>... interfaces) {
        if (interfaces.length == 0) {
            throw new IllegalArgumentException("Interfaces cannot be empty");
        }
        return this.getProxy(null, handler, interfaces);
    }
    
    /**
      * Creates a new proxy object implementing the interfaces supplied 
      * as {@code interfaces}, in order. <p>
      * 
      * The proxy object is registered as a proxy object for the
      * object supplied as {@code object}, if any. Also, if {@code interfaces}
      * is empty, all interfaces implemented by {@code object} 
      * will be used for the returned proxy. <p>
      * 
      * Proxies can be supplied as the {@code object} parameter, which
      * allows for nested proxies. 
      * 
      * @param  <T>        The type of {@code object}.
      * @param  object     The object to proxy; can be null.
      * @param  handler    The invocation handler to use by the
      *                    proxy; cannot be null.
      * @param  interfaces The interfaces to implement by the returned
      *                    proxy; cannot contain null entries
      *                    or non-interface classes. 
      * @return A new proxy instance; never null.
      * @throws NullPointerException If {@code handler} or {@code interfaces} are null, 
      *         or if {@code interfaces} contains a null entry.
      * @throws IllegalArgumentException If {@code interfaces} contain
      *         a non-interface class.
      * @throws ClassCastException If the returned proxy is cast
      *         to a non-interface class, or an interface not implemented
      *         by the proxy.
      */
    public <T> T getProxy(final T object, final InvocationHandler handler, Class<?>... interfaces) {
        return this.create(object, getInterfaces(object, interfaces), handler);
    }
    
    /**
      * Factory method used to create a new proxy object implementing
      * the interfaces supplied as {@code interfaces}, in order,
      * based on the object to proxy supplied as {@code object}, if any. <p>
      * 
      * If {@code object} is not null, the returned proxy will be registered 
      * with {@code object} as the associated proxied object using the 
      * {@link #register(Object, Object)} method.
      * 
      * @param  <T>        The type of {@code object}.
      * @param  object     The object to proxy; can be null.
      * @param  interfaces The interfaces to implement by the returned
      *                    proxy; cannot be null, or contain null entries
      *                    or non-interface classes. Annotations are ignored.
      * @param  handler    The invocation handler to use by the
      *                    proxy; cannot be null.
      * @return A new proxy instance; never null.
      * @throws NullPointerException If {@code interfaces} or {@code handler} are null, 
      *         or if {@code interfaces} contain a null entry.
      * @throws IllegalArgumentException If {@code interfaces} contain
      *         a non-interface class.
      */
    private <T> T create(T object, Set<Class<?>> interfaces, InvocationHandler handler) {
        // Adapt proxy with Proxy class so we can cast to it. If already assignable
        // to it, we are about to create a proxy to proxy...
        if (!interfaces.contains(Proxy.class)) {
            interfaces.add(Proxy.class);
        }
        Class<?>[] classes = new Class<?>[interfaces.size()];
        interfaces.toArray(classes);
        // Cannot use class.cast(..) here since the actual type 
        // may not be an interface, so we have to cast specifically 
        // and suppress the warnings...
        @SuppressWarnings("unchecked") T proxy = 
            (T)java.lang.reflect.Proxy.newProxyInstance(this.classLoader, 
                                                        classes, 
                                                        new Handler(handler));
        // Register, if possible..
        if (object != null) {
            this.register(proxy, object);
        }
        return proxy;
    }
    
    /**
      * Returns true if {@code object} is a proxy created by this
      * proxy factory, false if not.
      * 
      * @param  object The object to test; cannot be null.
      * @return True if a proxy created by this factory, false if 
      *         not.
      * @throws NullPointerException If {@code object} is null.         
      */
    public boolean isProxy(Object object) {
        if (object == null) {
            throw new NullPointerException("Object cannot be null");
        } else if (!(object instanceof Proxy)) {
            return false;
        }
        Proxy p = (Proxy)object;
        return this.proxies.containsKey(new Key(p));
    }
    
    /**
      * Returns the actual object proxied by {@code proxy}, or null
      * if no such proxy is handled by this proxy factory or if 
      * garbage collected.
      * 
      * @param  proxy The proxy; cannot be null.
      * @return The actual object, or null if garbage collected.
      * @throws NullPointerException If {@code proxy} is null. 
      * @throws IllegalArgumentException If {@code proxy} does not 
      *         represent a proxy. 
      */
    public Object getObject(Object proxy) {
        if (!(proxy instanceof Proxy)) {
            throw new IllegalArgumentException("Object is not a proxy: " + proxy);
        }
        Proxy p = (Proxy)proxy;
        WeakReference<Object> reference = this.proxies.get(new Key(p));
        return reference == null ? null : reference.get();
    }

    /**
      * Returns true if the object supplied as {@code object} 
      * is proxied by a proxy created by this factory, false if not.
      *  
      * @param  object The object to test; cannot be null.
      * @return True if proxied, false if not.
      * @throws NullPointerException If {@code object} is null.
      */
    public boolean isProxied(Object object) {
        if (object == null) {
            throw new NullPointerException("Object cannot be null");
        }
        for (WeakReference<?> reference : this.proxies.values()) {
            if ((reference != null) && (reference.get() == object)) {
                return true;
            }
        }
        return false;
    }
    
    /**
      * Returns all proxy objects for the object supplied
      * as {@code object}, if any.
      * 
      * @param  object The object; cannot be null.
      * @return A list containing the created proxies in no 
      *         particular order; never null, but can be empty.
      * @throws NullPointerException If {@code object} is null.         
      */
    public List<?> getProxies(Object object) {
        if (object == null) {
            throw new NullPointerException("Object cannot be null");
        }
        List<Object> proxies = new ArrayList<Object>();
        for (Entry<Key,WeakReference<Object>> entry : this.proxies.entrySet()) {
            WeakReference<Object> reference = entry.getValue();
            if ((reference != null) && (reference.get() == object)) {
                proxies.add(entry.getKey().proxy);
            }
        }
        return proxies;
    }
    
    /**
      * Returns a synchronised (proxy) version of {@code object} that can
      * be viewed as any of the interfaces supplied in {@code interfaces}. The
      * synchronisation lock used is {@code object} itself. <p>
      *  
      * All access to {@code object} via the methods defined in one of the 
      * types in {@code interfaces} will be synchronised on {@code object} 
      * itself <b>if and only if</b> {@code object} is accessed through the 
      * returned proxy object <b>only</b>. If {@code object} is accessed 
      * directly, synchronisation cannot be guaranteed! <p>
      * 
      * If no interfaces are supplied, all interfaces implemented by 
      * {@code object}, directly or indirectly (for example via a 
      * super-class), will be used. <p>
      * 
      * Since the returned object is a {@link java.lang.reflect.Proxy
      * java.lang.reflect.Proxy} instance, the normal rules for {@code Proxy} objects apply! An
      * implication of this is that even if a specific instance is supplied
      * as {@code object}, i.e. a specific class, the returned (proxy) 
      * object <b>cannot</b> be cast back into that type (the specific 
      * type information is "lost" as the class returned is an anonymous 
      * {@code Proxy} class). <p>
      * 
      * <b>Example</b>:
      * <pre>
      *   // Interfaces...
      *   public interface Foo {
      *     public String foo();
      *   }
      *
      *   public interface Bar {
      *     public String bar();
      *   }
      *
      *   ..
      *   
      *   // Implementation...
      *   public class FooBar implements Foo, Bar {
      *     public String foo() {
      *       return "foo";
      *     }     
      *     public String bar() {
      *       return "bar";
      *     }     
      *   }
      *   
      *   ..
      *   
      *   FooBar foobar = new FooBar();
      *   
      *   {@link ProxyFactory} factory = new {@link #ProxyFactory()};
      *   
      *   // No class casting needed, since we only use a single interface...
      *   Foo foo = factory.{@link #getSynchronisedObject(Object, Class...) getSynchronisedObject}(foobar, Foo.class);
      *          
      *   // All access to the methods defined in Foo will now be 
      *   // synchronised on "foobar" when accessing through "foo":
      *   
      *   String s = foobar.foo() // *not* synchronised
      *          s = foo.foo();   // synchronised
      *   
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)foo;
      *   
      *   ..
      *   
      *   // A class cast is now required, since more than one interface
      *   // is implemented by the returned proxy. A cast to either type
      *   // is valid...
      *   foo = (Foo)factory.{@link #getSynchronisedObject(Object, Class...) getSynchronisedObject}(foobar, Foo.class, Bar.class);
      *          
      *   // All access to the methods defined in Foo *and* Bar will now
      *   // be synchronised on "foobar" when accessing through "foo" and "bar":
      *   
      *   String s = foobar.foo(); // *not* synchronised
      *          s = foobar.bar(); // *not* synchronised
      *          s = foo.foo();    // synchronised
      *          
      *   // Must cast to access "bar"...          
      *   Bar bar = (Bar)foo;       
      *          s = bar.bar();    // synchronised 
      *                           
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)bar;
      *   
      *   // Default - all interfaces implemented by FooBar (including 
      *   // those from super-classes) - no class casting is required, 
      *   // but the returned proxy can simply be assigned to a proper type:
      *   
      *   foo = factory.{@link #getSynchronisedObject(Object, Class...) getSynchronisedObject}(foobar); // Foo.class & Bar.class
      *   bar = factory.{@link #getSynchronisedObject(Object, Class...) getSynchronisedObject}(foobar); // Foo.class & Bar.class
      * </pre>
      *  
      * Proxies can be supplied as the {@code object} parameter, which
      * allows for nested synchronisation. 
      *
      * @param  <T>        The type of {@code object}.
      * @param  object     The actual object to be used; cannot be null, and
      *                    must implement all the interfaces supplied
      *                    as {@code interfaces}.
      * @param  interfaces The interfaces to supply the methods to be 
      *                    synchronised; can be null, in which case
      *                    all interfaces implemented by the class of {@code object}
      *                    will be used.
      * @return A proxy object, synchronised on all methods defined in
      *         {@code interfaces}; never null.
      * @throws NullPointerException If {@code object} is null.
      * @throws IllegalArgumentException If {@code interfaces} contain
      *         an non-interface type.
      * @see    #getSynchronisedObject(Object, Object, Class...)
      * @see    #getLoggableObject(Object, Class...)
      * @see    #getSharedObject(Object, String, Class[], String...)         
      */
    public <T> T getSynchronisedObject(T object, Class<? super T>... interfaces) {
        return this.getSynchronisedObject(object, object, interfaces);
    }
 
    /**
      * Returns a synchronised (proxy) version of {@code object} that can
      * be viewed as any of the interfaces supplied in {@code interfaces},
      * synchronised on the object supplied as {@code lock}. <p>
      * 
      * All access to {@code object} via the methods defined in one of the 
      * types in {@code interfaces} will be synchronised on {@code lock} 
      * <b>if and only if</b> {@code object} is accessed through the 
      * returned proxy object <b>only</b>. If {@code object} is accessed 
      * directly, synchronisation cannot be guaranteed! <p>
      * 
      * If no interfaces are supplied, all interfaces implemented by 
      * {@code object}, directly or indirectly (for example via a 
      * super-class), will be used. <p>
      * 
      * Since the returned object is a {@link java.lang.reflect.Proxy
      * java.lang.reflect.Proxy} instance, the normal rules for {@code Proxy} objects apply! An
      * implication of this is that even if a specific instance is supplied
      * as {@code object}, the returned (proxy) object <b>cannot</b> be
      * cast back into that type (the specific type information is "lost" as
      * the class returned is an anonymous {@code Proxy} class). <p>
      * 
      * <b>Example</b>:
      * <pre>
      *   // Interfaces...
      *   public interface Foo {
      *     public String foo();
      *   }
      * 
      *   public interface Bar {
      *     public String bar();
      *   }
      *
      *   ..
      *   
      *   // Implementation...
      *   public class FooBar implements Foo, Bar {
      *     public String foo() {
      *       return "foo";
      *     }     
      *      public String bar() {
      *       return "bar";
      *     }     
      *   }
      *   
      *   ..
      *   
      *   Object lock = new Object();
      *   FooBar foobar = new FooBar();
      *   
      *   {@link ProxyFactory} factory = new {@link #ProxyFactory()};
      *   
      *   // No class casting needed, since we only use a single interface...
      *   Foo foo = factory.{@link #getSynchronisedObject(Object, Object, Class...) getSynchronisedObject}(foobar, lock, Foo.class);
      *           
      *   // All access to the methods defined in Foo will now be 
      *   // synchronised on "lock" when accessing through "foo":
      *   
      *   String s = foobar.foo() // *not* synchronised
      *          s = foo.foo();   // synchronised
      *   
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)foo;
      *   
      *   ..
      *   
      *   // A class cast is now required, since more than one interface
      *   // is implemented by the returned proxy. A cast to either type
      *   // is valid...
      *   foo = (Foo)factory.{@link #getSynchronisedObject(Object, Object, Class...) getSynchronisedObject}(foobar, lock, Foo.class, Bar.class);
      *          
      *   // All access to the methods defined in Foo *and* Bar will now
      *   // be synchronised on "lock" when accessing through "foo" and "bar":
      *   
      *   String s = foobar.foo(); // *not* synchronised
      *          s = foobar.bar(); // *not* synchronised
      *          s = foo.foo();    // synchronised
      *          
      *   // Must cast to access "bar"...          
      *   Bar bar = (Bar)foo;       
      *          s = bar.bar();    // synchronised 
      *                           
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)bar;
      *   
      *   // Default - all interfaces implemented by FooBar (including 
      *   // those from super-classes) - no class casting is required, 
      *   // but the returned proxy can simply be assigned to a proper type:
      *   
      *   foo = factory.{@link #getSynchronisedObject(Object, Object, Class...) getSynchronisedObject}(foobar, lock); // Foo.class & Bar.class
      *   bar = factory.{@link #getSynchronisedObject(Object, Object, Class...) getSynchronisedObject}(foobar, lock); // Foo.class & Bar.class
      * </pre>
      *  
      * Proxies can be supplied as the {@code object} parameter, which
      * allows nested synchronisation. 
      *  
      * @param  <T>        The type of {@code object}.
      * @param  object     The actual object to be used; cannot be null, and
      *                    must implement all the interfaces supplied
      *                    as {@code interfaces}.
      * @param  lock       The object to synchronise on; cannot be null.
      * @param  interfaces The interfaces to supply the methods to be 
      *                    synchronised; can be null, in which case
      *                    all interfaces implemented by the class of {@code object}
      *                    will be used.
      * @return A proxy object, synchronised on all methods defined in
      *         {@code interfaces}; never null.
      * @throws NullPointerException If {@code object} or {@code lock}
      *         are null.
      * @throws IllegalArgumentException If {@code interfaces} contain an 
      *         non-interface type.         
      * @see    #getSynchronisedObject(Object, Object, Class...)  
      * @see    #getLoggableObject(Object, Class...)
      * @see    #getSharedObject(Object, String, Class[], String...)
      */
    public <T> T getSynchronisedObject(final T object, final Object lock, Class<? super T>... interfaces) {
        if (lock == null) {
            throw new NullPointerException("Lock cannot be null");
        }
        InvocationHandler handler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                synchronized (lock) {
                    LogFactory.getLog(InvocationHandler.class).println("Synchronised access to: ", ObjectPolicy.ID.toString(proxy), " -> ", ObjectPolicy.ID.toString(object));
                    return Reflection.invoke(object, method, args);
                }
            }
        };
        return this.getProxy(object, handler, interfaces);
    }
    
    /**
      * Returns a (proxy) version of {@code object} that will record
      * all access to {@code object} when accessed (via the proxy) through
      * one of the interfaces supplied in {@code interfaces}. The access
      * is recorded to the {@link Log log} associated with the <i>original</i>
      * class of {@code object}. <p>
      *  
      * All access to {@code object} via the methods defined in one of the 
      * types in {@code interfaces} will be logged <b>if and only if</b> 
      * {@code object} is accessed through the returned proxy object 
      * <b>only</b>. If {@code object} is accessed directly, logging cannot 
      * be guaranteed! <p>
      * 
      * If no interfaces are supplied, all interfaces implemented by 
      * {@code object}, directly or indirectly (for example via a 
      * super-class), will be used. <p>
      * 
      * Since the returned object is a {@link java.lang.reflect.Proxy
      * java.lang.reflect.Proxy} instance, the normal rules for {@code Proxy} objects apply! An
      * implication of this is that even if a specific instance is supplied
      * as {@code object}, i.e. a specific class, the returned (proxy) 
      * object <b>cannot</b> be cast back into that type (the specific 
      * type information is "lost"). <p>
      * 
      * <b>Example</b>:
      * <pre>
      *   // Interfaces...
      *   public interface Foo {
      *     public String foo();
      *   }
      *
      *   public interface Bar {
      *     public String bar();
      *   }
      *
      *   ..
      *   
      *   // Implementation...
      *   public class FooBar implements Foo, Bar {
      *     public String foo() {
      *       return "foo";
      *     }     
      *     public String bar() {
      *       return "bar";
      *     }     
      *   }
      *   
      *   ..
      *   
      *   FooBar foobar = new FooBar();
      *   
      *   {@link ProxyFactory} factory = new {@link #ProxyFactory()};
      *   
      *   // No class casting needed, since we only use a single interface...
      *   Foo foo = factory.{@link #getLoggableObject(Object, Class...) getLoggableObject}(foobar, Foo.class);
      *          
      *   // All access to the methods defined in Foo will now be 
      *   // logged when accessing through "foo":
      *   
      *   String s = foobar.foo() // *not* logged
      *          s = foo.foo();   // logged
      *   
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)foo;
      *   
      *   ..
      *   
      *   // A class cast is now required, since more than one interface
      *   // is implemented by the returned proxy. A cast to either type
      *   // is valid...
      *   foo = (Foo)factory.{@link #getLoggableObject(Object, Class...) getLoggableObject}(foobar, Foo.class, Bar.class);
      *          
      *   // All access to the methods defined in Foo *and* Bar will now
      *   // be logged accessing through "foo" and "bar":
      *   
      *   String s = foobar.foo(); // *not* logged
      *          s = foobar.bar(); // *not* logged
      *          s = foo.foo();    // logged
      *          
      *   // Must cast to access "bar"...          
      *   Bar bar = (Bar)foo;       
      *          s = bar.bar();    // logged 
      *                           
      *   // Will throw a class cast exception! 
      *   FooBar goo = (FooBar)bar;
      *   
      *   // Default - all interfaces implemented by FooBar (including 
      *   // those from super-classes) - no class casting is required, 
      *   // but the returned proxy can simply be assigned to a proper type:
      *   
      *   foo = factory.{@link #getLoggableObject(Object, Class...) getLoggableObject}(foobar); // Foo.class & Bar.class
      *   bar = factory.{@link #getLoggableObject(Object, Class...) getLoggableObject}(foobar); // Foo.class & Bar.class
      * </pre>
      *  
      * Proxies can be supplied as the {@code object} parameter, which
      * allows for nested logging. 
      *
      * @param  <T>        The type of {@code object}.
      * @param  object     The actual object to be used; cannot be null, and
      *                    must implement all the interfaces supplied
      *                    as {@code interfaces}.
      * @param  interfaces The interfaces to supply the methods to be 
      *                    logged; can be null, in which case all interfaces 
      *                    implemented by the class of {@code object} will be used.
      * @return A proxy object, where all access on all methods defined in
      *         {@code interfaces} will be logged; never null.
      * @throws NullPointerException If {@code object} is null.
      * @throws IllegalArgumentException If {@code interfaces} contain
      *         an non-interface type.
      * @see    #getSynchronisedObject(Object, Class...)         
      * @see    #getSynchronisedObject(Object, Object, Class...)
      * @see    #getSharedObject(Object, String, Class[], String...)
      */
    public <T> T getLoggableObject(final T object, Class<? super T>... interfaces) {
        InvocationHandler handler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                return Reflection.invoke(object, method, args);
            }
        };
        return this.getProxy(object, handler, interfaces);
    }
    
    /**
      * Returns a (proxy) {@link Reference reference} to {@code object} 
      * that can be shared and <i>copied</i> by different contexts until a 
      * given <i>mutator</i> method is invoked on it, forcing the original 
      * object to be copied. Both <i>copy</i> and <i>mutator methods</i> 
      * are supplied as arguments to this method. <p>
      * 
      * All access to {@code object} is via the interfaces the class of it
      * implements. All access to {@code object} should be though the
      * returned proxy. <p>
      * 
      * Since the returned object is a {@link java.lang.reflect.Proxy
      * java.lang.reflect.Proxy} instance, the normal rules for {@code Proxy} objects apply! An
      * implication of this is that even if a specific instance is supplied
      * as {@code object}, i.e. a specific class, the returned (proxy) 
      * object <b>cannot</b> be cast back into that type (the specific 
      * type information is "lost"). <p>
      * 
      * <b>Example</b>:
      * <pre>
      *   // Interface of objects to be shared...
      *   public interface Foo {
      *     // Normal methods that can be shared...
      *     public int getFooCount();
      *     
      *     public String toString();
      *     ..
      *     
      *     // Mutator methods, which when invoked may cause a copy...
      *     public void setFooCount(int count);
      *     
      *     public void resetFooCount();
      *     .. 
      *     
      *     // Copy method to perform the actual copying...
      *     public Foo copy(boolean deep);
      *   }
      *
      *   // Implementation...
      *   public class FooBar implements Foo {
      *     ..
      *   }
      *
      *   ..
      *   
      *   // Specify copy method...
      *   String copyMethod = "copy";
      *   // "boolean" formal parameter (primitive types are ignored)...
      *   Class&lt;?> parameterTypes = {Boolean.class}; // or Boolean.TYPE
      *   // Mutator methods, based on name only...
      *   String[] mutators = {"getFooCount", "resetFooCount"};
      *      
      *   // "foo" is the instance to be shared (originally)...
      *   Foo foo = new FooBar();
      *   
      *   {@link ProxyFactory} factory = new {@link #ProxyFactory()};
      *   
      *   // No class casting needed...
      *   Foo foobar = factory.{@link #getSharedObject(Object, String, Class[], String...) getSharedObject}(foo, copyMethod, parameterTypes, mutatorMethods); // Can also use varargs for mutators
      *   
      *   // Create a copy using the normal copy method. This will increase the 
      *   // reference count to "foo"...
      *   Foo foofoo = foobar.copy(true);
      *   
      *   // Access to non-mutator methods will not cause a copy...
      *   System.out.println(foobar.getFooCount());
      *   System.out.println(foofoo); // toString()
      *   ..
      *   
      *   // Access to a mutator method in any copy will force "foo" to
      *   // be copied because more than one reference. The value 42 will
      *   // be set in a copy of "foo" used by "foofoo", not in other
      *   // references (here, "foobar")...
      *   foofoo.setFooCount(42);
      *   
      *   // Only one reference, so if "foobar" accesses mutator methods,
      *   // no copy will be made (until copied again)...
      *   foobar.resetFooCount();
      *   .. 
      * </pre>
      *  
      * @param  <T>            The type of {@code object}.
      * @param  object         The actual object to be shared; cannot be null.
      * @param  copyMethod     The name of the <i>copy</i> method to use; cannot 
      *                        be null. Must be declared in the class of 
      *                        {@code object}, but will be made accessible 
      *                        if not already. 
      * @param  parameterTypes The formal parameter types of {@code copyMethod},
      *                        if any, ignoring primitive types; can be null.                        
      * @param  mutatorMethods The names of the <i>mutator</i> methods that will 
      *                        force a copy of {@code object} before executed;
      *                        cannot be null or empty.
      * @return A proxy representing the shared object; never null.
      * @throws NullPointerException If {@code object} or {@code copyMethod} are null.
      * @throws IllegalArgumentException If {@code mutatorMethods} is empty, or
      *         if the copy method is a vararg method.
      * @throws Exception If no such copy method exists.
      * @see    Reference        
      * @see    #getSynchronisedObject(Object, Class...)         
      * @see    #getSynchronisedObject(Object, Object, Class...)
      * @see    #getLoggableObject(Object, Class...)
      */
    public <T> T getSharedObject(final T object, String copyMethod, Class<?>[] parameterTypes, String... mutatorMethods) throws Exception {
        return new ReferenceHandler<T>(this, object, copyMethod, parameterTypes, mutatorMethods).toProxy();
    }
    
    /**
      * Returns a to this factory unique <i>null object</i> of the type supplied 
      * as {@code type}. <p>
      * 
      * A null object is a unique object based on identity for its type, which can 
      * be used to mask null where null is not allowed. Hence, at most one instance of 
      * {@code type} will be created by this factory. <p>
      * 
      * Properties for returned null objects: <ul><p>
      * 
      * <li> If {@code type} is {@code java.lang.Object}, the returned null object
      *      is {@link Const#NULL}. <p>
      *      
      * <li> If {@code type} is an interface or abstract class, a unique proxy is 
      *      returned where all methods declaring a non-void return type will return null, 
      *      except {@code java.lang.Object} methods and methods having a primitive return 
      *      type; the latter will return the default primitive value of the return type 
      *      in question. <p> 
      * 
      * <li> If {@code type} is a primitive type or a primitive wrapper class, an 
      *      unique instance of the wrapper class is used with the default primitive 
      *      value. <p>
      *      
      * <li> If {@code type} is a neither a primitive type nor a primitive wrapper 
      *      class, a unique instance of the type is used, created using a public
      *      no-arg constructor. <p> 
      * 
      * </ul>
      * 
      * @param  <T>  The type of the null object.
      * @param  type The type as a class literal; cannot be null.
      * @return The unique null object of the type supplied as {@code type}; never null.
      * @throws Exception If {@code type} represents a class, which is neither a primitive 
      *         type nor a primitive wrapper class, without a public no-arg constructor; 
      *         or if {@code type} is an abstract class.
      */
    public <T> T getNullObject(final Class<T> type) throws Exception {
        if (type == null) {
            throw new NullPointerException("Type cannot be null");
        } else if (Object.class.equals(type)) {
            return type.cast(Const.NULL);
        }
        synchronized (this) {
            if (this.nulls == null) {
                this.nulls = new IdentityHashMap<Class<?>,Object>();
            }
            // Check primitive...
            Primitive p = Primitive.getPrimitive(type);
            if (p != null) {
                Object NULL = this.nulls.get(p.nonPrimitive);
                if (NULL == null) {
                    this.nulls.put(p.nonPrimitive, NULL = p.getDefaultValue());
                }
                // Cannot use type.cast on primitive types!
                @SuppressWarnings("unchecked") T t = (T)NULL;
                return t;
                
            } else {
                Object NULL = this.nulls.get(type);
                if (NULL == null) {
                    // If interface or abstract class, we need to create a proxy...
                    if ((type.isInterface()) || (Modifier.isAbstract(type.getModifiers()))) {
                        // Target for all dispatched Object methods...
                        final Object target = new Object();
                        NULL = this.getProxy(new InvocationHandler() {
                            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                                // Allow all Object methods, such as wait, notify, hashCode, toString, equals, etc...
                                if (method.getDeclaringClass().equals(Object.class)) { // ISSUE: test for methods overridden/re-defined by a sub-class!
                                    String name = method.getName();
                                    // Intercept toString to pretty print the object...
                                    if (name.equals("toString")) {                                     
                                        return new StringBuilder().append(type.getSimpleName()).append("[NULL]").toString();
                                    }
                                    return Reflection.invoke(target, method, args); // ISSUE: test for clone and serializable methods
                                } 
                                return Primitive.getDefaultValue(method.getReturnType());
                            }
                        }, type);
                        
                    // If a (non-primitive) class, we cannot create a proxy and must create a
                    // new instance...
                    } else {
                        NULL = type.newInstance();
                    }
                    this.nulls.put(type, NULL);
                }
                return type.cast(NULL);
            }
        }
    }
    
    /**
      * Returns the hash code of this factory.
      * 
      * @return The hash code.
      */
    public final int hashCode() {
        return System.identityHashCode(this);
    }

    /**
      * Returns true if {@code object} is the same
      * factory instance as this factory, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if same instance, false if not.
      */
    public final boolean equals(Object object) {
        return (this == object);
    }

    /**
      * Return the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; proxies: ").append(this.proxies.size()).append(']').toString();
    }
    
}
