/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
  * An <i>arguments</i> object represents a collection of String 
  * arguments that may or may not have an associated String value. <p>
  * 
  * Argument names are expected to be prefixed with a dash (-), 
  * and are looked up in a case-insensitive manner. <p>
  * 
  * Example:
  * <pre>
  *   -foo -bar barbar -goo goo goo goo -foobar
  * </pre>
  * 
  * Here, four argument names are present: {@code foo},
  * {@code bar}, {@code goo}, and {@code foobar}. {@code foo}
  * and {@code foobar} have no associated values, while
  * {@code bar} has the value {@code barbar} associated,
  * and {@code goo} has {@code goo goo goo}. <p>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Arguments implements Iterable<Arguments.Argument> {

    /**
      * An <i>argument</i> represents a single
      * argument, with or without an associated value.
      *  
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public final static class Argument {

        /**
          * The argument name (excluding prefixed -). <p>
          * 
          * Never null.
          */
        public final String name;
        /**
          * The associated value. <p>
          * 
          * Can be null.
          */
        public final String value;
        
        /**
          * Constructor.
          * 
          * @param name  The argument name, excluding prefixed -; never null.
          * @param value The associated value; can be null.
          */
        Argument(String name, String value) {
            this.name  = name;
            this.value = value;
        }

        /**
          * Returns the hash code of this argument.
          * 
          * @return The hash code.
          */
        public final int hashCode() {
            return this.name.hashCode();
        }

        /**
          * Returns true if {@code object} is an argument with
          * the same name as this argument, regardless of value.
          *  
          * @param  object The object to test; can be null.
          * @return True if equal, false if not.
          */
        public final boolean equals(Object object) {
            if (!(object instanceof Argument)) {
                return false;
            }
            return this.name.equals(((Argument)object).name);
        }
        
        /**
          * Returns the string representation of this argument.
          *
          * @return The string representation; never null.
          */
        public String toString() {
            return "-" + this.name + (this.value == null ? "" : " " + this.value);
        }
        
    }
    
    /**
      * Actual arguments stored according to insertion order. <p>
      * 
      * All names contained as keys do not have a prefixed dash (-). <p>
      * 
      * Never null, but can be empty. 
      */
    public Map<String,Argument> arguments;
    
    /**
      * Constructor. <p>
      * 
      * The arguments are parsed as described in the class documentation. 
      * 
      * @param args The actual arguments to parse; can be null.
      * @see   #parse(String...)
      */
    public Arguments(String... args) {
        this.arguments = parse(args);
    }

    /**
      * Copy constructor.
      * 
      * @param  arguments The arguments to copy; cannot be null.
      * @throws NullPointerException If {@code arguments} is null.
      */
    public Arguments(Arguments arguments) {
        this.arguments = new LinkedHashMap<String,Argument>(arguments.arguments);
    }
    
    /**
      * Returns the argument having the name supplied as {@code name}, 
      * regardless if it has an associated value or not. <p>
      * 
      * The comparison on argument name is performed in a 
      * case-insensitive manner.
      *  
      * @param  name The argument name; cannot be null, and may
      *              start with a dash.
      * @return The corresponding argument, or null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public Argument getArgument(String name) {
        return this.getArgument(name, false);
    }
    
    /**
      * Returns the argument having the name supplied as {@code name}, 
      * regardless if it has an associated value or not if 
      * {@code mustHaveValue} is false. If true, an argument is only
      * returned if it has an associated value. <p>
      * 
      * The comparison on argument name is performed in a 
      * case-insensitive manner.
      *  
      * @param  name          The argument name; cannot be null, and may
      *                       start with a dash.
      * @param  mustHaveValue True if an argument should only be returned 
      *                       if it has an associated value.              
      * @return The corresponding argument, or null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public Argument getArgument(String name, boolean mustHaveValue) {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        Argument argument = this.arguments.get(name = getName(name));
        if ((argument == null) || ((mustHaveValue) && (argument.value == null))) {
            return null;
        }
        return argument;
    }

    /**
      * Returns true if an argument has the name supplied as {@code name},
      * regardless if it has an associated value or not.
      *  
      * @param  name          The argument name; cannot be null, and may
      *                       start with a dash.
      * @return True if such an argument exist, false if not.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public boolean hasArgument(String name) {
        return (this.getArgument(name) != null);
    }
    
    /**
      * Returns the argument value associated with the argument
      * name supplied as {@code name}, if any. <p>
      * 
      * @param  name          The argument name; cannot be null, and may
      *                       start with a dash.
      * @return The argument value, or null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public String getArgumentValue(String name) {
        return this.getArgumentValue(name, (String)null);
    }

    /**
      * Returns the argument value associated with the argument
      * name supplied as {@code name}, if any. <p>
      * 
      * {@code defaultValue} is returned if no such argument
      * is found or it does not have an associated value. <p>
      * 
      * @param  name         The argument name; cannot be null, and may
      *                      start with a dash.
      * @param  defaultValue The default value; can be null.                       
      * @return The argument value, or {@code defaultValue} (which can
      *         be null).
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public String getArgumentValue(String name, String defaultValue) {
        Argument argument = this.getArgument(name, true);
        if (argument == null) {
            return defaultValue;
        } 
        return argument.value;
    }
    
    /**
      * Returns the argument value associated with the argument
      * name supplied as {@code name}, if any, transformed to
      * the type supplied as {@code type} using the static
      * {@code type.valueOf(String)} method. <p>
      * 
      * If no such method is found in {@code type}, null is returned. <p>
      * 
      * If the transformation fails, null is returned. 
      * 
      * @param  <V>  The type to transform the argument value into
      *              using its static {@code valueOf(String)} method.
      * @param  name The argument name; cannot be null, and may
      *              start with a dash.
      * @param  type The class of the desired type that the 
      *              argument value should be transformed 
      *              to by {@code type.valueOf(String)}; 
      *              cannot be null. 
      * @return The argument value as a {@code V} type; can be null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal. <p>
      */
    public <V> V getArgumentValue(String name, Class<V> type) {
        return this.getArgumentValue(name, null, type);
    }
    
    /**
      * Returns the argument value associated with the argument
      * name supplied as {@code name}, if any, transformed to
      * the type supplied as {@code type} using the static
      * {@code type.valueOf(String)} method. <p>
      * 
      * If no such method is found in {@code type}, null is returned. <p>
      * 
      * If no argument value is found or the transformation
      * fails, {@code defaultValue} is returned. <p>
      *  
      * @param <V>           The type to transform the argument value into
      *                      using its static {@code valueOf(String)} method.
      * @param  name         The argument name; cannot be null, and may
      *                      start with a dash.
      * @param  defaultValue The default value; can be null.                       
      * @param  type         The class of the desired type that the 
      *                      argument value should be transformed 
      *                      to by {@code type.valueOf(String)}; 
      *                      cannot be null. 
      * @return The argument value as a {@code V} type; can be null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal. <p>
      */
    public <V> V getArgumentValue(String name, V defaultValue, Class<V> type) {
        String string = this.getArgumentValue(name);
        if (string == null) {
            return defaultValue;
        } 
        try {
            Method method = type.getDeclaredMethod("valueOf", String.class);
            return type.cast(method.invoke(null, string));
            
        } catch (Exception e) {
            return defaultValue;
        }
    }
    
    /**
      * Updates or adds the argument having the name supplied
      * as {@code name} with the associated value supplied as 
      * {@code value}, if any. <p>
      * 
      * If an argument with that name is already contained, it
      * will be updated; if not, a new argument will be added.
      *  
      * @param  name  The argument name; cannot be null, and may
      *               start with a dash.
      * @param  value The associated argument value; can be null.              
      * @return True if an argument with the name {@code name} already
      *         exists, false if not.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    public boolean setArgument(String name, String value) {
        if (name == null) {
            throw new NullPointerException("Name cannot be null");
        }
        return (this.arguments.put(name = getName(name), new Argument(name, value)) != null);
    }

    /**
      * Returns a valid argument name based on {@code name}.
      * 
      * @param  name The name to make valid; cannot be null.
      * @return A valid argument name; never null.
      * @throws NullPointerException If {@code name} is null.
      * @throws IllegalArgumentException If {@code name} is empty or
      *         illegal.
      */
    private final static String getName(String name) {
        name = name.trim();
        if (name.length() == 0) {
            throw new IllegalArgumentException("Name cannot be empty");
        } else if ((name.length() == 1) && (name.charAt(0) == '-')) {
            throw new IllegalArgumentException("Illegal name format: " + name);
        }
        // Remove - prefix, if any...
        if (name.charAt(0) == '-') {
            name = name.substring(1); 
        }
        return name.toLowerCase();
    }
    
    /**
      * Returns an iterator that is backed by these arguments. <p>
      * 
      * @return An iterator; never null.
      */
    public Iterator<Argument> iterator() {
        return this.arguments.values().iterator();
    }
    
    /**
      * Returns a list of arguments having a name that starts 
      * with the prefix supplied as {@code prefix}, if any. <p>
      * 
      * If {@code prefix} is null, all arguments are returned. <p>
      * 
      * Modifying the returned list will not affect these arguments.
      *   
      * @param  prefix The argument prefix; cannot be null, and may
      *                start with a dash.
      * @return A list of matching arguments; never null, but can be 
      *         empty.
      * @throws IllegalArgumentException If {@code prefix} is empty or
      *         illegal.
      */
    public List<Argument> getArguments(String prefix) {
        if (prefix == null) {
            return new ArrayList<Argument>(this.arguments.values());
        } else {
            prefix = getName(prefix);
            List<Argument> arguments = new ArrayList<Argument>(this.size());
            for (String name : this.arguments.keySet()) {
                if (name.startsWith(prefix)) {
                    arguments.add(this.arguments.get(name));
                }
            }
            return arguments;
        }
    }
    
    /**
      * Parses the arguments supplied in {@code args}, and returns
      * a corresponding map, ordered according to the order of
      * {@code args}. <p>
      * 
      * Argument names are expected to be prefixed with a single dash.
      * Everything else is considered an argument value.
      * 
      * @param  args The arguments to parse; can be null or empty.
      * @return A map containing the arguments; never null, but can be empty.
      */
    public final static Map<String,Argument> parse(String... args) {
        Map<String,Argument> arguments = new LinkedHashMap<String,Argument>();
        if ((args == null) || (args.length == 0)) {
            return arguments;
        }
        String name = null, value = null;
        for (String arg : args) {
            if (arg.charAt(0) == '-') {
                if (name != null) {
                    arguments.put(name, new Argument(name, value));
                    value = null;
                }
                name = arg.substring(1).trim().toLowerCase();
                
            } else if (value == null) {
                value = arg.toString();
            } else {
                value += " " + arg.toString();
            }
        }
        // Add last...
        if ((args.length > 0) && (name != null)) {
            arguments.put(name, new Argument(name, value));
        }
        return arguments;
    }

    /**
      * Returns the number of arguments.
      *  
      * @return The number of arguments.
      */
    public int size() {
        return this.arguments.size();
    }
    
    /**
      * Returns the hash code of these arguments.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.toString().hashCode();
    }
    
    /**
      * Returns true if {@code object} is an {@code Arguments}
      * instance storing the same arguments as these arguments,
      * false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof Arguments)) {
            return false;
        }
        return this.arguments.equals(((Arguments)object).arguments);
    }
    
    /**
      * Returns the string representation of these arguments.
      *
      * @return The string representation; never null.
      */
    public String toString() {
        return this.arguments.values().toString();
    }
    
}
