/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * An <i>annotated observers sequence decorator</i> decorates
  * any {@link Sequence} to become an {@link ObservableSequence observable} 
  * sequence that uses the {@link Executor} annotation to identify notification 
  * methods for observers that accepts a {@code Sequence} type as the
  * first argument and a {@link dk.rode.thesis.meta.model.Sequence.State}
  * type as the second. The context {@link Executor#value() type} must
  * furthermore be a type the <i>decorated</i> sequence type is assignable 
  * to. <p><p>
  *  
  * <b>Implementation notes</b>:<br>
  * Because Java does not support functional multiple inheritance,
  * this decorator cannot inherit both {@link SequenceDecorator}
  * and {@link AnnotatedObserversSequence}. Hence, all {@code Sequence}
  * methods have to be implemented explicitly in this class, 
  * while the observable functionality is inherited from
  * {@code AnnotatedObserversSequence}.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Executor
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
@Participant("ConcreteSubject")
public class AnnotatedObserversSequenceDecorator<E> 
     extends AnnotatedObserversSequence<E> {
    
    /**
      * The decorated sequence. <p>
      * 
      * Never null.
      */
    private final Sequence<E> sequence;
    
    /**
      * Constructor. <p>
      * 
      * @param  sequence The sequence to decorate; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public AnnotatedObserversSequenceDecorator(Sequence<E> sequence) {
        super(sequence.getClass());
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The observable sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public AnnotatedObserversSequenceDecorator(AnnotatedObserversSequenceDecorator<E> sequence) {
        super(sequence);
        this.sequence = sequence.copy();
    }

    /**
      * 
      */
    public AnnotatedObserversSequenceDecorator<E> copy() {
        return new AnnotatedObserversSequenceDecorator<E>(this.sequence.copy());
    }
    
    /**
      * 
      */
    public boolean bounded() {
        return this.sequence.bounded();
    }

    /**
      * 
      */
    public boolean consistent() {
        return this.sequence.consistent();
    }

    /**
      * 
      */
    public boolean unique() {
        return this.sequence.unique();
    }
    
    /**
      * 
      */
    public E current() {
        return this.sequence.current();
    }
    
    /**
      * 
      */
    protected E doNext() {
        return this.sequence.next();
    }
        
    /**
      * 
      */
    public void reset() {
        this.sequence.reset();
    }
    

    /**
      * 
      */
    public State state() {
        return this.sequence.state();
    }
    
    /**
      * 
      */
    public int hashCode() {
        return this.sequence.hashCode();
    }

    /**
      * 
      */
    public boolean equals(Object object) {
        if (!(object instanceof AnnotatedObserversSequenceDecorator)) {
            return false;
        }
        AnnotatedObserversSequenceDecorator<?> aosd = 
            (AnnotatedObserversSequenceDecorator<?>)object;
        return this.sequence.equals(aosd.sequence); 
    }
    
    /**
      * 
      */
    public String toString() {
        return this.toString(null).toString();
    }
    
    /**
      * 
      */
    public StringablePolicy<? super Sequence<E>> getStringablePolicy(StringablePolicy<? super Sequence<E>> policy) {
        return this.sequence.getStringablePolicy(policy);
    }
    
    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        return this.sequence.toString(policy);
    }
    
}
