/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import dk.rode.thesis.adapter.SequenceAdapter;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.proxy.ImmutableSequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>sequence observers sequence</i> implements the basic
  * traits of any {@link ObservableSequence observable} sequence 
  * that uses a {@link SequenceObserver sequence observer} as 
  * the notification mechanism. <p>
  * 
  * The {@link ObservableSequence observable} sequence 
  * functionality is fully implemented by this class! Sub-classes
  * need simply implement the {@link #doNext()} method
  * (as opposed to {@link #next()}) without worrying about
  * notifying observers. Observers will be notified if the
  * {@link #getAspect()} method return a non-null aspect
  * value. It is invoked immediately after {@code doNext()}. <p>
  * 
  * This sequence is not thread-safe! It does not enforce
  * synchronisation, neither when adding or removing observers, nor
  * when observers are notified. It could be potentially dangerous
  * or least time consuming to notify observers from a synchronised 
  * context, holding a lock on the sequence itself. If synchronisation
  * is required, the context using the sequence must enforce it. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Observable sequences inheriting this class are double as an observer: 
  * observable sequence {@code A} can add observable sequence {@code B} as 
  * an observer. The notification of observers is thus 
  * {@linkplain dk.rode.thesis.composite composite} in a depth-first manner. <p>
  * 
  * This class applies the {@linkplain dk.rode.thesis.templatemethod Template Method}
  * pattern: {@link #next()} is the template method, while {@code doNext()} and
  * {@code getAspect()} are primitive operations. This ensures consistent state
  * during notification. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence.
  * @param  <A> The type of the aspect used by this sequence. 
  */
@Participant({"Subject", "Observer"})
public abstract class SequenceObserversSequence<E,A> 
              extends AbstractSequence<E> 
           implements AspectObservableSequence<SequenceObserver<A>,A,E> {

    /**
      * A set containing the current added 
      * {@link SequenceObserver observers} and their associated
      * aspects they subscribe to. <p>
      * 
      * If the associated set contains null, the observer subscribes
      * to all aspect values. <p>
      * 
      * Never null, but may be empty.
      */
    private final Map<SequenceObserver<A>,Set<A>> observers;
    
    /**
      * Sub-class no-arg constructor. <p>
      */
    protected SequenceObserversSequence() {
        // do not reset, because we are not done initialising yet...
        super(false); 
        this.observers = new HashMap<SequenceObserver<A>,Set<A>>();
    }

    /**
      * Copy constructor.
      * 
      * @param  sequence The observable sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected SequenceObserversSequence(AspectObservableSequence<SequenceObserver<A>,A,E> sequence) {
        super(sequence);
        // Copy observer references...
        this.observers = new HashMap<SequenceObserver<A>,Set<A>>();
        for (SequenceObserver<A> observer : sequence.getObservers()) {
            this.observers.put(observer, new HashSet<A>(sequence.getAspects(observer)));
        }
     }
    
    /**
      * {@inheritDoc}
      * 
      * Observers associated with this observable sequence are notified
      * with the (new) <i>aspect</i> of this sequence if {@link #getAspect()}
      * returns a non-null value.
      *
      * @see #doNext()
      * @see #getAspect()
      * @see #sequenceEvent(Sequence, Object)
      */
    public final E next() {
        // Fetch value via sub-class hook...
        E value = this.doNext();
        // Notify observers if the aspect has changed...
        A aspect = this.getAspect();
        if (aspect != null) {
            this.sequenceEvent(this, aspect);
        }
        return value;
    }

    /**
      * Hook for sub-classes to perform the actual {@code next()}
      * operation <b>without</b> notifying observers.
      *  
      * @return The next element; never null.
      */
    protected abstract E doNext();

    /**
      * Hook for sub-classes to deliver the changed <i>aspect</i> value 
      * to notify observers about immediately after {@link #doNext()} has been
      * invoked, if any. <p>
      * 
      * If this method returns null, no notification is performed as the
      * aspect has not changed.
      * 
      * @return The changed aspect; or null if not changed. 
      */
    protected abstract A getAspect();
    
    /**
      * 
      */
    public Collection<SequenceObserver<A>> getObservers() {
        return Collections.unmodifiableSet(this.observers.keySet());
    }

    /**
      * 
      */
    public Collection<SequenceObserver<A>> getObservers(A aspect) {
        List<SequenceObserver<A>> observers = new ArrayList<SequenceObserver<A>>(this.observers.size());
        for (SequenceObserver<A> observer : this.observers.keySet()) { 
            Set<A> aspects = this.observers.get(observer);
            if ((aspects.contains(aspect)) || (aspects.contains(null))) { // null = all 
                observers.add(observer);
            }
        }
        return observers;
    }

    /**
      * 
      */
    public Set<A> getAspects(SequenceObserver<A> observer) {
        if (observer == null) {
            throw new NullPointerException("Observer cannot be null");
        }
        Set<A> aspects = this.observers.get(observer);
        if (aspects == null) {
            return new HashSet<A>();
        }
        return Collections.unmodifiableSet(aspects);
    }
    
    /**
      * 
      */
    public boolean addObserver(@Participant("ConcreteObserver") SequenceObserver<A> observer) {
        return this.addObserver(observer, null);
    }
    
    /**
      * 
      */
    public boolean addObserver(@Participant("ConcreteObserver") SequenceObserver<A> observer, A aspect) {
        // Cannot add this instance as an observer because it will cause
        // an infinite loop, but there is still no way of avoiding other
        // observers to cascade notifications that might eventually
        // lead to infinite loops...
        if (observer == this) {
            throw new IllegalArgumentException("Cannot add this sequence as observer for itself");
        }
        Set<A> aspects = this.observers.get(observer);
        if (aspects == null) {
            this.observers.put(observer, aspects = new HashSet<A>());
        }
        if (aspects.contains(aspect)) {
            return false;
        } else if (aspect == null) { // all
            aspects.clear();
        }
        aspects.add(aspect);
        return true;
    }

    /**
      * 
      */
    public boolean removeObserver(Object observer) {
        return this.removeObserver(observer, null);
    }

    /**
      * 
      */
    public boolean removeObserver(Object observer, A aspect) {
       if (observer == null) {
           throw new NullPointerException("Observer cannot be null");
       }
       Set<A> aspects = this.observers.get(observer);
       if (aspects == null) {
           return false;
       }
       boolean removed = aspects.remove(aspect);
       if (aspects.size() == 0) {
           this.observers.remove(observer);
       }
       return removed;
    }
    
    /**
      * 
      */
    public void sequenceEvent(Sequence<?> sequence, A aspect) {
        // When we notify observers, we wrap this sequence in an 
        // immutable proxy sequence so it cannot be altered by 
        // observers. This will also ensure that notifications are 
        // issued only by the actual client...
        Log out = LogFactory.getLog(this);
        
        // Notify...
        if (sequence == this) {
            if (this.observers.size() > 0) {
                // To create an immutable sequence, we have to adapt the actual type... 
                ImmutableSequence<?> is = 
                    new ImmutableSequence<Object>(SequenceAdapter.create(this));
                // Get observers subscribing to the supplied aspect...
                for (SequenceObserver<A> observer : this.getObservers(aspect)) {
                    out.println("-> Invoking: ", this.toString(SequencePolicy.ID), ".", observer, ": ", aspect);
                    observer.sequenceEvent(is, aspect);
                }
            }
            
        // Notified...
        } else {
            out.println("-> ", this.toString(SequencePolicy.ID), " notified about: ", sequence.toString(SequencePolicy.ID), ": ", aspect);
        }
    }
    
}
