/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.proxy;

import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>non-resettable sequence</i> ensures that invocations of 
  * {@link #reset()} are ignored for the proxied {@link Sequence sequence} 
  * <b>if and only if</b> the proxied sequence is accessed through 
  * the proxy only. <p>
  *
  * A non-resettable sequence exhibits the same {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link Sequence#unique() unique}
  * properties as the proxied sequence. <p>
  *  
  * <b>Implementation notes</b>:<br>
  * A non-resettable sequence corresponds to a <i>protection proxy</i> as described
  * by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.208].
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    SequenceProxyFactory#getNonResettableSequence(Sequence)
  * 
  * @param  <E> The type of values delivered by this sequence. 
  */
@ParticipantUsage(value = "Subject", type = Sequence.class)
@Participant("Proxy")
public class NonResettableSequence<E> extends SequenceDecorator<E> implements Sequence<E> {

    /**
      * Constructor. <p>
      * 
      * @param  sequence The sequence to proxy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NonResettableSequence(@Participant("RealSubject") Sequence<E> sequence) {
        super(sequence, ProxiedPolicy.VISIBLE);
    }
    
    /**
      *  
      */
    public NonResettableSequence<E> copy() {
        return new NonResettableSequence<E>(this.getSequence(true));
    }
    
    /**
      * Does not reset the proxied sequence. 
      */
    public final void reset() {
        // Do not reset and declared final to ensure the we 
        // maintain the semantics henceforth...
        LogFactory.getLog(this).println("reset() disallowed: ", this.toString(SequencePolicy.ID));
    }
    
}
