/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.proxy;

import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.Set;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.reflect.proxy.ProxyFactory;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * 
  * A <i>sequence proxy factory</i> creates various dynamic proxies for
  * any type of {@link Sequence sequence}. <p>
  * 
  * The proxies are all instances of the 
  * {@link java.lang.reflect.Proxy java.lang.reflect.Proxy} 
  * class. Each created proxy can be cast into any interface implemented
  * by the proxied sequence. <p>
  * 
  * Six different kinds of dynamic sequence proxies can be created by this 
  * factory: <ol>
  * 
  * <li> {@link #getLoggableSequence(Sequence)}: returns a sequence proxy
  *      that will record and time all access to the proxied sequence
  *      <b>if and only if</b> all access is through the proxy only. <p>
  *  
  * <li> {@link #getSynchronisedSequence(Sequence)}: returns
  *      a sequence proxy that guarantees thread-safe access to the proxied 
  *      sequence <b>if and only if</b> all access is through the proxy only. 
  *      The lock used synchronise on is the proxied sequence. <p> 
  *
  * <li> {@link #getNonResettableSequence(Sequence)}: returns a sequence 
  *      proxy that disallows the proxied sequence to be {@link Sequence#reset() 
  *      reset} <b>if and only if</b> all access is through the proxy only. <p> 
  * 
  * <li> {@link #getImmutableSequence(Sequence)}: returns a sequence 
  *      proxy that disallows the proxied sequence to be {@link Sequence#reset() 
  *      reset}, {@link Sequence#next() advanced}, or {@link Sequence#copy()
  *      copied} <b>if and only if</b> all access is through the proxy only. 
  *      Copying will simply return the proxy instance, <b>not</b> the proxied 
  *      object. <p>
  * 
  * <li> {@link #getVirtualSequence(InstantiableTypeLiteral, Object...)}: 
  *      returns a <i>virtual sequence</i> proxy that will create the 
  *      actual sequence to use once the first method declared in 
  *      the specific {@code Sequence} type or a super-interface is invoked. <p> 
  *      
  * <li> {@link #getSharedSequence(Sequence)}: returns a <i>smart pointer</i> 
  *      proxy that will allow a given sequence to be shared among copies of
  *      it until a given copy invokes either {@link Sequence#next()} or
  *      {@link Sequence#reset()}, which will detach the copy from the
  *      reference and copy the sequence used internally. <p>
  *       
  * </ol>
  *
  * Factory methods 2-4 create dynamic proxies that mimics the behaviour of 
  * the fixed sequence proxy classes {@link SynchronisedSequence}, 
  * {@link NonResettableSequence}, and {@link ImmutableSequence}, respectively,
  * but the returned dynamic proxies can still be cast into any interface 
  * implemented by the proxied sequence. <p>
  * 
  * It is perfectly legal to create sequence proxies based on other
  * sequence proxies. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The factory defers the actual proxy creation to the 
  * {@link ProxyFactory} class, which allows creation of proxies
  * based on any type. The {@code ProxyFactory} could have been 
  * inherited, but composition is chosen instead to keep the
  * interface of this class clean and focused on sequences. <p>
  * 
  * Each method used to create a sequence proxy defined in this class is 
  * an application of the {@linkplain dk.rode.thesis.factorymethod Factory Method} 
  * pattern.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ProxyFactory
  */
public class SequenceProxyFactory {

    /**
      * The singleton proxy factory used to create the proxy objects
      * representing {@link Sequence sequences}. <p>
      * 
      * Never null.
      */
    final static ProxyFactory factory = new ProxyFactory();
    
    /**
      * The {@code reset()} method defined in the 
      * {@link Sequence} interface. <p>
      * 
      * Never null.
      */
    private final Method reset;
    /**
      * The {@code next()} method defined in the 
      * {@link Sequence} interface. <p>
      * 
      * Never null.
      */
    private final Method next;
    /**
      * The {@code copy()} method defined in the 
      * {@link Sequence} interface. <p>
      * 
      * Never null.
      */
    private final Method copy;
    
    /**
      * No-arg constructor.
      */
    public SequenceProxyFactory() {
        try {
            this.reset = Sequence.class.getMethod("reset");
            this.next  = Sequence.class.getMethod("next");
            this.copy  = Sequence.class.getMethod("copy");
            
        } catch (Throwable t) {
            LogFactory.getLog(SequenceProxyFactory.class).error("Cannot fetch Sequence methods").error(t);
            throw toRuntimeException(t);
        }
    }

    /**
      * Returns a sequence proxy implementing all interfaces {@code sequence}
      * implements that will log all access to {@code sequence}
      * <b>if and only if</b> all access is though the returned proxy. <p>
      * 
      * The returned proxy can be cast into any interface implemented 
      * by {@code sequence}, but <b>not</b> the actual sequence type
      * (class) itself. Trying to cast the returned proxy into the actual 
      * type (class) of {@code sequence} will cause a class cast exception 
      * to be thrown!
      * 
      * @param  <E>      The type of values delivered by {@code sequence}. 
      * @param  sequence The sequence to proxy; cannot be null.
      * @return The proxy sequence; never null.
      */
    @ParticipantUsage(value = "Proxy", type = Sequence.class)
    public <E> Sequence<E> getLoggableSequence(@Participant("RealSubject") final Sequence<E> sequence) {
        // ISSUE: NetBeans 5.5.1 and Java 1.6.0_01 generates warning because
        //        of creation of generic array (varargs)...
        return factory.getLoggableObject(sequence);
    }
    
    /**
      * Returns a sequence proxy implementing all interfaces {@code sequence}
      * implements that will ensure synchronised access to {@code sequence}
      * <b>if and only if</b> all access is though the returned proxy. <p>
      * 
      * The returned proxy has similar functionality to the 
      * {@link SynchronisedSequence} class, but the returned proxy can still be 
      * cast into any interface implemented by {@code sequence}, though 
      * <b>not</b> the actual sequence type (class) itself. Trying to cast 
      * the returned proxy into the actual type (class) of {@code sequence} 
      * will cause a class cast exception to be thrown!
      * 
      * @param  <E>      The type of values delivered by {@code sequence}. 
      * @param  sequence The sequence to proxy; cannot be null.
      * @return The proxy sequence; never null.
      * @see    SynchronisedSequence
      */
    @ParticipantUsage(value = "Proxy", type = Sequence.class)
    public <E> Sequence<E> getSynchronisedSequence(@Participant("RealSubject") final Sequence<E> sequence) {
        // ISSUE: NetBeans 5.5.1 and Java 1.6.0_01 generates warning because
        //        of creation of generic array (varargs)...
        return factory.getSynchronisedObject(sequence);
    }

    /**
      * Returns a sequence proxy implementing all interfaces {@code sequence}
      * implements that will disallow {@link Sequence#reset()} to be invoked 
      * for {@code sequence} <b>if and only if</b> all access is though the 
      * returned proxy. <p>
      * 
      * The returned proxy has similar functionality to the 
      * {@link NonResettableSequence} class, but the returned proxy can still be 
      * cast into any interface implemented by {@code sequence}, though 
      * <b>not</b> the actual sequence type (class) itself. Trying to cast 
      * the returned proxy into the actual type (class) of {@code sequence} 
      * will cause a class cast exception to be thrown!
      * 
      * @param  <E>      The type of values delivered by {@code sequence}. 
      * @param  sequence The sequence to proxy; cannot be null.
      * @return The proxy sequence; never null.
      * @see    NonResettableSequence
      */
    @ParticipantUsage(value = "Proxy", type = Sequence.class)
    public <E> Sequence<E> getNonResettableSequence(@Participant("RealSubject") final Sequence<E> sequence) {
        InvocationHandler handler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                if (SequenceProxyFactory.this.reset.equals(method)) {
                    LogFactory.getLog(InvocationHandler.class).println("reset() disallowed: ", ObjectPolicy.ID.toString(proxy), " -> ", sequence.toString(SequencePolicy.ID));
                    return null;
                }
                return Reflection.invoke(sequence, method, args);
            }
        };
        return factory.getProxy(sequence, handler);
    }
    
    /**
      * Returns a sequence proxy implementing all interfaces {@code sequence}
      * implements that will disallow {@link Sequence#reset()}, 
      * {@link Sequence#next()}, and {@link Sequence#copy()} to alter  
      * {@code sequence} <b>if and only if</b> all access is though the 
      * returned proxy. <p>
      * 
      * The {@code copy()} method simply returns the proxy instance, <b>not</b>
      * the proxied object. <p>
      * 
      * The returned proxy has similar functionality to the 
      * {@link ImmutableSequence} class, but the returned proxy can still be 
      * cast into any interface implemented by {@code sequence}, though 
      * <b>not</b> the actual sequence type (class) itself. Trying to cast 
      * the returned proxy into the actual type (class) of {@code sequence} 
      * will cause a class cast exception to be thrown!
      * 
      * @param  <E>      The type of values delivered by {@code sequence}. 
      * @param  sequence The sequence to proxy; cannot be null.
      * @return The proxy sequence; never null.
      * @see    ImmutableSequence
      */
    @ParticipantUsage(value = "Proxy", type = Sequence.class)
    public <E> Sequence<E> getImmutableSequence(@Participant("RealSubject") final Sequence<E> sequence) {
        // We cannot be a proxy to a non-resettable proxy here, because
        // the handler perform a call to any non-illegal method. Hence,
        // if the reset is not caught, it will call the supplied method,
        // which could be "next()" or "copy()" which this handler disallows.
        // Hence, we have to test for all scenarios in a separate handler
        // as done below...
        InvocationHandler handler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                // "next()"...
                if (SequenceProxyFactory.this.next.equals(method)) {
                    LogFactory.getLog(InvocationHandler.class).println("next() disallowed: ", ObjectPolicy.ID.toString(proxy), " -> ", sequence.toString(SequencePolicy.ID));
                    return sequence.current();
                    
                // "reset()"...
                } else if (SequenceProxyFactory.this.reset.equals(method)) {
                    LogFactory.getLog(InvocationHandler.class).println("reset() disallowed: ", ObjectPolicy.ID.toString(proxy), " -> ", sequence.toString(SequencePolicy.ID));
                    return null;
                    
                // "copy()"...
                } else if (SequenceProxyFactory.this.copy.equals(method)) {
                    LogFactory.getLog(InvocationHandler.class).println("copy() disallowed: ", ObjectPolicy.ID.toString(proxy), " -> ", sequence.toString(SequencePolicy.ID));
                    // Return same proxy instance...
                    return proxy;
                }
                return Reflection.invoke(sequence, method, args);
            }
        };
        return factory.getProxy(sequence, handler);
    }

    /**
      * Creates a <i>virtual sequence</i> based on the {@link Sequence sequence}
      * type {@code T}. <p>
      * 
      * Once the first method declared in {@code T}, or in
      * a super-type of {@code T}, is invoked on the returned proxy,
      * the proxy creates the actual proxied sequence that will be used 
      * as the subject hence forth. <p>
      * 
      * The returned proxy is <b>not</b> registered to this factory as the
      * proxied object is not yet created. Hence, any call to 
      * {@link #isProxy(Sequence)} with the returned proxy as the 
      * argument will return false! <p>
      * 
      * The returned proxy can be cast into any super-interface of 
      * {@code T}, or {@code T} itself if {@code T} is not a class.
      *  
      * @param  <E>       The type of sequence values delivered by the sequence.
      * @param  <T>       The actual type of sequence to create a virtual proxy for.
      * @param  type      The type literal representing the actual type; cannot be null.
      * @param  arguments The arguments used to construct the proxied sequence.
      * @return The virtual proxy representing the sequence to create; never null.
      * @throws NullPointerException If {@code type} is null.
      */
    @ParticipantUsage(value = "Proxy", type = Sequence.class)
    public <E,T extends Sequence<E>> T getVirtualSequence(@Participant("RealSubject") final InstantiableTypeLiteral<T> type, final Object... arguments) {
        final Set<Class<?>> types = Reflection.getInterfaces(type.getRawType());
        class VirtualSequence implements InvocationHandler {
            // Sequence will be created once the first method
            // declared in Sequence or a sub-type is invoked...
            private Sequence<E> sequence;
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                if (this.sequence == null) {
                    if (types.contains(method.getDeclaringClass())) {
                        this.sequence = type.newInstance(arguments);
                        LogFactory.getLog(InvocationHandler.class).println("Sequence created: ", this.sequence);
                    }
                }
                // If we have no created a sequence yet we simply use the
                // type literal as the target (hash code, equals, toString, etc).
                // This could - and probably will - go wrong, but this example
                // is contrived regardless... :) 
                return Reflection.invoke(this.sequence == null ? type : this.sequence, method, args);
            }
            public String toString() {
                StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
                return sb.append("; sequence: ").append(this.sequence).append(']').toString();
            }
        }
        // Create a proxy which is not registered. The cast to T is known to succeed here, but will cause
        // an exception in the calling context if the returned type is assigned to a non-interface type - 
        // the returned proxy must be assigned to one of the types in "types"!
        @SuppressWarnings("unchecked") T proxy = 
            (T)factory.getProxy(new VirtualSequence(), types.toArray(new Class<?>[types.size()])); 
        return proxy;
    }

    /**
      * Creates a <i>smart pointer</i> based on the {@link Sequence sequence}
      * supplied as {@code sequence}. <p>
      * 
      * Call the returned proxy X, using {@code sequence} as its internal
      * sequence, which may be shared. Hence, {@code sequence} will be
      * shared among all {@link Sequence#copy() copies} of X (inclusive), say 
      * X and Y, until a given copy, say Y, invokes either 
      * {@link Sequence#next() next()} or {@link Sequence#reset() reset()}. This 
      * will copy {@code sequence} and detach Y. Y will henceforth use the copy 
      * of {@code sequence} just made, and Y can be shared (copied) as well, say 
      * into Z, which will use the same internal sequence as Y, and so forth. <p>
      * 
      * The returned proxy will <b>not</b> be {@link #isProxy(Sequence) registered} 
      * to this factory.
      * 
      * @param  <E>      The type of sequence values delivered by {@code sequence}.
      * @param  sequence The sequence to share; cannot be null.
      * @return A proxy representing the shared sequence; never null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public <E> Sequence<E> getSharedSequence(@Participant("RealSubject") final Sequence<E> sequence) {
        // "copy" = copy method with no arguments, 
        // "next" and "reset" mutable methods...
        try {
            return factory.getSharedObject(sequence, "copy", null, "next", "reset");
        } catch (Exception e) {
            throw toRuntimeException(e);
        }
    }
    
    /**
      * Returns true if {@code sequence} is a proxy sequence created
      * by a sequence proxy factory, false if not.
      * 
      * @param  sequence The sequence to test; cannot be null.
      * @return True if {@code sequence} is a proxy sequence, false if not.
      * @throws NullPointerException If {@code sequence} is null.
      * @see    #isProxied(Sequence)
      */
    public boolean isProxy(Sequence<?> sequence) {
        return factory.isProxy(sequence);
    }
    
    /**
      * Returns true if {@code sequence} has at least one created 
      * proxy sequence associated with it, false if not. <p>
      *
      * {@code sequence} may be a proxy itself. 
      * 
      * @param  sequence The sequence to test; cannot be null.
      * @return True if {@code sequence} has at least one associated 
      *         proxy, false if not. 
      * @throws NullPointerException If {@code sequence} is null.
      * @see    #isProxy(Sequence)
      */
    public boolean isProxied(Sequence<?> sequence) {
        return factory.isProxied(sequence);
    }
    
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.ID.asString(this);
    }
    
}
