/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import static dk.rode.thesis.meta.reflect.Reflection.isOverriding;
import static dk.rode.thesis.meta.util.Throwables.unwind;

import java.lang.reflect.Constructor;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * <b>Singleton</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Singleton</b> tests. 
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Singleton tests...");
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
       
        boolean ok = true;

        // True for class based logging. Note, that this does not
        // cause the singleton classes to be initialised!
        // True for verbose logging...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(StatelessSingletonRegistry.class).setActive(log);
        LogFactory.getLog(StatefullSingletonRegistry.class).setActive(log);
        LogFactory.getLog(LoadableSingletonRegistry.class).setActive(log);

        // Create a loadable singleton registry to load and/or acquire the types. Since
        // we have to test for creation time, the classes are actually not loaded
        // via the singleton registry...
        LoadableSingletonRegistry<Sequence<String>> registry = 
            new LoadableSingletonRegistry<Sequence<String>>(){/*magic*/};
        
        LogFactory.getLog(DanishAlphabetSequence.class).setActive(log);
        LogFactory.getLog(NorwegianAlphabetSequence.class).setActive(log);
        LogFactory.getLog(SmileySequence.class).setActive(log);
        LogFactory.getLog(SimpsonsFamilySequence.class).setActive(log);
        LogFactory.getLog(SimpsonsAndBouvierFamilySequence.class).setActive(log);
        LogFactory.getLog(MutatedSimpsonsFamilySequence.class).setActive(log);
        LogFactory.getLog(StatelessSingletonRegistry.class).setActive(log);
        LogFactory.getLog(StatefullSingletonRegistry.class).setActive(log);
        
        // Creation time...
        out.heading("Number of sequence values in singleton sequences");
        out.println("(test of loading time)").println();
        out.println(DanishAlphabetSequence.class.getSimpleName(), "           : ", DanishAlphabetSequence.SIZE);
        out.println(NorwegianAlphabetSequence.class.getSimpleName(), "        : ", NorwegianAlphabetSequence.SIZE);
        out.println(SmileySequence.class.getSimpleName(), "                   : ", SmileySequence.SIZE);
        out.println(SimpsonsFamilySequence.class.getSimpleName(), "           : ", SimpsonsFamilySequence.SIZE);
        out.println(SimpsonsAndBouvierFamilySequence.class.getSimpleName(), " : ", SimpsonsAndBouvierFamilySequence.SIZE);
        
        MutatedSimpsonsFamilySequence msfs = null;
        EnglishAlphabetSequence eas = null;

        out.heading("Sequence instances");
        
        Sequence<?>[] sequences = {
            SimpsonsAndBouvierFamilySequence.getFamily(), 
            // hashCode/equals not overridden... 
            new LongSequence(),
            // Non singleton type inheriting a singleton type... (variable used in registry test as well)
            msfs = new MutatedSimpsonsFamilySequence(),  
            // hashCode/equals overridden... (variable used in registry test as well)
            eas = new EnglishAlphabetSequence(),
            // Singleton...
            SimpsonsFamilySequence.getFamily(), 
            new EnglishAlphabetSequence(),
            // Enumeration singleton
            DanishAlphabetSequence.Instance,
            // Singleton...
            NorwegianAlphabetSequence.getInstance(),
            new MutatedSimpsonsFamilySequence(), 
            // Singleton...
            NorwegianAlphabetSequence.getInstance(), 
            new LongSequence(),
            // Enumeration singleton...
            DanishAlphabetSequence.Instance,
            // Singleton...
            SimpsonsFamilySequence.getFamily(), 
            // Singleton...
            SmileySequence.getSequence()
        };
        
        // Print instances...
        for (int i = 0; i < sequences.length; i++) {
            out.println("Sequence ", i, ": ", sequences[i]);
        }

        // Equality (do not need to test both ways because of the standard
        // Java hashCode/equals contract...
        out.heading("Sequence equality");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println(i, ": ", sequence.toString(SequencePolicy.ID), " equals/same identity...");
            boolean overridden = ((isOverriding(sequence.getClass(), "equals", Object.class)) && (isOverriding(sequence.getClass(), "hashCode")));
            if (overridden) {
                out.println("(\"hashCode()\" and \"equals(Object)\" overridden)");
            }
            for (int j = i; j < sequences.length; j++) {
                Sequence<?> s = sequences[j];
                out.println("--> ", j, ": ", s.toString(SequencePolicy.ID), ": equal = ", sequence.equals(s), ", identity = ", sequence == s);
            }
        }
        
        out.heading("Sequence usage");
        for (int i = 0; i < 30; i++) {
            out.println("Iteration: ", i);
            for (int j = 0; j < sequences.length; j++) {
                Sequence<?> sequence = sequences[j];
                // Advance sequence...
                Object value = sequence.next();
                out.println(j, ": ", sequence, " -> ", value);
            }
            out.println();
        }
        
        out.heading("Sequence equality (copies)");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i].copy();
            out.println(i, ": ", sequence.toString(SequencePolicy.ID), " equals/same identity...");
            boolean overridden = ((isOverriding(sequence.getClass(), "equals", Object.class)) && (isOverriding(sequence.getClass(), "hashCode")));
            if (overridden) {
                out.println("(\"hashCode()\" and \"equals(Object)\" overridden)");
            }
            for (int j = i; j < sequences.length; j++) {
                Sequence<?> s = sequences[j].copy();
                out.println("--> ", j, ": ", s.toString(SequencePolicy.ID), ": equal = ", sequence.equals(s), ", identity = ", sequence == s);
            }
        }
        
        out.heading("Singleton test using registry");

        out.println("Created registry: ", registry).println();

        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.print(i, ": ", sequence.toString(SequencePolicy.ID), " is ");
            if (registry.isSingleton(sequence.getClass())) {
                out.println("a singleton");
            } else {
                out.println("not a singleton");
            }
        }
        
        out.heading("Getting singleton instances using registry");
 
        // Supposed to succeed!
        out.println("Getting singletons...").println();
        
        ok = this.getSingleton(out, registry, DanishAlphabetSequence.class,           DanishAlphabetSequence.Instance)              && ok;
        ok = this.getSingleton(out, registry, NorwegianAlphabetSequence.class,        NorwegianAlphabetSequence.getInstance())      && ok;
        ok = this.getSingleton(out, registry, SimpsonsFamilySequence.class,           SimpsonsFamilySequence.getFamily())           && ok;
        ok = this.getSingleton(out, registry, SimpsonsAndBouvierFamilySequence.class, SimpsonsAndBouvierFamilySequence.getFamily()) && ok;
        ok = this.getSingleton(out, registry, SmileySequence.class,                   SmileySequence.getSequence())                 && ok;

        out.println("Loading singletons...").println();
        
        // Cannot load DanishAlphabetSequence because it is an enumeration...
        ok = this.loadSingleton(out, registry, NorwegianAlphabetSequence.class.getName(),        NorwegianAlphabetSequence.getInstance())      && ok;
        ok = this.loadSingleton(out, registry, SimpsonsFamilySequence.class.getName(),           SimpsonsFamilySequence.getFamily())           && ok;
        ok = this.loadSingleton(out, registry, SimpsonsAndBouvierFamilySequence.class.getName(), SimpsonsAndBouvierFamilySequence.getFamily()) && ok;
        ok = this.loadSingleton(out, registry, SmileySequence.class.getName(),                   SmileySequence.getSequence())                 && ok;
        
        out.println("Getting non singletons (supposed to fail!)...").println();
        ok = !this.getSingleton(out, registry, EnglishAlphabetSequence.class,       eas)  && ok;
        ok = !this.getSingleton(out, registry, MutatedSimpsonsFamilySequence.class, msfs) && ok;
        
        out.println("Loading non singletons (supposed to fail!)...").println();
        ok = !this.loadSingleton(out, registry, MutatedSimpsonsFamilySequence.class.getName(), msfs) && ok;
        // Should fail because of type parameter!
        ok = !this.loadSingleton(out, registry, LongSequence.class.getName(), msfs) && ok;
        
        out.println("Singleton objects from registry: ", registry);
        int count = 0;
        for (Sequence<String> s : registry) {
            out.println("--> ", s);
            count++;
        }
        // Cause we acquired three singleton types...
        if (count != 5) {
            out.error("Illegal number of acquired singletons: ", count, " -> ", registry);
            ok = false;
        }

        out.heading("Testing singleton invariants");

        out.println("Trying to create SimpsonsFamilySequence instance using accessible constructor").println();
        try {
            SimpsonsFamilySequence sabfs = 
                new SimpsonsFamilySequence(new String[]{"Crusty the Clown"});
            out.error("Could create singleton instance using constructor: ", sabfs);
            ok = false;
            
        } catch (SingletonError se) {
            out.print("OK - could not create instance: ").println(se);
            
        } catch (Throwable t) {
            out.error("Unknown error").error(t);
            ok = false;
        }
        
        out.println("Trying to create SimpsonsFamilySequence instance using accessible copy constructor").println();
        try {
            SimpsonsFamilySequence sabfs = 
                new SimpsonsFamilySequence(SimpsonsAndBouvierFamilySequence.getFamily());
            out.error("Could create singleton instance using copy constructor: ", sabfs);
            ok = false;
            
        } catch (SingletonError se) {
            out.print("OK - could not create instance: ").println(se);
            
        } catch (Throwable t) {
            out.error("Unknown error").error(t);
            ok = false;
        }

        out.println("Trying to create SimpsonsFamilySequence instance using MutatedSimpsonsFamilySequence.evilCreation()").println();
        try {
            SimpsonsFamilySequence sabfs = msfs.evilCreation();
            out.error("Could create singleton instance using evilCreatuib(): ", sabfs);
            ok = false;
            
        } catch (SingletonError se) {
            out.print("OK - could not create instance: ").println(se);
            
        } catch (Throwable t) {
            out.error("Unknown error").error(t);
            ok = false;
        }
        
        out.println("Trying to create singletons from registry using reflection");

        for (Sequence<?> s : registry) {
            out.println().println("Testing (class): ", s);
            if (s instanceof Enum) {
                out.println("Sequence is an enumeration constant: ", s);
                continue;
            }
            Class<?> type = s.getClass();
            Constructor<?> constructor = null;
            
            // Fetch declared (private) constructor...
            try {
                constructor = type.getDeclaredConstructor();
                
            } catch (NoSuchMethodException nsme) {
                out.error("Could not get private constructor: ", type).error(nsme);
                ok = false;
                continue;
            }

            // Make accessible...
            try {
                // Make sure we can access it (if the security manager 
                // allows it). If not, then there is indeed no way to
                // create the instance, which is a good thing...
                constructor.setAccessible(true);
                
            } catch (SecurityException se) {
                out.warn("Security manager do not allow alteration of accessible property: ", type, " -> ", constructor).error(se);
                continue;
            }
            
            // Try create... 
            try {
                Object sequence = constructor.newInstance();
                out.error("Could create singleton instance using reflection: ", sequence);
                ok = false;
                
            // Java's reflection code will wrap exceptions in
            // various other types, so we have to unwind to get the
            // desired type, if at all possible...             
            } catch (Exception e) {
                SingletonError se = unwind(e, SingletonError.class);
                if (se != null) {
                    out.print("OK - could not create instance using reflection: ").println(se);
                    continue;
                }
                out.error("Could not invoke constructor: ", type).error(e);
                ok = false;
                continue;
            }
        }
        
        //
        // Clone tests
        //
        
        try {
            msfs = new MutatedSimpsonsFamilySequence().clone(); 
            out.println("Cloned mutated Simpsons family sequence: ", msfs);
            
        } catch (Exception e) {
            out.error("Could not clone mutated Simpsons family sequence: ").println(e);
            ok = false;
        }
        
        out.heading("Singleton tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Tests the singleton object supplied as {@code singleton} against the
      * return instance from {@code registry} when looked up using {@code type}.
      * 
      * @param  <T>       The super-type of singleton types supported by {@code registry}. 
      * @param  <S>       The actual singleton type to test.       
      * @param  out       The log to use; never null.
      * @param  registry  The registry to use to acquire the singleton instance based
      *                   on {@code type}; never null.
      * @param  type      The type of the singleton object to test; never null.
      * @param  singleton The known singleton object to test on; never null.
      * @return True if the test succeeded, false if not.
      */
    private <T,S extends T> boolean getSingleton(Log out, SingletonRegistry<T> registry, Class<S> type, S singleton) {
        out.println("Testing: ", type.getSimpleName(), " -> ", singleton);
        boolean ok = true;
        try {
            S instance = registry.getInstance(type);
            if (instance != singleton) {
                out.error("Returned instance different from supplied singleton: ", instance, " != ", singleton);
                ok = false;
                
            } else {
                out.println("Returned instance matches supplied singleton");
            }
            
        } catch (SingletonException se) {
            out.error("Could not acquire singleton").error(se);
            ok = false;
        }
        out.println();
        return ok;
    }

    /**
      * Tests the singleton object supplied as {@code singleton} against the
      * return instance from {@code registry} when looked up using {@code type}.
      * 
      * @param  <T>       The super-type of singleton types supported by {@code registry}. 
      * @param  <S>       The actual singleton type to test.       
      * @param  out       The log to use; never null.
      * @param  registry  The registry to use to acquire the singleton instance based
      *                   on {@code type}; never null.
      * @param  className The class name of the singleton type to test; never null.
      * @param  singleton The known singleton object to test on; never null.
      * @return True if the test succeeded, false if not.
      */
    private <T,S extends T> boolean loadSingleton(Log out, LoadableSingletonRegistry<T> registry, String className, S singleton) {
        out.println("Testing: ", className, " -> ", singleton);
        boolean ok = true;
        try {
            T instance = registry.getInstance(className);
            if (instance != singleton) {
                out.error("Loaded instance different from supplied singleton: ", instance, " != ", singleton);
                ok = false;
                
            } else {
                out.println("Loaded instance matches supplied singleton");
            }
            
        } catch (SingletonException se) {
            out.error("Could not acquire singleton").error(se);
            ok = false;
        } catch (ClassNotFoundException cnfe) {
            out.error("Could not load singleton").error(cnfe);
            ok = false;
        } catch (ClassCastException cce) {
            out.error("Illegal singleton type").error(cce);
            ok = false;
        }
        out.println();
        return ok;
    }
    
}
