/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.singleton;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>stateless singleton registry</i> is a registry with no
  * internal state that allows any type annotated correctly
  * with the {@link Singleton} annotation to be fetched, and
  * possibly created, using the {@link #getInstance(Class)}
  * method. <p> 
  *  
  * <b>Implementation notes</b>:</br>
  * This registry do <b>not</b> utilise an internal cache to store the
  * acquired singletons because that would require synchronisation.
  * Furthermore, we do not know how the actual singleton classes are 
  * implemented, i.e. static initialisation, lazily initialised, synchronised, 
  * or not synchronised!? Hence, each time the {@code getInstance(Class)} 
  * method is invoked the call is forwarded to the actual <i>singleton 
  * method</i> identified by the {@code Singleton} annotation used to 
  * annotate the type supplied as the {@code Class} argument. This means
  * that all singleton initialisation and creation is handled by the
  * individual singleton types! 
  *   
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The super type of all singleton types handled by
  *             this registry. Use {@code Object} if any type
  *             of singleton type can be handled.
  */
public class StatelessSingletonRegistry<T> implements SingletonRegistry<T> {

    /**
      * Constructor.
      */
    public StatelessSingletonRegistry() {
        super();
    }
    
    /**
      * @throws NullPointerException {@inheritDoc} 
      * @throws SingletonException   {@inheritDoc}
      */
    public <S extends T> S getInstance(Class<S> type) { 
        Singleton singleton = type.getAnnotation(Singleton.class);
        if (singleton == null) {
            throw new SingletonException(type);
        }
        Log out = LogFactory.getLog(this);
        try {
            // Fetch singleton method name...
            String singletonMethod = singleton.value();
            // Fetch fixed string arguments...
            Object[] arguments = singleton.arguments();
            // Generate formal parameter types...
            Class<?>[] parameterTypes = new Class<?>[arguments.length];
            for (int i = 0; i < arguments.length; i++) {
                parameterTypes[i] = String.class;
            }
            out.println("Singleton method: ", type.getName(), " -> ", singletonMethod);
            // Fetch actual singleton method...
            Method method = type.getDeclaredMethod(singletonMethod, parameterTypes);
            // Make accessible if not already...
            if (!Modifier.isPublic(method.getModifiers())) {
                method.setAccessible(true);
            }
            // Static, so no target...
            S instance = type.cast(method.invoke(null, arguments));
            out.println("Acquired singleton instance: ", type.getName(), " -> ", instance);
            return instance;
            
        } catch (Exception e) {
            out.error("Could not acquire singleton type: ", type).error(e);
            throw new SingletonException(e);
        }
    }

    /**
      * @throws NullPointerException {@inheritDoc} 
      */
    public boolean isSingleton(Class<?> type) {
        return (type.getAnnotation(Singleton.class) != null);
    }
    
    /**
      * Returns the string representation of this registry.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.IDENTITY.asString(this);
    }
 
}

