/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.templatemethod;

import static dk.rode.thesis.meta.util.Const.INFINITE;
import static dk.rode.thesis.meta.util.Throwables.toThrowableType;

import java.io.Closeable;
import java.util.Arrays;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>sequence template</i> forms a template for sequences
  * requiring several steps to construct, deliver, and dispose 
  * sequence values. An example could be values fetched from a
  * database that requires steps like <i>open connection</i>, 
  * <i>begin transaction</i>, <i>fetch value</i>, 
  * <i>end transaction</i>, and <i>close connection</i>. <p>  
  * 
  * <b>Sub-class implementation</b>:<br>
  * The basic sequence methods {@link #current()}, {@link #next()}, 
  * and {@link #reset()} as well as {@link #close()} are declared 
  * final - they are <i>template methods</i>. Instead, a sequence 
  * template defines the following <i>(abstract) primitive operations</i> 
  * that <b>can</b> or <b>must</b> be implemented, i.e. if abstract, in 
  * sub-class implementations: <ul> <p>
  * 
  * <li> <b>Sequence related operations</b>: sequences must be opened before values
  *      are delivered, reset for reuse, and closed after use: <ul><p>
  *  
  * <li> {@link #sequenceOpen(Object...)}: performs initialisation of 
  *      this sequence before any sequence values are delivered ("<i>open 
  *      connection</i>"). Is invoked upon {@link #SequenceTemplate(Object...) 
  *      construction} of new sequences and 
  *      {@link #SequenceTemplate(SequenceTemplate, Object...) copying} of 
  *      existing sequences when needed. Sub-class constructors should therefore
  *      <b>not</b> perform initialisation. <b>Must be implemented</b>. <p>
  * 
  * <li> {@link #sequenceReset()}: resets this sequence for reuse when
  *      this sequence is restarted or {@link #reset() reset} ("<i>close and
  *      re-open connection</i>"). <b>Default implementation does nothing</b>. <p>
  *      
  * <li> {@link #sequenceClose()}: closes this sequence after use, so no more 
  *      sequence values can be delivered by this sequence ("<i>close connection</i>").
  *      Once this sequence has been closed, trying to acquire sequence values
  *      from it will cause an {@link SequenceValueException exception} to
  *      be thrown. <b>Default implementation does nothing</b>. Is invoked 
  *      when sequences are {@link #close() closed}. <p>
  * 
  * </ul>
  * 
  * <li> <b>Value related operations</b>: values delivered by {@link #next() next()} are 
  *      created in three steps: <ul><p>
  * 
  * <li> {@link #nextValueOpen(Object) nextValueOpen(E)}: prepares the {@link #next() next}
  *      sequence value to be delivered ("<i>begin transaction</i>"). The 
  *      argument supplied is the current (previous) 
  *      sequence value and the returned value is a <i>correlation key</i> used 
  *      to identify the value to sub-sequence calls to 
  *      {@link #nextValue(Object, Object) nextValue(K, E)} and 
  *      {@link #nextValueClose(Object, Object) nextValueClose(K, E)}
  *      methods for the same value. If no correlation key is used, simply
  *      return null. <b>Must be implemented</b>. <p>
  * 
  * <li> {@link #nextValue(Object, Object) nextValue(K, E)}: fetches or constructs the 
  *      actual sequence value to be delivered as the next value, or returns null if 
  *      this sequence has no more values ("<i>fetch value</i>"). The type of value 
  *      is determined by the type parameter {@code E}. The first supplied argument 
  *      is the <i>correlation key</i> returned by the {@link #nextValueOpen(Object)
  *      nextValueOpen(E)} method, and the second the current (previous) value. 
  *      <b>Must be implemented</b>. <p>
  * 
  * <li> {@link #nextValueClose(Object, Object) nextValueClose(K, E)}: finalises the 
  *      sequence value just constructed with {@link #nextValue(Object, Object)
  *      nextValue(K, E)} ("<i>end transaction</i>"). The supplied arguments are in 
  *      order the <i>correlation key</i> generated by {@link #nextValueOpen(Object) 
  *      nextValueOpen(E)} and the actual sequence value generated by 
  *      {@link #nextValue(Object, Object) nextValue(K, E)}. <b>Default implementation 
  *      does nothing</b>. <p>
  * 
  * </ul></ul>
  *  
  * In <i>pseudo-code</i>, when delivering the {@link #next() next} sequence value,
  * the functionality of a sequence template is thus:
  * <pre>
  *   // current = old value
  *   E current = this.{@link #current()};
  *   
  *   // Open and get correlation key...
  *   K key = {@link #nextValueOpen(Object) nextValueOpen(current)};
  *   
  *   // Get actual value based on correlation key and *old* value...
  *   value = {@link #nextValue(Object, Object) nextValue(key, current)};
  *   
  *   // Close newly created value (e.g. resources)...
  *   {@link #nextValueClose(Object, Object) nextValueClose(key, value)};
  *   
  *   // Update current...
  *   this.current = value;
  *   return value;
  * </pre>
  *
  * As default, sequence templates are declared as {@link #bounded() bounded},
  * {@link #consistent() consistent}, and {@link #unique() unique}, but sub-classes
  * should override this behaviour as appropriate. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * All attributes are declared private in this class to prevent sub-classes 
  * from modifying them outside the scope of the (abstract) operations. <p>
  * 
  * Java makes it easy for sub-classes to determine which (abstract) operations
  * must be implemented: those declared abstract. Furthermore, to ensure that
  * the template methods are not overridden, they should be declared final. This
  * also gives the sub-class implementer a clear view of which methods can
  * be overridden. Gamma et al. also emphasise the need for consistent 
  * naming among (abstract) operations for clarity [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.329]. Here, 
  * all operations related to overall sequence functionality are prefixed with 
  * {@code sequence}, and operations related to generating the next value with 
  * {@code nextValue}. The last part of the names of the (abstract) operations 
  * determine their functionality, e.g. {@code Open}
  * and {@code Close} as in {@code nextValueOpen} and {@code nextValueClose}. <p>
  * 
  * As described above, the initialisation of sub-class sequence implementations 
  * is performed using the {@link #sequenceOpen(Object...)} method, invoked 
  * automatically via the constructors defined in this class when
  * needed. Sub-class constructors must invoke one of these constructors, but will 
  * declare explicit arguments types as needed to create the specific type of 
  * sequence. They will thus pass matching arguments to the super-constructor, which then
  * invokes the {@code sequenceOpen(Object...)} initialisation method. However, casts
  * are still required in the method implementation, but are known to succeed since the 
  * arguments originate from one of the parameter type-safe constructors in the same
  * class. This is in our view a very elegant use of varargs. An alternative could
  * be to use an additional generic type parameter on this class, say {@code P}, to 
  * narrow the argument type(s) for the constructors and the {@code sequenceOpen(Object...)}
  * method (and thus in sub-classes overriding it) even further, e.g. 
  * {@code sequenceOpen(P...)}. But, if more than one argument is needed with 
  * different type hierarchies, the actual type parameter for {@code P} would still 
  * need to be {@code Object}, so the question is how much will be gained?! Here, no 
  * such type parameter is used. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <K> The type of <i>correlation keys</i> used by this sequence. Use
  *             {@code Void} if not used.
  * @param  <E> The type of values delivered by this sequence. 
  */
@Participant("AbstractClass")
public abstract class SequenceTemplate<K,E> extends AbstractSequence<E> implements Closeable {
    
    /**
      * Policy use for logging of this sequence. <p>
      * 
      * Change at compile time for more verbose output.
      */
    protected final static SequencePolicy POLICY = SequencePolicy.ID;

    /**
      * The current sequence value. <p>
      * 
      * Never null.
      */
    private E current;
    /**
      * The size of this sequence, i.e. the number of values delivered. <p>
      * 
      * Calculated once this sequence has restarted. <p>
      * 
      * Initialised to {@link Const#INFINITE}. 
      */
    private int size;
    /**
      * The current index into this sequence, used to calculate the size
      * of this sequence. <p>
      * 
      * Interval: 0 <= index <= {@link #size} (once size has been calculated).
      */
    private int index;
    /**
      * True if this sequence has been {@link #close() closed}. If so, no more 
      * values can be delivered by this sequence.
      */
    private boolean closed;
    
    /**
      * Constructor, which creates this sequence template. <p> 
      *
      * The {@link #sequenceOpen(Object...) initialisation} of
      * this sequence is performed with {@code arguments} as arguments. <p>
      * 
      * @param  arguments The arguments, if any, to use in the 
      *                   initialisation of this sequence; can be null.
      * @throws SequenceValueException If the initialisation fails, or if the 
      *         initial current value of this sequence cannot be fetched.
      */
    protected SequenceTemplate(Object... arguments) throws SequenceValueException {
        super(false);
        this.index = 0;
        this.size = INFINITE;
        try {
            // Perform sub-class specific sequence initialisation, if any. A
            // returned true value indicates that the initialisation
            // succeed, false that it failed...
            LogFactory.getLog(this).println("Opening sequence: ", this.toString(POLICY), " -> arguments: ", Arrays.asList(arguments));
            this.closed = !this.sequenceOpen(arguments); // abstract operation
            
            // Reset - and set current value!...
            this.reset();
            
            // Make sure reset did in fact reset the current value...
            if (this.current == null) {
                throw new SequenceValueException("Current value is null");
            }
            
            // Initial state to maintain sequence semantics...
            this.state = State.START;
            
        // Catch all, transform if needed, and re-throw...
        } catch (Exception e) {
            throw toThrowableType(e, SequenceValueException.class);
        }
    }
    
    /**
      * Copy constructor, which creates this sequence template based on
      * {@code sequence}, i.e. to have the same size, use the same
      * current index, etc. <p> 
      *
      * Initialisation of this sequence template is still performed unless
      * {@code sequence} is {@link #isClosed() closed}. The 
      * {@link #sequenceOpen(Object...) initialisation} is thus performed with 
      * {@code arguments} as arguments. Alternatively, if {@code sequence} 
      * is closed, this sequence will be closed as well.
      * 
      * @param  sequence  The sequence to copy; cannot be null.
      * @param  arguments The arguments, if any, to use in the 
      *                   initialisation of this sequence; can be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws SequenceValueException If the initialisation fails.
      * @see    #SequenceTemplate(Object...)
      */
    protected SequenceTemplate(SequenceTemplate<K,E> sequence, Object... arguments) throws SequenceValueException {
        super(sequence);
        this.index = sequence.index;
        this.size = sequence.size;
        this.current = sequence.current;
        if (!(this.closed = sequence.closed)) {
            // Open...
            try {
                // Perform specific sequence initialisation, if any. A
                // returned true value indicates that the initialisation
                // succeed, false that it failed... 
                LogFactory.getLog(this).println("Opening sequence: ", this.toString(POLICY), " -> arguments: ", arguments);
                this.closed = !this.sequenceOpen(arguments); // (abstract) operation

            // Catch all, transform, and re-throw...
            } catch (Exception e) {
                throw toThrowableType(e, SequenceValueException.class);
            }
        }
    }
    
    /**
      * @return Default implementation returns true.
      */
    public boolean bounded() {
        return true;
    }

    /**
      * @return Default implementation returns true.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * @return Default implementation returns true.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * Guard method that will throw an {@link SequenceValueException exception} 
      * if this sequence is not open, or otherwise return silently. 
      */
    protected final void ensureOpen() {
        if (this.closed) {
            throw new SequenceValueException("Sequence is closed");
        }
    }

    /**
      * Closes this sequence. <p>
      * 
      * Once this sequence has been {@link #isClosed() closed}, sequence values
      * can no longer be delivered. <p>
      * 
      * The context should <b>always</b> remember to close a sequence template
      * to ensure that it proper finalisation is made. <p>
      * 
      * Closing this sequence twice has no effect.
      * 
      * @see    #sequenceClose()
      */
    public final void close() { // Template method
        if (!this.closed) {
            Log out = LogFactory.getLog(this);
            CharSequence s = this.toString(POLICY);
            try {
                this.current = null;
                out.println("Closing sequence: ", s);
                this.sequenceClose(); // (abstract) operation
                
            } catch (Exception e) {
                out.error("Could not close sequence ", s);
                out.error(e);
                // Do not re-throw, fail gracefully instead...
                
            } finally {
                this.closed = true;
            }
        }
    }

    /**
      * Returns true if this sequence has been {@link #close() closed}, 
      * false if not. <p>
      * 
      * A closed sequence cannot deliver sequence values, and trying to
      * do so will cause a {@link SequenceValueException exception} to
      * be thrown.
      * 
      * @return True if this sequence is closed, false if not.
      * @see    #ensureOpen()
      */
    public final boolean isClosed() {
        return this.closed;
    }
    
    /**
      * @throws SequenceValueException If this sequence has been
      *         {@link #close() closed}, or if the reset fails
      *         otherwise.
      * @see    #sequenceReset()         
      */
    public final void reset() { // Template method
        this.ensureOpen();
        CharSequence s = this.toString(POLICY);
        Log out = LogFactory.getLog(this);
        
        // Calculate size after first restart...
        if ((this.size == INFINITE) && (this.index > 0)) {
            this.size  = this.index;
            out.println("Sequence size: ", s, " -> size: ", this.size);
        }
        this.index = 0;
        
        // Perform sequence specific reset, if any...
        try {
            this.sequenceReset(); // (abstract) operation
        } catch (Exception e) {
            out.error("Could not reset sequence: ", s).error(e);
            throw toThrowableType(e, SequenceValueException.class);
        }

        // Clear old current...
        this.current = null;
        // ..and fetch next one...
        this.next();
        
        // State...
        super.reset();
    }
    
    /**
      * @throws SequenceValueException If this sequence has been
      *         {@link #close() closed}.
      */
    public final E current() { // Template method
        this.ensureOpen();
        return this.current;
    }
    
    /**
      * @throws SequenceValueException If this sequence has been
      *         {@link #close() closed}, or if the next value cannot
      *         be fetched.
      *         
      * @see #nextValueOpen(Object) nextValueOpen(E)
      * @see #nextValue(Object, Object) nextValue(K, E)
      * @see #nextValueClose(Object, Object) nextValueClose(K, E)
      */
    public final E next() { // Template method
        this.ensureOpen();
        Log out = LogFactory.getLog(this);
        // For logging purposes...
        CharSequence s = this.toString(POLICY);
        
        this.state = State.NORMAL;
        
        try {
            // Old value...
            E value = this.current();
            
            out.println("Calling nextValueOpen: ", s, " -> value = ", value);
            K key = this.nextValueOpen(value); // (abstract) operation
            
            // Fetch the actual value... A null value indicates restart of 
            // this sequence...
            out.println("Calling nextValue: ", s, " -> [key, value] = [", key, ", ", value, "]");
            if ((value = this.nextValue(key, value)) == null) { // (abstract) operation

                // If not bounded, null can never be returned!
                if (!this.bounded()) {
                    throw new SequenceValueException("Value creation returned null value: " + s);
                }
                // Restart will invoke this method as well. After it
                // returns, the current value will be set...
                this.reset();
                this.state = State.RESTART;
                
                return this.current;
            }

            // Close this entry, so we can continue with the next...
            out.println("Calling nextValueClose: ", s, " -> [key, value] = [", key, ", ", value, "]");
            this.nextValueClose(key, value); // (abstract) operation

            // Update index...
            this.index++;
            
            // Update current and return value...
            return this.current = value;
            
        // Catch all, transform if needed, and re-throw...
        } catch (Exception e) {
            throw toThrowableType(e, SequenceValueException.class);
        }
    }
    
    /**
      * Returns the number of sequence elements in this sequence. <p>
      * 
      * The size is not calculated until this sequence has restarted.
      * Until then, a value of {@link Const#INFINITE} is returned. 
      * 
      * @return The size, or {@code INFINITE} if not calculated yet.
      */
    public final int size() {
        return this.size;
    }

    /**
      * Returns the index of the current sequence value in this sequence. 
      *  
      * @return The index.
      */
    protected final int index() {
        return this.index; 
    }
    
    /**
      * Simply invokes {@link #close()}.
      * 
     * @throws Throwable If the finalising fails.
      */
    protected final void finalize() throws Throwable {
        try {
            this.close();
        } finally {
            super.finalize();
        }
    }
    
    //
    // (Abstract) Operations
    //

    /**
      * Performs sub-class specific initialisation of this sequence before 
      * sequence values will be delivered, and returns true upon success. In 
      * case the initialisation could not be performed, this method returns 
      * false, or alternatively throws an exception. If the initialisation 
      * fails, this sequences is considered {@link #isClosed() closed}. <p>
      * 
      * Initialisation is performed upon {@link #SequenceTemplate(Object...) 
      * construction} of new sequences and when 
      * {@link #SequenceTemplate(SequenceTemplate, Object...) copying} existing 
      * sequences, if needed. The arguments supplied to this (abstract) operation 
      * is the varargs arguments supplied at construction time (e.g. via sub-class 
      * implementations). <p>
      *     
      * @param  arguments Arguments required by the initialisation, if any; 
      *                   never null, but can be empty.
      * @return True if the initialisation succeeded, false if it failed.
      * @throws Exception If the initialisation fails.
      * @see    #sequenceReset()
      * @see    #sequenceClose()
      */
    protected abstract boolean sequenceOpen(Object... arguments) throws Exception;

    /**
      * Performs sub-class specific {@link #reset() reset} of this 
      * sequence before sequence values will be delivered. <p>
      * 
      * Default implementation does nothing.
      * 
      * @throws Exception Not thrown by the default implementation.
      * @see    #sequenceOpen(Object...)
      * @see    #sequenceClose()
      */
    protected void sequenceReset() throws Exception {
        // Default: no special reset needed
    }
    
    /**
      * Performs sub-class specific procedures for {@link #close() closing}
      * this sequence. Once this sequence is closed, sequence values can
      * no longer be delivered. <p>
      * 
      * Default implementation does nothing.      
      *      
      * @throws Exception Not thrown by the default implementation.
      * @see    #sequenceReset()
      * @see    #sequenceOpen(Object...)
      */
    protected void sequenceClose() throws Exception {
        // Default: no closing necessary...
    }
    
    /**
      * Prepares this sequence for the {@link #next() next} sequence value to
      * be delivered based on the {@link #current() current} value supplied
      * as {@code current}. This is step 1/3 in value creation, where
      * the next steps are {@link #nextValue(Object, Object) nextValue(K, E)} and
      * {@link #nextValueClose(Object, Object) nextValueClose(K, E)}. <p>
      *
      * The returned <i>correlation key</i> will be used to identify the 
      * <i>same creation process</i> in subsequent calls to {@code nextValue}
      * and {@code nextValueClose}. 
      * 
      * @param  current The current (soon to be previous) sequence value; never null 
      *                 unless initialising this sequence (once).
      * @return The correlation key to identify the next sequence value; can be null
      *         if not used.                 
      * @throws Exception In case the opening fails.
      */
    protected abstract K nextValueOpen(E current) throws Exception;
  
    /**
      * This abstract operation creates the {@link #next() next} sequence 
      * value to be delivered. This is step 2/3 in value creation, where 
      * the previous step was {@link #nextValueOpen(Object) nextValueOpen(E)} 
      * and the last is {@link #nextValueClose(Object, Object) nextValueClose(K, E)}. <p>
      * 
      * @param  key     The <i>correlation key</i> just created by 
      *                 {@link #nextValueOpen(Object) nextValueOpen(E)}; can be null.
      * @param  current The current (soon to be previous) sequence value; never null 
      *                 unless initialising this sequence (once).
      * @return The next sequence value, which will be set as the new current 
      *         value once {@link #next()} returns; null means sequence restart, 
      *         which is only allowed if this sequence is {@link #bounded() 
      *         bounded}.
      * @throws Exception In case value construction fails.                  
      */
    protected abstract E nextValue(K key, E current) throws Exception;
  
    /**
      * Ensures that this sequence finalises the sequence value just created, if
      * required. This is step 3/3 in value creation, where
      * the previous steps were {@link #nextValueOpen(Object) nextValueOpen(E)} and
      * {@link #nextValue(Object, Object) nextValue(K, E)}. <p>
      * 
      * Default implementation does nothing. 
      * 
      * @param  key   The <i>correlation key</i> just created by 
      *               {@link #nextValueOpen(Object) nextValueOpen(E)}; 
      *               can be null.
      * @param  value The sequence value just created by 
      *               {@link #nextValue(Object, Object) nextValue(K, E)}; never null. 
      * @throws Exception In case the closing fails; not thrown by the 
      *         default implementation
      */
    protected void nextValueClose(K key, E value) throws Exception {
        // Default: no closing necessary...
    }
    
}
