/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>integer valued visitable sequence</i> allows any 
  * {@link Sequence sequence} delivering {@code Integer} values
  * to become {@link ValueVisitableSequence value} visitable using the
  * {@link SequenceValueVisitor#visitIntegerValued(Sequence, Object)} 
  * visitation method <b>and</b> is {@link TypeVisitableSequence type}
  * visitable for any known sub-interface of {@code Sequence}
  * that is implemented by the sequence. <p>
  * 
  * <b>Implementation notes</b>:<br/>
  * Like {@link StringValuedVisitableSequence}, this sequence type could
  * have declared a type parameter like {@code <E extends Number>} to
  * allow this class to be utilised with any sub-class of {@code Number},
  * e.g. not only {@code Integer}, but also {@code Long}, {@code Double}, 
  * {@code BigDecimal}, etc. But since type parameters are erased at
  * runtime, this class would have to call a single visitation method
  * like {@code visitNumberValued(Sequence<? extends Number>, ..)}, 
  * and the possibility to distinguish on a specific numeric value type 
  * delivered by a sequence is lost (unless of cause one would be so careless
  * as to use the {@code instanceof} operator on actual sequence values,
  * which is a really bad idea).  
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DateValuedVisitableSequence
  * @see    StringValuedVisitableSequence
  * @see    ReflectiveVisitableSequence
  */
@Participant("ConcreteElement")
public class IntegerValuedVisitableSequence extends AbstractVisitableSequence<Integer> { 

    /**
      * Constructor.
      * 
      * @param sequence The decorated sequence that will become visitable;
      *                 cannot be null.
      * @throws NullPointerException If {@code sequence} is null.                  
      */
    public IntegerValuedVisitableSequence(Sequence<Integer> sequence) {
        super(sequence);
    }

    /**
      * {@inheritDoc} 
      *  
      * The decorated sequence is <b>not</b> copied. 
      */
    public IntegerValuedVisitableSequence copy() {
        return new IntegerValuedVisitableSequence(this.getSequence(false));
    }
    
    /**
      * @see SequenceValueVisitor#visitIntegerValued(Sequence, Object)
      */
    public <P> void accept(SequenceValueVisitor<P> visitor, P argument) {
        visitor.visitIntegerValued(this.getSequence(false), argument); // no copy
    }

}
