/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import java.lang.reflect.Method;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.HexDigitSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>reflective visitable sequence</i> decorates a 
  * {@link Sequence sequence} to become {@link ValueVisitableSequence value} 
  * <b>and</b> {@link TypeVisitableSequence type} visitable. <p>
  * 
  * <b>Implementation notes</b>:<br/>
  * The {@linkplain dk.rode.thesis.decorator Decorator} pattern is used
  * to implement this visitable sequence. <p>
  * 
  * The type visitable functionality is inherited from 
  * {@link AbstractVisitableSequence}, while the value visitable
  * functionality is implemented via reflection in this class
  * using {@link TypeLiteral type literals}. <p>
  * 
  * In case the decorated sequence is generic itself, i.e. the type 
  * of values to be delivered is specified using a type parameter, 
  * the value type cannot be inferred. In such cases, the 
  * {@link SequenceVisitor#visitUnknown(Sequence, Object)} value
  * visitation method will be used. The {@code visitUnknown} method
  * will also be used in case the type can be inferred, but is not
  * supported by the {@link SequenceValueVisitor} interface, for 
  * example a sequence delivering {@link java.math.BigDecimal
  * java.math.BigDecimal} values. Also note that only sequences delivering 
  * {@link java.lang.String java.lang.String} values can be inferred to the 
  * {@link SequenceValueVisitor#visitStringValued(Sequence, Object)} visitation
  * method, though it accepts any sequence delivering 
  * {@link java.lang.CharSequence java.lang.CharSequence} values. Hence,
  * decorating the {@link HexDigitSequence}, for example, will cause the
  * {@code visitUnknown} method to be used because no {@code visitCharSequenceValued}
  * visitation method is defined in the {@code SequenceValueVisitor} 
  * interface. An alternative to allow more flexible visitation method
  * identification could be to use the {@link Executor} annotation instead. <p>
  * 
  * While reflection is used to invoke the value visitation method, type
  * safety is still guaranteed as the invocation is made privately, using
  * parameter types and values known to be correct. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DateValuedVisitableSequence
  * @see    IntegerValuedVisitableSequence
  * @see    StringValuedVisitableSequence
  * 
  * @see    TypeLiteral
  *    
  * @param  <E> The type of values delivered by this visitable sequence.
  */
@Participant("ConcreteElement")
public class ReflectiveVisitableSequence<E> extends AbstractVisitableSequence<E> { 

    /**
      * Default value visitation method, i.e. visit unknown type. <p>
      * 
      * Never null.
      */
    private final static Method visitUnknown;
    static {
        try {
            visitUnknown = SequenceVisitor.class.getMethod("visitUnknown", Sequence.class, Object.class);
        } catch (Exception e) {
            throw new Error(e);
        }
    }
    /**
      * The value visitation method used by this sequence. <p>
      * 
      * Never null.
      */
    private final Method visitationMethod;
    
    /**
      * Constructor. <p>
      * 
      * In case {@code sequence} is generic or delivers sequence values
      * that cannot be visited by a {@link SequenceValueVisitor value}
      * visitor, the {@link SequenceVisitor#visitUnknown(Sequence, Object)}
      * value visitation method will be used.  
      *  
      * @param sequence The decorated sequence that will become visitable;
      *                 cannot be null.
      * @throws NullPointerException If {@code sequence} is null.                  
      */
    public ReflectiveVisitableSequence(Sequence<E> sequence) {
        super(sequence);
        TypeLiteral<?> type = TypeLiteral.create(sequence.getClass());
        Method method = null;
        String name = null;
        try {
            // This may intentionally return null (if "sequence" is generic)
            // and thus generate a null pointer below...
            Class<?> clazz = type.getRawType();
            // Invariant: all visitation methods named "visit" + simple type name + "Valued"...
            name = "visit" + clazz.getSimpleName() + "Valued";
            // Object class can be used because of type erasure...
            method = SequenceValueVisitor.class.getMethod(name, Sequence.class, Object.class);
            // Found visitation method!
            LogFactory.getLog(this).println("Found value visitation method for: ", sequence.toString(SequencePolicy.ID), " -> ", name);
            
        // Default: visitUnknown...
        } catch (Exception e) {
            Log out = LogFactory.getLog(this);
            if (name != null) {
                out.warn("Cannot find visitation method named: ", name);
            }
            out.warn("Using default value visitation type for: ", sequence.toString(SequencePolicy.ID), " -> ", visitUnknown.getName());
            method = visitUnknown;
        }
        // Will be public because declared in a public interface...
        this.visitationMethod = method;
    }

    /**
      * {@inheritDoc} 
      *  
      * The decorated sequence is <b>not</b> copied. 
      */
    public ReflectiveVisitableSequence<E> copy() {
        return new ReflectiveVisitableSequence<E>(this.getSequence(false)); // no copy
    }
    
    /**
      * 
      */
    public <P> void accept(SequenceValueVisitor<P> visitor, P argument) {
        LogFactory.getLog(this).println("Dispatching ", visitor.getClass().getSimpleName(), ".", this.visitationMethod.getName(), "...");
        // P ensures correct type of argument to "visitor"!
        Reflection.invoke(visitor, this.visitationMethod, this.getSequence(false), argument);
    }

}
